# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Alexis Sukrieh
# - Stephane Jourdois
# - Philippe Blayo
# - Antoine Hulin
package FDN::Sql::Basic;

use strict;
use Exporter ;

our @ISA = ( 'Exporter' ) ;

our @EXPORT = qw (
	get_next
	get_current
	get_stats
	dochar
	dolike
	dodate
	do_null_or_char
	do_null_or_date
	do_null_or_int
	dobool
	array_has_no_rows

	lock_soft
	unlock_soft
	is_locked_soft

	lock_tables
	unlock_tables

	__par_id
	__par_parent
	__id_par_nom
	__id_par_clef
	__id_par_clefs
	__champ_par_id
	__tous
	__tous_h
	__tous_cache
	__par_nom
	__par_nom_like
	__par_champ_ext
	__par_champ_test
	__par_clef
	__par_clefs
	__max
	__min
	__new
	__update
	__set_tbl_val
	__update_relation
	__set_rel_val
	__set_relation
	__set_self_relation
	__get_self_relation
	__get_self_relation_ordonnee
	__get_relation
	__get_relation_ordonnee
	__par_multi_rel
	__get_related_a
	__get_related_b
	__get_properties
	__del_properties
	__set_properties
	__update_properties
	__del
	__del_par_clef
	__del_par_clefs
	__del_par_champ
	__del_relation
	__par_date
	__rotate_table

	__count

	__add_index
	__drop_index

	__jointure

	__open_all
	__read_all
	__close_all

	__open_par_champ_ext
	__read_par_champ_ext
	__close_par_champ_ext

	__make_h
	__make_cache
	);

our @EXPORT_OK = qw (
	__return_none
	__return_errors
	__return_one
	__return_all
	__return_all_or_one
	__return_count
	__return_one_col
	__return_all_col
	__return_all_or_one_col
	);
our %EXPORT_TAGS = ();

use FDN::Common::Utils;
use FDN::Common::Logger;
use FDN::Common::Constantes;
use FDN::Sql::DB_Access;

# Section 1: La fonction de debug

our $debug = 0;
sub __debug($)
{
	my $msg = shift;
	$msg = "[vide]" unless defined $msg;
	debug $msg if $debug;
}

# Section 2: La fonction pour avancer dans les sequences
#

sub __seq_name($$) {
	my ( $dbh, $what ) = @_;
	if ( ! $dbh->{'suffix'} ) {
		return $what;
	}
	my $suffix = $dbh->{'suffix'};
	if ( $what =~ /_[A-Z]+$/ ) {
		$what =~ s/(_[A-Z]+)$/$suffix$1/;
	} else {
		$what .= $suffix;
	}
	return $what;
}

sub get_next( $$ ) {
	my ( $dbh,$what ) = @_;
	my $id = sequence_data($dbh,__seq_name($dbh,$what));
	if ( ! defined $id ) {
		info "No next ".$what;
	} else {
		__debug "Got next ".$what;
	}
	return $id;
}

sub get_current($$) {
	my ( $dbh, $what ) = @_;
	return sequence_data($dbh, __seq_name($dbh,$what), 'current');
}

sub get_stats($$) {
	my ( $dbh, $what ) = @_;
	return sequence_data($dbh, __seq_name($dbh,$what), 'stats');
}

# Section 3: Les fonctions de mise en forme des arguments
# dochar: fait une chaine de caracteres encadres par " en quotant
#   proprement (re-exporte de Sql::DB_Access)
# dolike: comme dochar, mais ajoute des % autour
# dodate: mets en forme une date en prenant pour cas particulier 'now'
# do_null_or_char: comme dochar, sauf si l'argument est indefini ou
#   'null', permet de supprimer une valeur.
# do_null_or_date: idem (ne test pas la validit de la date)
# do_null_or_int: idem, reconnait 'null' en lieu et place d'un entier
#   s'il le faut
# do_bool: renvoit le modele 'CK' de booleen: -1 pour indefini, 0
#   pour faux, 1 pour vrai.

sub dolike( $ ) { return dochar('%'.$_[0].'%');}

sub dodate( $ ) {
	my $date = $_[0];
	if(defined($date) and $date ne '' and $date ne 0) {
		if ( $date =~ /now/i ) {
			return 'now()' ; # la date de tout de suite
		} else {
			$date =~ s/'/\\'/g;
			return "'".$date."'";
		}
	} else {
		return "'0000-00-00 00:00:00'" ;
	}
}

sub do_null_or_char( $ ){
	my $string = $_[0];
	if ( !defined($string) or $string =~ /^null$/i ) {
		return 'NULL';
	} else {
		return dochar( $string );
	}
}

sub do_null_or_date( $ ) {
	my $string = $_[0];
	if ( !defined($string) or $string =~ /^null$/i ) {
		return 'NULL';
	} else {
		return dodate( $string );
	}
}

sub do_null_or_int( $ ){
	my $string = $_[0];
	if ( defined($string) and $string =~ /^-?[0-9]+$/ ){
		return "'$string'";
	} else {
		return 'NULL';
	}
}

sub dobool( $ ) {
	my ( $val ) = @_;
#	__debug 'dobool '.$val;
	if ( !defined($val) or $val eq '-1' ) {
		return "'-1'";
	} elsif($val) {
		return "'1'";
	} else {
		return "'0'";
	}
}

sub __return_none() {
	if ( wantarray ) {
		return ();
	} else {
		return undef;
	}
}

sub __return_errors($) {
	my ( $res ) = @_;
	my $query = 'undef';
	$query = $res->{'query'} if defined $res->{'query'};
	if ( $res->{'err_db'} ) {
		error "There was an error with query $query";
	}
	my $count = 'undef';
	$count = $res->{'count_rows'} if defined $res->{'count_rows'};
	__debug "Found $count lines of data from query $query";
}

sub __return_one($) {
	my ( $res ) = @_;
	__return_errors($res);
	if ( ! defined $res->{'count_rows'} || $res->{'count_rows'} == 0 || $res->{'err_db'} ) {
		return undef;
	} else {
		return $res->{'rows'}->[0];
	}
}

sub __return_all($) {
	my ( $res ) = @_;
	__return_errors($res);
	if ( ! defined $res->{'count_rows'} || $res->{'count_rows'} == 0 || $res->{'err_db'} ) {
		return ();
	} else {
		return @{$res->{'rows'}};
	}
}

sub __return_all_or_one($) {
	my ( $res ) = @_;
	if ( wantarray ) {
		return __return_all($res);
	} else {
		return __return_one($res);
	}
}

sub __return_count($) {
	my ( $res ) = @_;
	__return_errors($res);
	if ( $res->{'ok'} && defined $res->{'count_rows'} ) {
		return $res->{'count_rows'};
	} else {
		return undef;
	}
}
		

sub __return_one_col($$) {
	my ( $res, $col ) = @_;
	__return_errors($res);
	if ( $res->{'count_rows'} == 0 || $res->{'err_db'} ) {
		return undef;
	} else {
		return $res->{'rows'}->[0]->{$col};
	}
}

sub __return_all_col($$) {
	my ( $res, $col ) = @_;
	__return_errors($res);
	if ( $res->{'err_db'} || ! exists $res->{'count_rows'} || $res->{'count_rows'} == 0 ) {
		return ();
	} else {
		my @ret;
		for my $i ( @{$res->{'rows'}} ) {
			push @ret, $i->{$col};
		}
		return @ret;
	}
}

sub __return_all_or_one_col($$) {
	my ( $res, $col ) = @_;
	if ( wantarray ) {
		__return_all_col($res, $col);
	} else {
		__return_one_col($res, $col);
	}
}

# From a list of rows, return a hash base on a key (which is supposed unique)
sub __make_h($$) { # Test
	my ( $l, $k ) = @_;
	my $res = {};
	return $res if ( ref($l) ne 'ARRAY' );
	for my $in ( @$l ) {
		$res->{$in->{$k}} = $in;
	}
	return $res;
}

# From a list of rows, return a hash base on a key (which is supposed multiple)
sub __make_h_multi($$) {
	my ( $l, $k ) = @_;
	my $res = {};
	return $res if ( ref($l) ne 'ARRAY' );
	for my $in ( @$l ) {
		$res->{$in->{$k}} = [] if ( !defined $res->{$in->{$k}} );
		push @{ $res->{$in->{$k}} }, $in;
	}
	return $res;
}

sub __make_queries_hash($) {
	my ( $queries ) = @_;
	my $h = {};
	for my $q ( @$queries ) {
		my ( $id, $clef ) = @$q;
		$h->{$clef} = $id;
	}
	return $h;
}

sub __sql_from_queries($$) {
	my ( $table, $queries ) = @_;

	if ( ! defined $table ) {
		error "Missing parameter";
		return undef;
	}

	if ( ! defined $queries ) {
		error "Invalid queries";
		return undef;
	}

	if ( ref($queries) eq 'ARRAY' ) {
		error "Deprecated syntax. Should use a pretty hash-ref.";
		$queries = __make_queries_hash($queries);
	}

	if ( ref($queries) ne 'HASH' ) {
		error "Invalid queries ";
		return undef;
	}

	my $sql = [];
	my ( $clef, $id );
	while ( ( $clef, $id ) = each %$queries ) {
		if ( $id eq 'NULL' ) {
			push @$sql, $table.'.'.$clef.' IS NULL';
		} else {
			push @$sql, $table.'.'.$clef."=".dochar($id);
		}
	}
	return $sql;
}



# From a list of rows, return a cache, that is a multiple index hash, each index
# is supposed unique if it is a scalar, and multiple if it is as ARRAYREF
# 
# E.g.:
# $cache = __make_cache(tous_machins(), id => 'MACHIN_ID', name => 'MACHIN_NAME' )
# returns a two levels hash table, basically like that:
# $cache = {
#	'id' => {
#		1 => {
#			'MACHIN_ID' => 1,
#			'MACHIN_NAME' => 'toto'
#		}, ...
#	},
#	'name' => {
#		'toto' => $cache->{id}->{1},
#	}
# };
# $cache = __make_cache(tous_machins(), id => 'MACHIN_ID', name => [ 'MACHIN_NAME' ] )
# Same, but supposes the 'name' is not a unique key.
sub __make_cache($%) {
	my ( $l, %keys ) = @_;
	my $res = {};
	for my $k ( keys %keys ) {
		if ( ref($keys{$k}) eq 'ARRAY' ) {
			$res->{$k} = __make_h_multi($l, $keys{$k}->[0]);
		} else {
			$res->{$k} = __make_h($l, $keys{$k});
		}
	}
	return $res;
}


# Section 4: Recherche d'objet par id
#

sub __par_id( $$$ ) {
	my ( $dbh, $what, $id ) = @_;
	return undef if ! defined $id;
	my $suffix = '';
	$suffix = $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $table  = uc($what).$suffix;
	my $tmp = get_data_reuse ( $dbh,
			[ $table ],
			[ $table.'.'.uc($what).'_ID='.$id ] );
	__debug "Recherche ".$what." par id (reuse): ".$tmp->{'query'};
	if ( ! defined $tmp->{'count_rows'} ) {
		__debug "Pas trouve ".$what.", echec de la requete (count_rows non defini dans la reponse)";
		return undef;
	} elsif ( $tmp->{'count_rows'} != 1 ) {
		__debug "Pas trouve ".$what.", puisque ".$tmp->{'count_rows'}." ligne(s)!";
		return undef;
	} else {
		return $tmp->{'rows'}[0];
	}
}


# Section 6: Recherche d'un id par nom (et d'un seul)
# avec un test de type "=" par defaut et possibilit de faire un like
# avec '~', '!~' voire meme 'IS NULL'
# __id_par_nom($dbh, 'TBL', 'toto', '~', 'TBL_FIELD');
# __id_par_nom($dbh, 'TBL', 'toto', '=');
# __id_par_nom($dbh, 'TBL', 'toto');
sub __id_par_nom( $$$;$$ ) {
	my ( $dbh, $what, $nom, $testtype ,$whatnom) = @_;
	my $clause;
	if (!defined($whatnom)) {
		$whatnom = $what.'_NOM';
	}
	if (!defined($testtype)) {
		$testtype = '=';
	}
	if ( $testtype eq '=' ){
		$clause = $whatnom.' = '.dochar($nom);
	} elsif( $testtype eq '~' ) {
		$clause = $whatnom.' ~ '.dolike($nom);
	} elsif( $testtype eq '!~' ) {
		$clause = $whatnom.' !~ '.dolike($nom);
	} elsif( $testtype eq 'IS NULL' ){
		$clause = $whatnom.' IS NULL ';
	}
	my $tmp = get_data_reuse ( $dbh, [ $what.'_ID' ], [ $clause ] );
	return __return_all_or_one_col($tmp,$what.'_ID');
}

# Section 6 bis : Recherche d'une liste d'id par clef etrangere
sub __id_par_clef( $$$$;$ ) {
	my ( $dbh, $what, $id, $clef, $order ) = @_;
	return __id_par_clefs($dbh, $what, { $clef => $id }, $order );
}

sub __id_par_clefs( $$$;$ ) {
	my ( $dbh, $what, $queries, $order ) = @_;

	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $sql = __sql_from_queries($table, $queries);
	if ( ! defined $sql ) {
		error "Invalid parameters";
		return __return_none();
	}

	if ($order) {
		$order = [ $order ];
	} else {
		$order = 'none';
	}

	my $tmp = get_data_reuse ( $dbh, [ $table.'.'.$what.'_ID' ], $sql, $order );
	__debug "__id_par_clefs with query ".$tmp->{'query'};
	return __return_all_or_one_col($tmp,$what.'_ID');
}



# Section 7: Recherche d'un ou plusieurs objets par nom avec un like
# 

sub __tous($$) {
	my ( $dbh, $what ) = @_;
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $tmp = get_data ( $dbh, [$table], [] );
	return __return_all_or_one($tmp);
}

sub __tous_h($$;$) {
	my ( $dbh, $what, $key ) = @_;
	$key = $what.'_ID' if ( ! defined $key );
	my $tmp = [ __tous($dbh,$what) ];
	return __make_h($tmp,$key);
}

sub __tous_cache($$%) {
	my ( $dbh, $what, %keys ) = @_;
	my $tmp = [ __tous($dbh,$what) ];
	return __make_cache($tmp,%keys);
}

sub __par_nom_like( $$$;$ ) {
	my ( $dbh, $what, $nom, $whatnom ) = @_;
	if (!defined($whatnom)) {
		$whatnom = $what.'_NOM';
	}
	if (! defined $what or ! defined $whatnom or ! defined $nom) {
		error "Missing parameter";
		return __return_none();
	}
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $tmp = get_data ( $dbh, [ $table ], [ $whatnom."~".dolike( $nom) ] );
	return __return_all_or_one($tmp);
}

sub __par_nom($$$;$) {
	my ( $dbh, $what, $nom, $whatnom ) = @_;
	$whatnom = $what.'_NOM' if ( ! defined ($whatnom) );
	if ( ! defined $what or ! defined $whatnom or ! defined $nom ) {
		error "Missing parameter";
		return __return_none();
	}
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $tmp = get_data ( $dbh, [ $table ], [$whatnom.'='.dochar($nom) ] );
	return __return_all_or_one($tmp);
}

# Section 7.1: Recherche d'une ou plusieurs entres avec une clause de type
# 	       "~", "!~", "=" ou "IS NULL" ou <, >, !=

my $known_tests = {
	'=' => 1,
	'!=' => 1,
	'~' => 1,
	'!~' => 1,
	'<' => 1,
	'>' => 1,
	'<=' => 1,
	'>=' => 1,
	'IS NULL' => '0null',
	'IS NOT NULL' => '0null',
	'gt' => '0rw>',
	'lt' => '0rw<',
	'le' => '0rw<=',
	'ge' => '0rw>=',
	'ne' => '0rw!=',
	'eq' => '0rw=',
};

=head1 __rewrite_condition($;$$)

=head2 Paramtres

=over

=item * $clause: ARRAYREF [ nom, test, val ]

nom: le champ contre lequel on effectue le test [[TABLE.]TABLE_]CHAMP

test: le test  effectuer =, !=, ~, !~, <, >, <=, >=, IS NULL, IS NOT NULL,
la syntaxe perl est aussi accept: gt, lt, le, ge, ne, eq

val: la valeur  tester. (undef pour les tests IS NULL et IS NOT NULL)

=item * $table: (optionnel)

Nom de table utilis pour complter le nom du champ si celui-ci est imcomplet
(pas de TABLE_ au dbut du nom)

=item * $suffix: (optionnel)

Sufix utilis dans les bases multi-instances

=back

=head2 Utilisation

	__rewrite_condition(['USER.PERSON_ID','=','42']); # USER.PERSON_ID = 42
	__rewrite_condition(['PERSON_ID','=','42']); # PERSON.PERSON_ID = 42
	__rewrite_condition(['PERSON_ID','IS NULL',undef]); # PERSON.PERSON_ID IS NULL
	__rewrite_condition(['ID','IS NULL',undef], 'PERSON'); # PERSON.PERSON_ID IS NULL
	__rewrite_condition(['USER.PERSON_ID','=','42'], undef, '123456'); # USER123456.PERSON_ID = 42
	__rewrite_condition(['PERSON_ID','=','42'], undef, '123456'); # PERSON123456.PERSON_ID = 42
	__rewrite_condition(['ID','IS NULL',undef], 'PERSON', '123456'); # PERSON123456.PERSON_ID IS NULL

=cut
sub __rewrite_field($;$$) {
	my ( $nom, $table, $suffix ) = @_;
	if ( $nom !~ /^(?:(?:([A-Z][A-Z0-9]*\.))?([A-Z][A-Z0-9]*)_)?([^._]+)$/ ) {
		error "Invalid field name: $nom";
		return undef;
	}
	my ( $tbl, $ftbl, $fld ) = ( $1, $2, $3 );
	if ( ! defined $tbl ) {
		$tbl = '';
	}
	if ( ! defined $ftbl ) {
		if ( ! defined $table ) {
			error "Ambiguous field name: $nom";
			return undef;
		}
		$ftbl = $table;
	}
	if ( defined $suffix ) {
		if ( $tbl eq '' ) {
			$tbl = $ftbl.$suffix.'.';
		} elsif ( $tbl !~ s/\.$/$suffix./ ) {
			error "Can't concat suffix before dot in : '$tbl'";
			return undef;
		}
	}
	return $tbl.$ftbl.'_'.$fld;
}
sub __rewrite_condition($;$$);
sub __rewrite_condition($;$$) {
	my ( $clause, $table, $suffix ) = @_;
	if ( ref($clause) ne 'ARRAY' ) {
		error "The condition should be an array ref";
		return undef;
	}
	if ( ref($clause->[0]) eq 'ARRAY' ) {
		my $res = [];
		for my $cl ( @$clause ) {
			my $tmp = __rewrite_condition($cl, $table, $suffix);
			push @$res, $tmp;
		}
		return undef if ! @$res;
		return join ' or ', @$res;
	}
	if ( $#$clause != 2 ) {
		error "The condition should have precisely 3 entries: [ field, test, val ]";
		return undef;
	}
	if ( ! $known_tests->{$clause->[1]} ) {
		error "The condition '".$clause->[1]."' is not valid";
		return undef;
	}
	if ( ! defined $clause->[0] ) {
		error "In the condition, the first entry (field) cannot be omited";
		return undef;
	}
	my ($nom, $test, $val) = @$clause;
	$nom = __rewrite_field($nom, $table, $suffix);
	return undef if !defined $nom;
	if ( defined $val && $val !~ /^['"]/ ) {
		my $tmp = __rewrite_field($val, $table, $suffix) if $val =~ /^(?:(?:([A-Z][A-Z0-9]*\.))?([A-Z][A-Z0-9]*)_)([^._]+)$/;
		$val = $tmp if defined $tmp;
	}

	my $res;
	if ( $known_tests->{$test} eq '1' ) {
		$res = $nom.' '.$test.' '.$val;
	} elsif ( $known_tests->{$test} =~ /^0rw(.*)$/ ) {
		$res = $nom.' '.$1.' '.$val;
	} elsif ( $known_tests->{$test} =~ /^0null/ ) {
		$res = $nom.' '.$test;
	} else {
		error "Unknown test type: $test";
		return undef;
	}
	return $res;
}

=head1 __rewrite_conditions($;$$)

Appel en boucle __rewrite_condition() pour traiter chaque condition.

=cut
sub __rewrite_conditions($;$$) {
	my ( $conds, $table, $suffix ) = @_;
	return [] if ( ! defined $conds );
	if ( ref($conds) ne 'ARRAY' ) {
		error "Invalid conditions";
		return undef;
	}
	my $res = [];
	for my $c ( @$conds ) {
		my $r = __rewrite_condition($c,$table, $suffix);
		if ( ! defined $r ) {
			error "There was an invalid condition";
			return undef;
		}
		push @$res, $r;
	}
	return $res;
}

# 10-19 -> offset 10, limit 10
# 10+10 -> offset 10, limit 10
# +10   -> offset  0, limit 10
# -9    -> offset  0, limit 10
# 10+   -> offset 10, limit undef
# 10-   -> offset 10, limit undef
sub __rewrite_range_option($$) {
	my ( $option, $opts ) = @_;
	my ( $offset, $type, $limit );
	if ( $option =~ /^(\d*)([-+])(\d*)$/ ) {
		( $offset, $type, $limit ) = ( $1, $2, $3 );
		if ( ! $offset ) {
			$offset = 0;
		}
		if ( $type eq '-' ) {
			if ( $limit ne '' ) {
				$limit = $limit + 1 - $offset;
			} else {
				$limit = undef;
			}
		} else {
			if ( ! $limit ) {
				$limit = undef;
			}
		}
	} else {
		return FALSE;
	}
	# Should be large enough to mean no limit.
	$limit = '18446744073709551615' if ( ! defined $limit );
	$opts->{'offset'} = $offset;
	$opts->{'limit'} = $limit;
	$opts->{'range'} = [ "limit($limit)", "offset($offset)" ];
	__debug "Detected range limit=$limit offset=$offset";
	return TRUE;
}

# up:DATE		-> order by TABLE_DATE ASC
# orderup(DATE) 	-> order by TABLE_DATE ASC
# down:TABLE_DATE	-> order by TABLE_DATE DESC
# orderdown(DATE)	-> order by TABLE_DATE DESC
sub __rewrite_order_option($$) {
	my ( $option, $opts ) = @_;
	if ( $option =~ /^(down|up):(.*)$/ ) {
		push @{$opts->{'order'}}, "order$1($2)";
		__debug "Detected order$1($2).";
		return TRUE;
	} elsif ( $option =~ /^order(up|down)\(.*\)$/ ) {
		__debug "Detected DB_Access style $option.";
		push @{$opts->{'order'}}, $option;
		return TRUE;
	} else {
		error "Invalid order $option.";
		return FALSE;
	}
}

sub __rewrite_archives_option($$) {
	my ( $option, $opts ) = @_;
	unless ($option =~ /(?:(\d{8})(\d{6})?|)([<>])(?:(\d{8})(\d{6})?|)/) {
		error "archives incorrect ($option)";
		return FALSE;
	}
	my ($datemin, $datemax) = (0, 0);
	my ( $dmin, $hmin, $order, $dmax, $hmax ) = ( $1, $2, $3, $4, $5 );
	$hmin = '000000' if ( ! defined $hmin );
	$hmax = '235959' if ( ! defined $hmax );
	$datemin = "$dmin$hmin" if ( defined $dmin );
	$datemax = "$dmax$hmax" if ( defined $dmax );
	$opts->{'datemin'} = $datemin;
	$opts->{'datemax'} = $datemax;
	if ( $order eq '>' ) {
		$opts->{'dateorder'} = 'down:ROTATION_DATE';
		$opts->{'datelast'}  = 'before';
	} else {
		$opts->{'dateorder'} = 'up:ROTATION_DATE';
		$opts->{'datelast'}  = 'after';
	}
	return TRUE;
}

sub __find_archives_table($) {
	my ( $opts ) = @_;
	my $table = $opts->{'table'};
	# On commence par chercher les tables dont la date de rotation est
	# comprise dans l'intervalle de recherche
	my @conds = ( ['ROTATION_TABLE', '=', dochar($table) ] );
	if ( $opts->{'datemin'} ) {
		__debug "There is a date min";
		push @conds, [ 'ROTATION_DATE', '>=', dodate($opts->{'datemin'}) ];
	}
	if ( $opts->{'datemax'} ) {
		__debug "There is a date max";
		push @conds, [ 'ROTATION_DATE', '<=', dodate($opts->{'datemax'}) ];
	}
	my @tables;
	my $dbh = $opts->{'dbh'};
	@tables = __par_champ_ext( $dbh, 'ROTATION_DEST', \@conds,
			{ 'order' => $opts->{'dateorder'} } );
	@tables = map { $_->{'ROTATION_DEST'} } @tables;
	__debug "Les tables impactees (etape 1): ".join(' ',@tables);
	my $last;
	if ( ! $opts->{'datemax'} ) {
		# Il n'y a pas de borne sup dans le temps
		# => Il nous faut egalement ajouter la table courante
		$last = $table;
	} else {
		# Il nous manque la premiere table a avoir ete tournee apres datemax.
		my $tmp = __par_champ_ext( $dbh, 'ROTATION_DEST',
					[ [ 'ROTATION_TABLE', '=', dochar($table) ],
					  [ 'ROTATION_DATE' , '>', dodate($opts->{'datemax'}) ] ],
					{ 'order' => 'up:ROTATION_DATE' } );
		if ( defined $tmp ) {
			$last = $tmp->{'ROTATION_DEST'};
		} else {
			# Il n'y a pas eu de rotation suivante.
			$last = $table;
		}
	}
	if ( $opts->{'datelast'} eq 'after' ) {
		push @tables, $last;
	} else {
		unshift @tables, $last;
	}
	# S'il n'y a qu'une table a explorer, ne pas faire croire qu'il y en
	# a plusieurs.
	if ( $#tables > 0 ) {
		$opts->{'tables'} = \@tables;
	} else {
		$opts->{'table'} = $tables[0];
	}
}

sub __rewrite_option($$$) {
	my ( $option, $val, $opts ) = @_;
	if ($option eq "range" ) {
		return __rewrite_range_option($val,$opts);
	}
	if ($option eq "order" ) {
		if ( ref($val) ne 'ARRAY' ) {
			return __rewrite_order_option($val,$opts);
		} else {
			my $res = 1;
			for my $v ( @$val ) {
				$res &&= __rewrite_order_option($v,$opts);
			}
			return $res;
		}
	}
	if ($option eq "group" ) {
		$opts->{'group'} = [];
		if ( ref($val) eq 'ARRAY' ) {
			push @{$opts->{'group'}}, $val;
		} else {
			$opts->{'group'} = [ $val ];
		}
		error "NOT YET IMPLEMENTED GROUP";
	}
	if ($option eq "archives" && $opts->{'archives-ok'} ) {
		if ( ! __rewrite_archives_option($val,$opts) ) {
			return FALSE;
		}
		__find_archives_table($opts);
		if ( $opts->{'tables'} ) {
			__debug "tables (+suprieure) = ".join(' ',@{$opts->{'tables'}});
		} else {
			__debug "Une seule table suffira: ".$opts->{'table'};
		}
		return TRUE;
	}
	if ($option eq "misc") {
		if ( ref($val) eq 'ARRAY' ) {
			$opts->{'misc'} = $val;
		} else {
			$opts->{'misc'} = [ $val ];
		}
		__debug "Miscelaneous options ".join('/',@{$opts->{'misc'}});
		return TRUE;
	}
}

sub __rewrite_options($;$) {
	my ( $options, $opts ) = @_;
	return [] if ( ! defined $options );
	return [] if ( $options eq '' );
	$opts = {} if ( ! defined $opts );
	if ( ref($opts) ne 'HASH' ) {
		alert "Opts should be a hash ref";
		return undef;
	}
	if ( ref($options) ne 'HASH' )  {
		error "Options should be a hash ref";
		return undef;
	}
	$opts->{'limit'} = undef;
	$opts->{'offset'} = undef;
	$opts->{'order'} = [];
	$opts->{'misc'} = [];
	while (my ($option, $valeur) = each %$options) {
		if ( ! __rewrite_option($option, $valeur, $opts ) ) {
			error "Option invalide: $option -> $valeur";
			return FALSE;
		}
	}
	$opts->{'options'} = [ ];
	push @{$opts->{'options'}}, @{$opts->{'group'}} if (defined $opts->{'group'});
	push @{$opts->{'options'}}, @{$opts->{'order'}} if (defined $opts->{'order'});
	push @{$opts->{'options'}}, @{$opts->{'range'}} if (defined $opts->{'range'});
	push @{$opts->{'options'}}, @{$opts->{'misc'}}  if (defined $opts->{'misc'});
	return $opts->{'options'};
}

# La fonctions __par_champ_ext
# __par_nom est un cas particulier de cette fonction
#
# XXX documentation deprecated XXX
# Cette fonction prend les arguments suivants :
# $dbh     : handle de connection  la base de donnes, renvoy par FDN::DBC
# $what    : ce qu'il est demand de retourner sous la forme [TABLE.]TABLE[_CHAMP] (ce qui est entre crochets est optionnel)
# $queries : une rfrence optionnelle sur un tableau de tableaux sous la forme [ [ $val, $nom, $test ], [ $val, $nom, $test ], ... ]
#     $nom  : le champ contre lequel on effectue le test (CHAMP) (pas ncessaire de mettre TABLE_, mais accept)
#     $val  : la valeur  tester (scalaire)
#     $test : le test  effectuer (un parmi =, ~, !~, IS NULL, <, >, !=) # TODO
# $options : une rfrence optionnelle sur un hash contenant les cls optionnelles suivantes :
#     range    : scalaire de la forme \d*[-+]\d*
#         le premier chiffre optionnel est la valeur de l'offset
#         le caractre sparateur est - pour signifier "de " ou + pour signifier "n enregistrements  partir de l'offset"
#         le deuxieme chiffre est soit la limite haute du range, soit le nombre d'enregistrements  retourner
#         Notes :
#             un offset sans clause ORDER sur une valeur unique peut donner des rsultats tranges...
#             la limite est force  1 si le contexte d'appel de la fonction est scalaire
#     order    : scalaire de la forme (up:|down:)?CHAMP (pas ncessaire de mettre TABLE_, mais accept)
#     		 OU BIEN rfrence sur un tableau de la mme forme
#         up/down permet d'ordonner le classement pour ce champ
#         CHAMP donne le champ sur lequel on effectue le classement.
#         Notes :
#             Le classement est effectu table par table en SQL, mme en combinaison avec l'option archives.
#             Le tri ne renverra donc peut-tre pas ce qui tait attendu...
#     archives : scalaire sous la forme DATE[<>]DATE
#         DATE est de la forme (aaaammjj(hhmmss)?)?
#         si hhmmss n'est pas prsent, il est 000000 pour la premire date, et 235959 pour la deuxieme
#         si la date n'est pas prsente, la premire est 0, et la deuxieme est la date courante
#         Effectue une recherche dans les tables tournes (cf. __rotate_table) dans le range de dates spcifi.
#         le caractre central indique l'ordre dans lequel les tables seront lues
#         (de la plus ancienne  la plus rcente pour '<' ou le contraire pour '>')
#         Attention : La date la plus ancienne doit toujorus se trouver  gauche ! (FIXME)
#         Notes :
#             ARCHIVES permet de chercher dans les tables correspondant aux dates. Aucun test n'est effectu
#                 sur un champ donn pour la correspondance au range. Ajoutez des valeurs au tableau queries pour a.
#             ORDER sera fait en mmoire, et pas transmis  la base de donnes.
#             La recherche aura lieu dans une seule table en contexte scalaire
#         
=head1 __par_champ_ext($$;$$)

Prend les mme paramtes que __open_par_champ_ext() et effectue l'enchainement:
__open_par_champ_ext(), __read_par_champ_ext(), __close_par_champ_ext()

Si le retour attendu de la fonction est un ARRAY le __read_par_champ_ext est
effectu tant qu'il y a des rsultat.

=cut
sub __par_champ_ext($$;$$) {
	my ($dbh, $what, $queries, $options) = @_;
	my $mrqh = __open_par_champ_ext($dbh,$what,$queries,$options);
	if ( ! defined $mrqh ) {
		error "Something went wrong or there is no line.";
		return __return_none();
	}
	if ( ! wantarray ) {
		__debug "Asked for a single line, not much to do.";
		my $res = __read_par_champ_ext($mrqh);
		__close_par_champ_ext($mrqh);
		return $res;
	}
	my @res;
	__debug "Reading first line.";
	my $tmp = __read_par_champ_ext($mrqh);
	while ( defined $tmp ) {
		push @res, $tmp;
		$tmp = __read_par_champ_ext($mrqh);
		__debug "One more line. We are up to ".$#res;
	}
	__debug "No more line to read.";
	__close_par_champ_ext($mrqh);
	return @res;
	
}

sub __par_champ_test( $$$$$ ) {
	my ( $dbh, $what, $val, $whatnom, $testtype ) = @_;
	return __par_champ_ext($dbh, $what, [[$whatnom, $testtype,$val]]);
} 


# Section 7bis: Recherche par clef etrangere
#

sub __par_clef( $$$$;$ ) {
	my ( $dbh, $what, $id, $clef, $order ) = @_;
	return __par_clefs($dbh, $what, { $clef => $id }, $order);
}

sub __par_clefs( $$$;$ ) {
	my ( $dbh, $what, $queries, $order ) = @_;

	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $sql = __sql_from_queries($table, $queries);
	if ( ! defined $sql ) {
		error "Invalid parameters";
		return __return_none();
	}

	if ($order) {
		$order = [ $order ];
	} else {
		$order = 'none';
	}

	my $tmp = get_data_reuse ( $dbh, [ $table ], $sql, $order, 'none' );
	__debug "__par_clefs with query ".$tmp->{'query'};
	return __return_all_or_one($tmp);
}


#sub xx_par_yy_id( $ ) { return __par_clef($dbh,'XX',$_[0],'YY_ID');}

# Section 8: Creation d'entrees dans les tables, avec sequencage
# 

sub __new( $$@ ) {
	my ($dbh, $what, @values) = @_;
	my $table = $what;
	# get_next va ajouter le suffix si besoin.
	my $whatid = get_next($dbh, $table.'_ID');
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	if (!defined($whatid)) {
		return undef;
	}
	unshift @values , $what.'_ID='.$whatid;
	my $tmp = new_data ( $dbh, [ $table ], \@values );
	if ($tmp->{'err_db'} || !$tmp->{'ok'}) {
		error "Echec crea d'un ".$what." par ".$tmp->{'query'};
		info "Erreur: ".$tmp->{'err_mess'} if ($tmp->{'err_mess'});
		return undef;
	} else {
		__debug "Creation d'un ".$what." par ".$tmp->{'query'};
		return $whatid;
	}
}

# Section 9: Changement des champs d'une table

sub __update($$$@) {
	my ( $dbh, $what, $id, @values ) = @_;
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $vals = [];
	for my $v ( @values ) {
		$v =~ s/^$what\./$table./ if defined $dbh->{'suffix'};
		$v = $table.'.'.$v if $v !~ /^$table\./;
		push @$vals, $v;
	}
	my $tmp = set_data($dbh, $vals, [ $table.'.'.$what.'_ID='.$id ] );
	__debug "Updates value(s) on table $what, id $id, with query ".$tmp->{'query'}.".";
	return __return_count($tmp);
}

sub __set_tbl_val( $$$$$ ) {
	my ( $dbh, $what, $key, $whatid, $val ) = @_;
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $tmp = set_data ( $dbh, [ $table.'.'.$key.'='.$val ], [ $table.'.'.$what.'_ID='.$whatid ] );
	__debug "Set value: $table.$key=$val for id $whatid with query ".$tmp->{'query'}.".";
	if ( $tmp->{'ok'} ) {
		return TRUE;
	} else {
		return FALSE;
	}
}

#sub set_xx_yy( $$ ) { __set_tbl_val('XX','XX_YY',$_[0],$_[1]); }
#sub set_xx_yy( $$ ) { __set_tbl_val('XX','XX_YY',$_[0],dochar($_[1])); }
#sub set_xx_yy( $$ ) { __set_tbl_val('XX','XX_YY',$_[0],dobool($_[1])); }

#Pour vi, quand on a les bonnes macros de dfinies (demander  Benj):
#:Fct FWD NOM

# Section 10: Gestion des tables de relation

sub  __update_relation($$$@) {
	my ( $dbh, $what, $clefs, @values ) = @_;
	unless (defined ($dbh) and
		defined ($what) and
		defined ($clefs) and
		@values) {
		error "Invalid parameters";
		return FALSE;
	}

	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $conds = [];
	my $conds_str = [];
	while (my ($key, $val) = each (%{$clefs})) {
		push @$conds, "$table.$key=$val";
		push @$conds_str, "$key=$val";
	}
	$conds_str = join ', ', @$conds_str;
	my $vals = [];
	for my $v ( @values ) {
		$v =~ s/^$what\./$table./ if defined $dbh->{'suffix'};
		$v = $table.'.'.$v if $v !~ /^$table\./;
		push @$vals, $v;
	}

	my $tmp = set_data($dbh, $vals, $conds);
	__debug "Updates value(s) on table $what, conds=($conds_str), with query ".$tmp->{'query'}.".";
	return __return_count($tmp);
}


# ------------------------------------------------------------------------------------
# nom : __set_rel_val
# auteur : Alexis Sukrieh <alexis@sitadelle.com>  Mon Aug 18 10:06:04 CEST 2003
# 
# description : permet de modifier la valeur d'un champ d'une table de relation, la fonction
# 		__set_tbl_val ne permet de modifier le champ d'une table qu'avec une seule
# 		clef. Ici, on se base sur une srie indfinie de couples clefs/valeurs pour la 
# 		clause WHERE de l'update.
#
# parametres : 
# 	0 - $dbh    : le handler vers la base de donnes
# 	1 - $table  : la table  mettre  jour
# 	2 - $clefs  : un HREF contenant TABLE1_ID => valeur, TABLE2_ID => valeur, etc... 
# 	3 - $colone : le nom de la colone  mettre  jour
# 	4 - $valeur : la valeur  utiliser
#
# retours :
# 	0 - si une erreur se produit
# 	1 - si tout va bien
# ------------------------------------------------------------------------------------
sub __set_rel_val ($$$$$) {
	my ($dbh, $table, $clefs, $colone, $valeur) = @_;
	my $conds = [];
	
	unless (defined ($dbh) and
		defined ($table) and
		defined ($clefs) and
		defined ($colone) and
		defined ($valeur)) {
		error "Invalid parameters";
		return FALSE;
	}
	
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	while (my ($key, $val) = each (%{$clefs})) {
		push @$conds, "$table.$key=$val";
	}
	
	my $tmp = set_data ( $dbh, ["$table.$colone=$valeur"], $conds );
	if ( $tmp->{'ok'} ) {
		return TRUE;
	} else {
		error "Echec de la mise a jour pour $table";
		return FALSE;
	}

}

# ----------------------------------------------------------------------
# nom 		: __set_self_relation
# date		: Wed Nov 26 15:31:42 CET 2003
# description 	: permet de crer une relation d'une table vers elle meme
#		  la convention veut qu'une table de relation vers elle-mme
#		  soit nomme : TABLESELF et contienne deux colones :
#		  TABLE_ID et TABLE_DEPID
# parametres 	: 
# 			0 - le handler vers la base de donnes
# 			1 - la table contenant les ID a lier
# 			2 - l'ID pere 
# 			3 - l'ID fils 
# 			4 - une liste de valeurs a ajouter par la meme
# 			occasion
# retours 	:
# 			1 si la requte russie
# 			0 sinon
# ----------------------------------------------------------------------
sub __set_self_relation ($$$$@) 
{
	my ( $dbh, $table, $id_pere, $id_fils, @values ) = @_;

	return __set_relation($dbh, $table, 'SELF', $id_pere, $id_fils, @values);
}


sub __set_relation( $$$$$@) {
	my ( $dbh, $whata, $whatb, $aid, $bid, @values ) = @_;
	map { $_=$whata.$whatb.'_'.$_ } @values;
	my $tmp = new_data ( $dbh, [ $whata.$whatb ],
				[ $whata.'_ID='.$aid,
				  $whatb.'_ID='.$bid,
				  @values ]);
	if ( $tmp->{'ok'} ) {
		return TRUE;
	} else {
		return FALSE;
	}
}

# ----------------------------------------------------------------------
# nom 		: __get_self_relation
# description 	: permet de rcuprer des relations vers elles-meme
# parametres 	: 
# retours 	:
# ----------------------------------------------------------------------
sub __get_self_relation( $$$$@) {
       my ( $dbh, $table, $which, $id, @values ) = @_;
       return __get_relation($dbh, $table, 'SELF', $which, $id, @values);
}


sub __get_relation( $$$$$@) {
	my ( $dbh, $whata, $whatb, $what, $id, @values ) = @_;

	if (! defined $id or ! defined $whata or ! defined $whatb or ! defined $what) {
		error "Missing parameter";
		return __return_none();
	}
	my $tbl = $whata.$whatb;
	map { if ($_ !~ /_[A-Z0-9]+([!=~<>]+|IS NULL).*$/) {$_=$tbl.'_'.$_;} } @values;
	map { if ($_ !~ /^[A-Z]*\./ )  {$_=$tbl.'.'.$_;} } @values;

	my $tmp = {};
	$tmp = get_data ( $dbh, [ $tbl ], [ $tbl.'.'.$what.'_ID='.$id, @values ]);
	__debug "query: ".$tmp->{'query'};
	return __return_all_or_one($tmp);
}

sub __get_self_relation_ordonnee($$$$;$@) {
	my ( $dbh, $table, $which, $id, $ordre, @values ) = @_;
	return __get_relation_ordonnee($dbh,$table,'SELF',$which,$id,$ordre,@values);
}

sub __get_relation_ordonnee($$$$$;$@) {
	my ($dbh, $whata, $whatb, $what, $id, $ordre, @values) = @_;
	my $tbl = $whata.$whatb;
	map { if ($_ !~ /_[A-Z0-9]+$/) {$_=$tbl.'_'.$_;} } @values;
	map { if ($_ !~ /^[A-Z]*\./ )  {$_=$tbl.'.'.$_;} } @values;
	if ( ref($ordre) ne 'ARRAY' ) {
		$ordre = [ $ordre ];
	}
	my $tmp = get_data($dbh, [$tbl], [$tbl.'.'.$what.'_ID='.$id, @values],
                           $ordre, 'none');
	__debug "query: ".$tmp->{'query'};
	return __return_all_or_one($tmp);
}

sub __par_multi_rel ( $$$@ ) {
	my ( $dbh, $conds, $order, @tables ) = @_;
	my @jointures;
	my @tbls;
	for my $c ( @$conds ) {
		next if ( ! defined $c );
		next if ( $c eq '' );
		if ( ref($c) ne 'ARRAY' ) {
			error 'The second parameter should be a *list* of conditions';
			return __return_none();
		}
		push @jointures, __rewrite_condition($c);
	}
	my $lasttbl;
	while ( $#tables >= 0 ) {
		my $cpl  = shift @tables;
		my $tbla = $cpl->[0];
		if ( !defined($tbla) || ! $tbla ) {
			__debug "Missing first table in a couple, I guess it can be the last table of the previous couple.";
			$tbla = $lasttbl;
		}
		if ( !defined($tbla) ) {
			error "Missing first table in the first couple.";
			return __return_none();
		}
		my $tblb = $cpl->[1];
		if ( !defined($tblb) ) {
			error "Missing second table in a couple.";
			return __return_none();
		}
		__debug "Ajout jointure 1: ".$tbla.$tblb.'.'.$tbla.'_ID='.$tbla.'.'.$tbla.'_ID';
		__debug "Ajout jointure 2: ".$tbla.$tblb.'.'.$tblb.'_ID='.$tblb.'.'.$tblb.'_ID';
		push @jointures, $tbla.$tblb.'.'.$tbla.'_ID='.$tbla.'.'.$tbla.'_ID';
		push @jointures, $tbla.$tblb.'.'.$tblb.'_ID='.$tblb.'.'.$tblb.'_ID';
		push @tbls, $tbla;
		push @tbls, $tblb;
		push @tbls, $tbla.$tblb;
		$lasttbl = $tblb;
	}
	if ( ref($order) ne 'ARRAY' ) {
		$order = [ $order ];
	}
	my $tmp = get_data ( $dbh, \@tbls, \@jointures, $order, 'none');
	__debug "Cherche relations multiples par ".$tmp->{'query'};
	return __return_all_or_one($tmp);
}

# ----------------------------------------------------------------------
# nom           : __get_related
# description   : renvoie les entres lies via une table de relation
#                 par exemple, TBL, lie  XX via TBLXX
#                 renvoie les TBL.* lis  un XX_ID ou le contraire.
# parametres    : $dbh, le handler
#                 $whata, la table de gauche dans la relation
#                 $whatb, la table de droite dans la relation
#                 $from, la table dont on donne l'ID
#                 $target, la table dont on veut les entres
#                 $id, l'id pour lequel on veut les entres lies
# 		  @values, liste des conditions supplmentaires
# ----------------------------------------------------------------------
sub __get_related($$$$$$;@) {
        my ( $dbh, $whata, $whatb, $from, $target, $id, @values ) = @_;
        my $tbl = $whata.$whatb;
	map { if ($_ !~ /_[A-Z0-9]+([!=~<>]+|IS NULL).*$/) {$_=$tbl.'_'.$_;} } @values;
	map { if ($_ !~ /^[A-Z]*\./ ) {$_=$tbl.'.'.$_;} } @values;
        my $tmp = get_data($dbh,[$target],["$tbl.${from}_ID=$id","$tbl.${target}_ID=$target.${target}_ID",@values]);
        return __return_all_or_one($tmp);
}

# ----------------------------------------------------------------------
# nom           : __get_related_b
# description   : renvoie les entres lies via une table de relation
#                 on prcise l'identifiant de la table A on reoit les
#                 entres de la table B rattaches.
# ----------------------------------------------------------------------
sub __get_related_b($$$$;@) {
        my ( $dbh, $whata, $whatb, $aid, @values ) = @_;
        return __get_related($dbh,$whata,$whatb,$whata,$whatb,$aid,@values);
}

# ----------------------------------------------------------------------
# nom           : __get_related_a
# description   : renvoie les entres lies via une table de relation
#                 on prcise l'identifiant de la table B on reoit les
#                 entres de la table A rattaches.
# ----------------------------------------------------------------------
sub __get_related_a($$$$;@) {
        my ( $dbh, $whata, $whatb, $bid, @values ) = @_;
        return __get_related($dbh,$whata,$whatb,$whatb,$whata,$bid,@values);
}



# ----------------------------------------------------------------------
# nom 		: __jointure
# description 	: permet d'excuter une requete faisaint une jointure 
#                 entre plusieurs tables
#                 cette fonction prend comme parametre une liste de 
#                 couple de table
#                 pour chacun de ces couples, la permiere table fera
#                 fois pour la jointure. exemple
#                 [ tableA, tableB ] => 
#                           tableA.tableA_id = tableB.tableA_id
# parametres 	: $dbh         : le connecteur pour la db.
# 		  $tables      : referenece sur une liste de couple de 
# 		                 table
# 		  $conditions  : liste de condition SQL, sous forme 
# 		                  d'une ref de ARRAY (cf. __rewrite_conditions())
# 		  $options     : ref de hash contenant => 
#                 		- clef : le nom de l'option a generer 
#                                 limit, orderby, groupby
#                 		- val  : les donnees specifque a l'option
#                                 . limit = ref de hash 
#                                    { offset(opt) => a partir de quel row  
#                                      nb => combien de row
#                                    }
#                                 . orderby et groupby 
#                                   un scalaire ou une liste du type
#                                   (ASC:DESC)?:TABLE.TABLE_CHAMP
#                 $ra_field : liste des champs que le remontes par la 
#                             requete SQL
# retour 	: 0 si erreur sinon ref.
# ----------------------------------------------------------------------
sub __jointure ($$;$$$) {
	my ( $dbh, $couples, $conds, $options, $fields) = @_;

	if ( ref($couples) ne 'ARRAY' ) {
		error "List of couples of tables is invalid.";
		return undef;
	}
	if ( ref($fields) ne 'ARRAY' ) {
		$fields = [];
	}
	
	# on verifie que les conditions sont valides
	my $conds2;
	$conds2 = __rewrite_conditions($conds);
	if ( ! defined $conds2 ) {
		error "Conditions are not valid";
		return __return_none();
	}
	
	# on verifie que les options sont valides
	my $options2;
	$options2 = __rewrite_options($options);
	if ( ! defined $options2 ) {
		error "Failed parsing options";
		return __return_none();
	}

	# On analyse la liste des couples de tables
	my $tables = [];
	foreach my $tbls ( @$couples ) {
		if ( $#$tbls != 1 ) {
			error "An element of the list of couples is not a couple";
			return __return_none();
		}
		my ( $a, $b ) = @$tbls;
		if ( ! defined $a || ! defined $b || ! $a || ! $b ) {
			error "One of the tables in the couple [ $a, $b ] is not valid.";
			return __return_none();
		}
		push @$tables, $a if ( ! grep /^$a$/, @$tables ) ;
		push @$tables, $b if ( ! grep /^$b$/, @$tables ) ;
		
		# on rajoute les conditions de jointures
		push @$conds2, " $a.${a}_ID = $b.${a}_ID ";
	}

	if ( $#$tables < 0 ){
		error "There shoul be at least 2 tables involved...";
		return __return_none();
	}
	__debug "Talbes involved: ".join(',',@$tables);

        my $tmp;
	if ( $#$fields >= 0 ) {
                $tmp = get_data ($dbh, $fields, $conds2, $options2, $tables );
        } else {
                $tmp = get_data ($dbh, $tables, $conds2, $options2 );
        }
	return __return_all_or_one($tmp);
}

# Section 11: Gestion des proprietes hierarchisees

# selectionner toutes les proprits d'un objet et les stocker convenablement dans une table de hash
sub __get_properties( $$$ ) {
	my ( $dbh, $what, $whatid ) = @_;
	my $propwhat='PROP'.$what;

	return undef if (! defined $what or ! defined $whatid);
	my $tmp = get_data($dbh, [$propwhat],
			[ $propwhat.'.'.$what."_ID='".$whatid."'" ] );
	__debug "Recherche props par id par: ".$tmp->{'query'};
	if ( $tmp->{'err_db'} ) {
		warning "Echec";
		return undef;
	}
	
	my $ptmp={};
	foreach my $p (@{$tmp->{'rows'}} ) {
		$ptmp->{$p->{$propwhat.'_ID'}} = $p;
		if ( !defined($p->{$propwhat.'_VAL'}) ) {
			$p->{'val'}={};
		} else {
			$p->{'val'}=$p->{$propwhat.'_VAL'};
		}
	}
	my $props={};
	foreach my $p ( @{$tmp->{'rows'}} ) {
		my ( $par, $nom, $attache, $toadd );
		$par=$p->{$propwhat.'_PAR'};
		$nom=$p->{$propwhat.'_NOM'};
		__debug "Examen de $nom, valant $p->{'val'}, fille de " . (defined $par ? $par : "(undef)") ;
		$attache=$props;
		if ( $par ) {
			$attache=$ptmp->{$par}->{'val'};
		}
		$toadd = $p->{'val'};
		# si la propriete ? un parent
		if ( $nom =~ /^@(.*)$/ ) {
			my $nom = $1;
			# on l'ajoute aux proprietes de meme nom
			if ( !defined($attache->{$nom}) ) {
				$attache->{$nom}=[];
			}
			push ( @{$attache->{$nom}}, $toadd);
		} else {
			$attache->{$nom}=$toadd;
		}
	}

	return $props;
}

sub __del_properties( $$$ ) {
	my ( $dbh, $what, $whatid ) = @_;
	my $propwhat='PROP'.$what;
	my $tmp = del_data($dbh, [$propwhat], [$propwhat.'.'.$what.'_ID='.$whatid]);
	return __return_count($tmp);
}



sub __set_properties( $$$$ ) {
	my ( $dbh, $what, $whatid, $props ) = @_;
	__internal_set_properties ( $dbh, $what, $whatid, $props, undef );
}
sub __internal_set_properties( $$$$$ );
sub __internal_set_properties( $$$$$ ) {
	my ( $dbh, $what, $whatid, $props, $parent ) = @_;
	my $propwhat = 'PROP'.$what;
	my $me;
	my $vals = [ $what.'_ID='.$whatid ];
	push @$vals, $propwhat.'_PAR='.$parent if ( defined($parent) );
	for my $p ( keys %$props ) {
		if ( ref($props->{$p}) eq 'ARRAY' ) {
			# Multi-value
			__debug "$p est multivalue";
			push @$vals, $propwhat.'_NOM="@'.$p.'"';
			for my $v ( @{$props->{$p}} ) {
				if ( ref($v) eq 'HASH' ) {
					__debug "$p est multivalue ET est un hash";
					$me=__new($dbh, $propwhat,@$vals);
					__internal_set_properties($dbh, $what,$whatid,$v,$me);
				} else {
					push @$vals, $propwhat.'_VAL='.dochar($v);
					$me=__new($dbh,$propwhat,@$vals);
					pop @$vals;
				}
			}
		} elsif (ref($props->{$p}) eq 'HASH' ) {
			push @$vals, $propwhat.'_NOM="'.$p.'"';
			__debug "$p est un hash";
			# Noeud d'arbre
			$me=__new($dbh,$propwhat,@$vals);
			__internal_set_properties($dbh,$what,$whatid,$props->{$p},$me);
		} else {
			push @$vals, $propwhat.'_NOM="'.$p.'"';
			__debug "$p est un scalaire";
			# C'est prie d'etre un scalaire, et une feuille
			push @$vals, $propwhat.'_VAL='.dochar($props->{$p});
			$me=__new($dbh,$propwhat,@$vals);
			pop @$vals;
		}
		pop @$vals;
	}
}

sub __update_properties( $$$$ ) {
	my ( $dbh, $what, $whatid, $props ) = @_;
	__del_properties( $dbh, $what, $whatid );
	__set_properties( $dbh, $what, $whatid, $props );
}

# Fonction de del
sub __del ( $$$ ) {
	my ( $dbh, $what, $whatid ) = @_;
	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $tmp = del_data ( $dbh,[$table], [ $table.".".$what."_ID =".dochar($whatid)  ]);
	return __return_count($tmp);
}

sub __del_par_clef ( $$$$ ) {
	my ( $dbh, $what, $id, $clef ) = @_;
	return __del_par_clefs($dbh, $what, { $clef => $id });
}

sub __del_par_clefs ( $$$ ) {
	my ( $dbh, $what, $queries ) = @_;

	my $table = $what;
	$table .= $dbh->{'suffix'} if defined $dbh->{'suffix'};
	my $sql = __sql_from_queries($table, $queries);
	if ( ! defined $sql ) {
		error "Invalid parameters";
		return undef;
	}

	my $tmp = del_data ( $dbh,[$table], $sql);
	return __return_count($tmp);
}

sub __del_par_champ ( $$$$ ) {
	my ( $dbh, $what, $champs, $val ) = @_;
	my $tmp = del_data ( $dbh,[$what], [ $what.".".$what."_".$champs." = ".$val ]);
	return __return_count($tmp);
}

sub __del_relation ( $$$$$@ ) {
	my ( $dbh, $whata, $whatb, $aid, $bid, @values ) = @_;
	if ( $whata eq $whatb ) {
		$whatb = 'SELF';
	}
	my $table = $whata.$whatb;
	map { $_=$table.'_'.$_ } @values;
	my $tmp = del_data ( $dbh, [ $table ],
			[ $table.'.'.$whata."_ID=".$aid,
			  $table.'.'.$whatb."_ID=".$bid,
			  @values ] );
	return __return_count($tmp);
}

# Fonction de max
sub __max ( $$$ ) {
	my ($dbh, $table, $champ) = @_;
	my $tmp = get_data ( $dbh, [ $table.'.'.$champ ], 'none', ['max'] );
	__debug "Search max for ".$champ.": ".$tmp->{'query'};
	return __return_one_col($tmp,'max('.$table.'.'.$champ.')');
}

sub __min ( $$$ ) {
	my ( $dbh, $table, $champ) = @_;
	my $tmp = get_data ( $dbh, [$table.'.'.$champ ], 'none', ['min'] );
	__debug "Search min for ".$champ.": ".$tmp->{'query'};
	return __return_one_col($tmp,'min('.$table.'.'.$champ.')');
}

# Recherche sur une date
# sub __clefs_par_date($dbh, $table, $champ, $unite, $date)
sub __par_date ($$$$$) {
	my ($dbh, $table, $champ, $unite, $date) = @_;
	if ( $date =~ /^\d{14}$/ ) {
		$date = timestamp14_2_db($date);
	}
	$date = dochar($date);
	my $dtcond = [];
	if ($unite eq 'an') {
		push @$dtcond, 'YEAR('.$date.') = YEAR('.$champ.')';
	} elsif ($unite eq 'mois') {
		push @$dtcond, 'YEAR('.$date.') = YEAR('.$champ.')';
		push @$dtcond, 'MONTH('.$date.') = MONTH('.$champ.')';
	} elsif ($unite eq 'jour') {
		push @$dtcond, 'TO_DAYS('.$date.') = TO_DAYS('.$champ.')';
	} elsif ($unite eq 'heure') {
		push @$dtcond, 'TO_DAYS('.$date.') = TO_DAYS('.$champ.')';
		push @$dtcond, 'HOUR('.$date.') = HOUR('.$champ.')';
	} elsif ($unite eq 'minute') {
		push @$dtcond, 'TO_DAYS('.$date.') = TO_DAYS('.$champ.')';
		push @$dtcond, 'HOUR('.$date.') = HOUR('.$champ.')';
	} elsif ($unite eq 'seconde') {
		push @$dtcond, $champ.' = '.$date;
	} else {
		info "__par_date, unite ".$unite." inconnue, pris a la seconde";
		push @$dtcond, $champ.' = '.$date;
	}
	my $tmp = get_data ($dbh, [ $table ], $dtcond );
	__debug "Recherche par date ".join('/',@$dtcond).": ".$tmp->{'query'};
	return __return_all_or_one($tmp);
}

sub __champ_par_id ($$$$) {
	my ($dbh, $table, $champ, $id) = @_;
	my $tmp = get_data($dbh, [$table.'.'.$champ], [$table."_ID = '$id'"] );
	return __return_one_col($tmp,$champ);
}

sub __rotate_table ($$;$) {
	my ( $dbh, $table, $ts ) = @_;
	my $res;

	if ( ! defined $ts ) {
		$ts = epoch_2_timestamp14(time);
	}
	
	$res = clone_table_to($dbh, $table, 'NEW'.$table);
	if (!$res) {
		error 'Rotation de la table echouee';
		return undef;
	}
	$res = rename_tables ( $dbh, [  { 'old' => $table, 'new' => $table.$ts },
					{ 'old' => 'NEW'.$table, 'new' => $table } ] );
	if (!$res) {
		error 'Rotation de la table echouee';
		return undef;
	}
	$res = __new($dbh,'ROTATION',
		'ROTATION_TABLE='.dochar($table),
		'ROTATION_DATE='.$ts,
		'ROTATION_DEST='.dochar($table.$ts));
	return $res;
}

# Useless function kept for backward compatibility.
sub array_has_no_rows ($) {
	my ($array) = shift;
	return ( $#$array == -1 );
}

# section open

# ----------------------------------------------------------------------
# nom 		: __open_all 
# description 	: ouvre une requete de type SELECT *  
# parametres 	: $dbh   : le connecteur pour la db.
# 		  $table : le nom de la table sur le quelle sera 
# 		  effectuee la requete.
# retour 	: une session a passer a __read_all sinon undef.
# ----------------------------------------------------------------------
sub __open_all ($$) {
	 my ($dbh, $table) = @_;
	 return open_sql($dbh, [$table], 'none', 'none');
}

# ----------------------------------------------------------------------
# nom 		: __read_all 
# description 	: retourne ligne par ligne le resultat de __open_all  
# parametres 	: $session : la session ouverte par __open_all
# retour 	: un hash ref contenant le resultat de la requete passe 
#                 a __open_all
#                 undef quand il n'y a plus de ligne
# ----------------------------------------------------------------------
sub __read_all ($) {
	my ($rqh) = @_;
	return read_sql($rqh);
}

sub __close_all($) {
	my ( $rqh ) = @_;
	return close_sql($rqh);
}

=pod

=head1 DESCRIPTION

__open_par_champ_ext, __read_par_champ_ext et __close_par_champ_ext

=head1 UTILISATION

    Il ne faut pas s'carter de la squence classique suivante :

    $session = __open_par_champ_ext(params_de_la_requete_sql);
    return unless $session;
    while( my $row = __read_par_champ_ext($session) ) {
        faire des choses avec $row
    }
    __close_par_champ_ext($session)

* Il faut absoluement faire au moins un __read_par_champ_ext, sinon la
  connexion reste bloque (cette limitation vient de MySQL, pas de notre
  systme interne de verrou).

* Il faut finir par un __close_par_champ_ext pour relacher notre verrou
  interne.

=cut


# ----------------------------------------------------------------------
# nom 		: __open_par_champ_ext
# description 	: clone de la fonction __par_champ_ext mais fonctionnant
#                 avec open_SQL a la place de Get_Data
#                 la partie archivage de __par_champ_ext n'a pas t
#                 garde
# parametres 	: $dbh     : handle de connection  la base de donnes
#                 $what    : la table  attaquer [TABLE.]TABLE[_CHAMP] 
#                            (ce qui est entre crochets est optionnel)
#                 $queries (opt) : une liste contenant des tests 
#                                   effectuer
#                                  ( cf. __rewrite_conditions() )
#                            ARRAYREF [ 
#                                       ARRAYREF [ nom, test, val ]
#                                     ]
#                             nom : le champ contre lequel on effectue 
#                                   le test [[TABLE.]TABLE_]CHAMP
#                            test : le test  effectuer 
#                             val : la valeur  tester (peut valoir
#                                   undef pour certain test)
#                 $options (opt) : un hash contenant les cls 
#                                  optionnelles suivantes :
#                            range : de la forme \d*[-+]\d*
#                                  le premier chiffre optionnel est la 
#                                  valeur de l'offset le caractre 
#                                  sparateur est - pour signifier 
#                                  "de " ou + pour signifier "n 
#                                  enregistrements  partir de l'offset" 
#                                  le deuxieme chiffre est soit la 
#                                  limite haute du range, soit le nombre 
#                                  d'enregistrements  retourner
#                                  Notes :
#                                  un offset sans clause ORDER sur une 
#                                  valeur unique peut donner des 
#                                  rsultats tranges...
#				   ( cf. __rewrite_range_option )
#                            order : scalaire de la forme 
#                                  (up:|down:)?CHAMP (pas ncessaire de 
#                                  mettre TABLE_, mais accept)
#     		                   OU BIEN rfrence sur un tableau de 
#     		                   la mme forme up/down permet 
#     		                   d'ordonner le classement pour ce 
#     		                   champ CHAMP donne le champ sur lequel 
#     		                   on effectue le classement.
#				   ( cf. __rewrite_order_option )
#			     archives : ( cf. __rewrite_archives_option )
# retour 	: une session a passer a __read_par_champ_ext 
#                 sinon undef.
# ----------------------------------------------------------------------
sub __open_par_champ_ext ($$;$$) {
	my ($dbh, $what, $queries, $options) = @_;
	my $opts = {};
	$opts->{'dbh'} = $dbh;
	$opts->{'archives-ok'} = 1;
	my $field = '';
	my $suffix = '';
	$suffix = $dbh->{'suffix'} if defined $dbh->{'suffix'};
	
	if ($what !~ /^(?:([^_. ,]+)\.)?([^_. ,]+)(_[^. ,]+)?/) {
		error "what invalide : $what";
		return __return_none();
	}
	my ( $tbl, $ftbl, $fld ) = ( $1, $2, $3 );
	if ( ! $ftbl ) {
		error "Pas de table prcise dans what !";
		return __return_none();
	}
	$tbl = $ftbl if ( !defined $tbl );
	$opts->{'table'} = $tbl.$suffix;
	$field = '.'.$ftbl.$fld if ( defined $fld );

	my $conds = __rewrite_conditions($queries, $tbl, $dbh->{'suffix'});
	if ( ! defined $conds ) {
		error "Conditions are not valid.";
		return __return_none();
	}
	__debug "Condition rewritten are: ".join('/',@$conds);

	$options = __rewrite_options($options, $opts);
	if ( ! defined $options ) {
		error "Failed option parsing.";
		return __return_none();
	}

	if ( ! defined $opts->{'tables'} ) {
		$opts->{'tables'} = [ $opts->{'table'} ];
	} elsif ( $suffix ne '' ) {
		@{$opts->{'tables'}} = map { $_.$suffix } @{$opts->{'tables'}};
	}

	if ( $#{$opts->{'tables'}} == 0 ) {
		my $mrqh = {
			'opts' => $opts
		};
		my $table = shift @{$opts->{'tables'}};
		__debug "This a query on a single table: $table.";
		my $tmp = open_sql( $dbh, [$table.$field], $conds, $options );
		if (!defined $tmp) {
			error "Failed to open the SQL request";
			return undef;
		}
		$mrqh->{'rqh'} = $tmp;
		return $mrqh;
	} else {
		my $mrqh = {
			'multi' => 1,
			'rqh'   => undef,
			'opts'  => $opts,
			'field' => $field,
			'conds' => $conds,
			'dbh'   => $dbh
		};
		__debug "This is a query with several tables: ".join(',',@{$opts->{'tables'}});
		__open_next_request($mrqh);
		if ( ! defined $opts->{'offset'} ) {
			__debug "No offset, so no line to skip.";
			return $mrqh;
		}
		__debug "We have ".$opts->{'offset'}." line(s) to skip.";
		while ( $opts->{'offset' } > 0 ) {
			my $res = __read_par_champ_ext($mrqh);
			if ( ! defined $res ) {
				return __return_none();
			}
			$opts->{'offset'}--;
			$opts->{'limit'}++;
			__debug "We still have ".$opts->{'offset'}." line(s) to skip.";
		}
		return $mrqh;
	}
}

sub __open_next_request($) {
	my ( $mrqh ) = @_;
	my $opts = $mrqh->{'opts'};
	my $dbh  = $mrqh->{'dbh'};
	my $table = shift @{$opts->{'tables'}};
	my $options;
	push @$options, @{$opts->{'group'}} if (defined $opts->{'group'});
	push @$options, @{$opts->{'order'}} if (defined $opts->{'order'});
	push @$options, @{$opts->{'misc'}} if (defined $opts->{'misc'});
	if ( defined $opts->{'limit'} && defined $opts->{'offset'} && $opts->{'offset'} == 0 ) {
		push @$options, "limit(".$opts->{'limit'}.")";
	}
	if ( defined $mrqh->{'rqh'} ) {
		close_sql($mrqh->{'rqh'});
	}
	__debug "Opening request on the next table: $table.";
	$mrqh->{'rqh'} = open_sql($dbh,
				[ $table.$mrqh->{'field'} ],
				$mrqh->{'conds'}, $options);
	if ( defined $mrqh->{'rqh'} ) {
		return TRUE;
	} else {
		return FALSE;
	}
}

sub __read_par_champ_ext ($) {
	my ($mrqh) = @_;
	my $rqh = $mrqh->{'rqh'};
	my $res = read_sql($rqh);
	if ( ! $mrqh->{'multi'} ) {
		return $res;
	}
	my $opts = $mrqh->{'opts'};
	my $dbh  = $mrqh->{'dbh'};
	my $tables = $opts->{'tables'};
	if ( defined $opts->{'limit'} && $opts->{'limit'} == 0 ) {
		__debug "Limit is 0, so the number of lines must have been achieved.";
		return undef;
	}
	while ( ! defined $res ) {
		__debug "This is the end of the current table.";
		if ( $#$tables == -1 ) {
			__debug "There is no more table to explore.";
			return undef;
		}
		__debug "Getting to the next table if we can.";
		if ( ! __open_next_request($mrqh) ) {
			error "Failed to switch to the next SQL request.";
			return undef;
		}
		# The request handler have been modified.
		$rqh = $mrqh->{'rqh'};
		$res = read_sql($rqh);
	}
	$opts->{'limit'}-- if ( defined $opts->{'limit'} );
	return $res;
}

# ----------------------------------------------------------------------
# nom 		: __close_par_champ_ext 
# description 	: relache le verrou pour permettre un nouvel
#                 __open_par_champ_ext  
# parametres 	: $session : la session ouverte par __open_par_champ_ext
# retour 	: pas de valeur de retour pertinente
# ----------------------------------------------------------------------
sub __close_par_champ_ext ($) {
	my ($mrqh) = @_;
	close_sql($mrqh->{'rqh'});
	delete $mrqh->{'dbh'};
}

1;
