package FDN::Sql::Basic2;

use strict;
use warnings;
use base 'Exporter';

use Carp;
use FDN::Common::Logger;
use FDN::Sql::Basic;

our @EXPORT = qw (
	minimal_functions
	search_functions
	update_functions
	delete_functions
);

our $debug = 0;
sub __debug {
	my ( $msg ) = @_;
	debug $msg if $debug;
}
sub __info {
	my ( $msg ) = @_;
	info $msg if $debug;
}

my $seen = {};
sub __make_func($$$) {
	my ( $pack, $name, $code ) = @_;
	__info "For $name, code is: $code";
	$seen->{$pack} //= {};
	if ( $seen->{$pack}->{$name} ) {
		confess "Redefining $name in $pack";
	}
	$seen->{$pack}->{$name} = 1;

	no strict 'refs';
	if ( defined &{$pack.'::'.$name} ) {
		__debug("Function $name already exists.");
	} else {
		eval '*{'.$pack.'::'.$name.'} = '.$code.';';
		if ( $@ ) {
			confess "Failed to build $name: $@";
		}
	}
	push @{$pack.'::EXPORT'}, $name;
	use strict 'refs';
}

sub __make_new($$$) {
	my ( $pack, $type, $fields ) = @_;

	my $name = 'new_'.lc($type);
	my $code = 'sub { my @vals = (); my $tmp; ';
	for my $f ( @$fields ) {
		$f = $type.'_'.$f if $f !~ /_/;
		$code .= '$tmp = shift;  '.
			 'if ( ! defined $tmp ) { FDN::Common::Logger::error "'.$f.' is required parameter"; return undef; } '.
			 'push @vals, "'.$f.'=".dochar($tmp); ';
	}
	$code .= 'return __new($'.$pack.'::dbh, "'.$type.'", @vals); }';
	__make_func($pack, $name,$code);
}

sub __make_par_id($$) {
	my ( $pack, $t ) = @_;
	my $name = lc($t).'_par_id';
	my $code = 'sub { my $id = shift; return __par_id($'.$pack.'::dbh, "'.$t.'", dochar($id)); }';
	__make_func($pack,$name,$code);
}

sub __make_del($$) {
	my ( $pack, $t ) = @_;
	my $name = 'del_'.lc($t);
	my $code = 'sub { my $id = shift; return __del($'.$pack.'::dbh, "'.$t.'", $id); }';
	__make_func($pack,$name,$code);
}

sub __make_par_clefs($$$) {
	my ( $pack, $type, $fields ) = @_;
	my $name = lc($type).'_par';
	my $func = '__par_clefs';
	$func = '__id_par_clefs' if $type =~ s/_id$//i;
	my $code = 'sub { my $hash = {}; my $tmp; ';
	for my $f ( @$fields ) {
		$name .= '_'.lc($f);
		$f = $type.'_'.$f if $f !~ /_/;
		$code .= '$tmp = shift; ';
		$code .= '$hash->{"'.$f.'"} = $tmp; ';
	}
	$code .= 'return '.$func.'($'.$pack.'::dbh, "'.$type.'", $hash); }';
	__make_func($pack,$name,$code);
}

sub __make_set($$$;$) {
	my ( $pack, $type, $clef, $hook ) = @_;
	my $name = 'set_'.lc($type).'_'.lc($clef);
	if ( $hook ) {
		$hook = 'my $col = "'.$clef.'"; '.$hook.' ';
	} else {
		$hook = '';
	}
	$clef = $type.'_'.$clef if $clef !~ /_/;
	my $code = 'sub { my $id = shift; my $val = shift; '.$hook.
		   'return __set_tbl_val($'.$pack.'::dbh, "'.$type.'", "'.$clef.'", dochar($id), do_null_or_char($val)); }';
	__make_func($pack,$name,$code);
}

sub __make_del_par_clefs($$$) {
	my ( $pack, $type, $fields ) = @_;
	my $name = 'del_'.lc($type).'_par';
	my $code = 'sub { my $hash = {}; my $tmp; ';
	for my $f ( @$fields ) {
		$name .= '_'.lc($f);
		$f = $type.'_'.$f if $f !~ /_/;
		$code .= '$tmp = shift; ';
		$code .= '$hash->{"'.$f.'"} = $tmp; ';
	}
	$code .= 'return __del_par_clefs($'.$pack.'::dbh, "'.$type.'", $hash); }';
	__make_func($pack,$name,$code);
}

sub minimal_functions($) {
	my ( $types ) = @_;

	my ( $pack, $file, $name ) = caller;

	if ( ref($types) ne 'HASH' ) {
		confess "Invalid parameters";
	}

	my $errors;
	for my $t ( keys %$types ) {
		if ( ref($types->{$t}) ne 'ARRAY' ) {
			$errors .= "Invalid parameters for type $t (should be an array-ref)\n";
			next;
		}
		__make_new($pack, $t, $types->{$t});
		__make_par_id($pack, $t);
		__make_del($pack, $t);
	}
	if ( $errors ) {
		confess "Failed to build some of the functions: \n$errors";
	}
	
}

sub search_functions($) {
	my ( $types ) = @_;
	my ( $pack, $file, $name ) = caller;

	if ( ref($types) ne 'ARRAY' ) {
		confess "Invalid parameters";
	}
	if ( @$types % 2 ) {
		confess "Should be an even number of keys...";
	}

	my $errors;
	while ( @$types ) {
		my $type = shift @$types;
		my $clefs = shift @$types;
		if ( ref($clefs) ne 'ARRAY' ) {
			$errors .= "Invalid parameters for type $type (should be an array-ref)\n";
			next;
		}
		__make_par_clefs($pack, $type, $clefs);
	}
	if ( $errors ) {
		confess "Failed to build some of the functions:\n$errors";
	}
}

sub update_functions($;$) {
	my ( $types, $hooks ) = @_;
	my ( $pack, $file, $name ) = caller;
	$hooks = {} if ! defined $hooks;

	if ( ref($types) ne 'ARRAY' ) {
		confess "Invalid parameters";
	}
	if ( @$types % 2 ) {
		confess "Should be an even number of keys...";
	}

	my $errors;
	while ( @$types ) {
		my $type = shift @$types;
		my $clef = shift @$types;
		my $hook;
		$hook = $hooks->{$type} if exists $hooks->{$type};
		if ( ref($clef) ) {
			$errors .= "Invalid parameters for type $type (should be a field name)\n";
			next;
		}
		__make_set($pack, $type, $clef, $hook);
	}
	if ( $errors ) {
		confess "Failed to build some of the functions:\n$errors";
	}
}

sub delete_functions($;$) {
	my ( $types, $hooks ) = @_;
	my ( $pack, $file, $name ) = caller;
	$hooks = {} if ! defined $hooks;

	if ( ref($types) ne 'ARRAY' ) {
		confess "Invalid parameters";
	}
	if ( @$types % 2 ) {
		confess "Should be an even number of keys...";
	}

	my $errors;
	while ( @$types ) {
		my $type = shift @$types;
		my $clef = shift @$types;
		if ( ref($clef) ne 'ARRAY' ) {
			$errors .= "Invalid parameters for type $type (should be an array-ref)\n";
			next;
		}
		__make_del_par_clefs($pack, $type, $clef);
	}
	if ( $errors ) {
		confess "Failed to build some of the functions:\n$errors";
	}
}

1;
