# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Christophe Caillet
# - Alexis Sukrieh
# - Philippe Blayo
# - Thomas Parmelan
# - Jerome Bourgeois
package FDN::Sql::DB_Access ;

use strict;

use Exporter ;
use Encode;

our @ISA = ( 'Exporter' ) ;
our @EXPORT;

push @EXPORT, qw (
	db_trace
	dochar
);

# DB connection
push @EXPORT, qw (
	connect_db
	disconnect_db
	clone_dbh
);

# Direct SQL processing
push @EXPORT, qw (
	sql_transaction
	sql_transaction_reuse_prepare
	sql_transaction_reuse_execute
);

# Unsuported
push @EXPORT, qw (
	Init_DB_Structure
	Get_DB_Structure
);

# Manipulatin data
push @EXPORT, qw (
	get_data
	set_data
	new_data
	del_data

	get_data_reuse

	sequence_data
);

# Locking
push @EXPORT, qw (
	lock_tables
	unlock_tables
	lock_soft
	unlock_soft
	is_locked_soft
);

# Table manipulations
push @EXPORT, qw (
	get_table_create
	clone_table_to
	rename_tables
	drop_table
);

# Cursor like
push @EXPORT, qw (
	open_sql
	read_sql
	close_sql
) ;

use DBI;
use Storable;
use FDN::Common::Logger;
use FDN::Common::Constantes;

our $debug = 0;
sub __debug($) {
	my ( $msg ) = @_;
	return if ( ! $debug );
	debug $msg;
}

sub __get_script_name () {
	return $ENV{'SCRIPT_FILENAME'} if $ENV{'MOD_PERL'};
	return $0;
}

# Contiendra les handlers vers les requtes prpares par Get_Data_Sequence
our @OPEN_DBH;

END {
	for my $h ( @OPEN_DBH ) {
		# If the connexion was opened by someone else, I may have lost it.
		# Not really a problem.
		next if $h->{'pid'} != $$;
		if ( exists $h->{'SQL_REUSE'} ) {
			error "The script forgot to close its databases.";
		}
	}
}

sub init_dbaccess () {
}

# permet d'activer un tracer dans /var/log/dbaccess.log
# en mode exclusif, on efface d'abbord le fichier de trace,
# en mode append on ecrit dedans a la suite, apres avoir ecrit la ligne
# "DBT-> DB_Access Tracer, nouvelle session"
sub db_trace(;$$$) {
	my ( $level, $mode, $file ) = @_;
	$level = 1 if ( ! defined $level );
	$mode = 'append' if ( ! defined $mode );
	$file = "/var/log/dbaccess.log" if ( ! defined $file );

	if ($mode eq 'append') {
		__debug "Tracer appele en mode append";
		if (open TRACEFILE, ">>$file") {
			print TRACEFILE "\n\nDBT-> DB_Access Tracer, nouvelle session.\n";
			close TRACEFILE;
			return DBI->trace($level, $file);
		} else {
			error "Impossible d'ouvrir le fichie de trace en ecriture.";
			return undef;
		}
	} else {
		__debug "Tracer appele en mode exclusif";
		unlink $file;
		return DBI->trace($level, $file);
	}
	# Should not happen...
	return undef;
}


sub __parse_condition($$;$) {
	my ( $condition, $tables, $values ) = @_;
	if ( $condition =~ /^
			((([^\._\s()]+)\.)?(([^()\._\s]+)(_([^()\._\=\~\!\s]+))*))   # Field name (TBL.STUF_OTHER)
				\s*
			(=|~|!=|!~|<=|>=|<|>|IS(?:\s+NOT)?)                    # Operator
				\s*
			("([^"\\]|\\"|\\\\)*"|'([^'\\]|\\'|\\\\)*'|[^'"]\S*)             # Value
			$/x ) {
		my ( $fullf, $tab, $tabf, $f, $cond, $val ) = ( $1, $3, $5, $4, $8, $9 ) ;
		my @allready;
		if ( defined $tables ) {
			if ( defined $tab ) {
				@allready = grep /^$tab(\d{14})?$/, @$tables;
			} else {
				@allready = grep /^$tabf(\d{14})?$/, @$tables;
			}
			if ( $#allready == 0 ) {
				$tabf = $allready[0];
				__debug "Field-table-name from known list: $tabf";
			}
			if ( defined $tab && $tabf =~ /^$tab\d{14}$/ ) {
				__debug "Explicit table name $tab becomes $tabf";
				$fullf =~ s/^$tab/$tabf/;
				$tab = $tabf;
			}
		}
		if ( ! defined $tab ) {
			$tab = $tabf;
			__debug "Implicit table name: $tab from $fullf";
			$fullf = $tab.'.'.$fullf;
		}
		if ( defined $tables && $#allready == -1 ) {
			if ( ! grep /^(\s*\S+\s+(as|AS)\s+)?$tab\s*$/, @$tables ) {
				push @$tables, $tab;
				__debug "Adding table $tab from field $fullf";
				__debug "Table list is now: ".join(", ",@$tables);
			}
		}
		if ( $val =~ /^((\w+)\.)?([^_]+)_(\w+)$/ ) {
			__debug "Value $val is a field name";
			my ( $etab, $itab, $fld ) = ( $2, $3, $4 );
			if ( ! defined $etab ) {
				$etab = $itab;
				$val = $etab.'.'.$val;
			}
			if ( defined $tables && ! grep /^$etab$/, @$tables ) {
				push @$tables, $etab;
				__debug "Adding table $etab from value $val";
			}
		}
		if ( $cond eq '~' ) {
			$cond = 'LIKE';
		}
		if ( $cond eq '!~' ) {
			$cond = 'NOT LIKE';
		}
		if ( defined $values ) {
			if ( $val =~ /^(('|")(.*[^%].*)\2|(\d+))$/ ) {
				my ( $quote, $text, $digit ) = ($2, $3, $4);
				if ( defined $text ) {
					$text =~ s/\\(\\|$quote)/$1/g;
					push @$values, $text;
				}
				if ( defined $digit ) {
					push @$values, $digit;
				}
				$val = '?';
			}
		}
		$condition = $fullf.' '.$cond.' '.$val;
#	} else {
#		print STDERR "Condition BROKEN: <$condition>\n";
	}
	if ( $condition =~ /^[\w\.]+ LIKE ("%+"|'%+')$/ ) {
		$condition = 'none';
	}
	return $condition;
}

sub __parse_conditions($$;$) {
	my ( $conditions, $tables, $values ) = @_;

	return 'none' if ( ! defined $conditions );
	return 'none' if ( ref($conditions) ne 'ARRAY' );
	
	my @ands;
	for my $tmp ( @$conditions ) {
		my @ors;
#		my @conds = split ( /\s+or\s+/i, $cond ) ;
		my $cond = $tmp;
# Dans la regexp, la recherche d'une fonction apres l'operateur
# risque de lire trop loin, en mangeant tout jusqu'a une parenthese
# fermante.
# Ca marche, tant que l'appel de fonction est la dernier element dans
# la serie des 'or' qu'on lit dans cette condition.
		while ( $cond =~ s/^
				(
					(?:[\(\s]*\S+?[\s\)]*) \s*
					(?:\=|\~|\!=|\!~|\<\=|\>\=|\<|\>|\sIS(?:\s+NOT)?\s) \s*
					(?:"(?:[^"\\]|\\"|\\\\)*"|'(?:[^'\\]|\\'|\\\\)*'|\w+\(.*\)|[^'"]\S*)
				) (\s+or\s+)? (.*)$/$3/x ) {
			my $part = $1;
			while ( $part =~ s/^\s*\(\s*(.*)\s*\)\s*$/$1/ ) {}
			my $res = __parse_condition($part,$tables,$values);
			if ( $res ne 'none' ) {
				push @ors, $res;
			}
		}
		if ( $cond !~ /^\s*$/ ) {
			__debug "Remaining condition after split '$cond' added untouched\n";
#			print STDERR "BROKEN Condition <$cond>";
			push @ors, $cond;
		}
		if ( $#ors >= 0 ) {
			my $res = join (') OR (', @ors);
			push @ands, '('.$res.')';
		}
	}
	my $condition;
	if ( $#ands >= 0 ) {
		$condition = join ( ') AND (', @ands ) ;
		$condition	= '('.$condition.')' ;
	} else {
		$condition = 'none';
	}
	return $condition;
}

sub __parse_insert($$) {
	my ( $fields, $table ) = @_;
	my ( @cols, @vals );
	for my $v ( @$fields ) {
		if ( $v =~ /^([^=]+)\s*=\s*(.*)$/s ) {
			my ( $col, $val ) = ( $1 , $2 ) ;
			if ( $col !~ /_/ ) {
				$col = $table.'_'.$col;
			}
			push @cols, $col;
			push @vals, $val;
			__debug 'Insert per column. Col is '.$col.', Val is '.$val;
		}
	}
	my $values;
	if ( $#cols >=0 || $#vals >= 0 ) {
		if ( $#cols != $#vals || $#cols != $#$fields ) {
			error 'Inconsistency in the field list. Expect an SQL error.';
		}
		my $col = join ',',@cols;
		my $val = join ',',@vals;
		$values = ' ( '.$col.' ) VALUES ( '.$val.' )';
	} else {
		$values = ' VALUES ( '.(join ',',@$fields).' )';
	}
	return $values;
}

my $select_options = {
	'distinct' => 0,
	'nocache'  => 0,
	'group'    => 2,
	'orderup'  => 2,
	'orderdn'  => 2,
	'orderdown'=> 2,
	'limit'    => 2,
	'offset'   => 2,
	'count'    => 1,
	'sum'      => 1,
	'max'      => 1,
	'min'      => 1
};

sub __set_opt_arg($$$) {
	my ( $what, $opt, $arg ) = @_;
	if ( defined $arg ) {
		if ( $what =~ /$arg/ ) {
			$what =~ s/$arg/$opt($arg)/;
		} else {
			error "Cannot $opt $arg since not selected.";
		}
	} else {
		$what = $opt.'('.$what.')';
	}
	return $what;
}

our $first_order = 'ORDER BY';
our $first_group = 'GROUP BY';

sub __add_query_opt($$) {
	my ( $opt, $arg ) = @_;
	if ( $opt eq 'limit' ) {
		return " LIMIT $arg";
	}
	if ( $opt eq 'offset' ) {
		return " OFFSET $arg";
	}
	my $res;
	if ( $opt eq 'orderup' ) {
		$res = " $first_order $arg ASC";
		$first_order = ',';
		return $res;
	}
	if ( $opt eq 'orderdn' || $opt eq 'orderdown' ) {
		$res = " $first_order $arg DESC";
		$first_order = ',';
		return $res;
	}
	if ( $opt eq 'group' ) {
		$res = " $first_group $arg";
		$first_group = ',';
		return $res;
	}
	error "Unexpected option $opt($arg), ignored. This is a bug.";
	return "";
}

sub __parse_options($$$$) {
	my ( $options, $what, $table, $cond ) = @_;
	if ( ! defined $options ) {
		$options = [];
	}
	if ( ref($options) ne 'ARRAY' ) {
		$options = [];
	}
	my $query = 'SELECT ';
	my $opts  = '';
	$first_order = 'ORDER BY';
	$first_group = 'GROUP BY';
	for my $opt ( @$options ) {
		next if ( ! defined $opt );
		next if ( $opt eq '' );
		my $arg;
		if ( $opt =~ /^([a-z]+)(\((.*)\))?$/i ) {
			$opt = $1;
			$arg = $3;
		} else {
			error "Option '$opt' is not valid.";
			next;
		}
		my $o2 = lc($opt);
		error "Options should be lowercase." if ( $opt ne $o2 );
		$opt = $o2;
		if ( ! defined $select_options->{$opt} ) {
			error "Option '$opt' is not valid.";
			next;
		}
		if ( defined $arg and ! $select_options->{$opt} ) {
			error "Options '$opt' cannot have an argument. Ignoring it.";
			$arg = undef;
		}
		if ( $opt eq 'distinct' ) {
			$query .= 'DISTINCT ';
		} elsif ( $opt eq 'nocache' ) {
			$query .= 'SQL_NO_CACHE ';
		} elsif ( $select_options->{$opt} == 1 ) {
			$what = __set_opt_arg($what,$opt,$arg);
		} elsif ( $select_options->{$opt} == 2 ) {
			$opts .= __add_query_opt($opt,$arg);
		} else {
			error "Unexpected option. That is a bug.";
		}
	}
	my $res;
	$res = $query.$what.' FROM '.$table;
	if ( $cond ne 'none' ) {
		$res .= ' WHERE '.$cond;
	}
	$res .= $opts;
	return $res;
}

# Desc	: Fonction construisant les requetes SQL correspondantes aux objects demandes
# Proto	: STR Build_SQL_query ( STR $type, ARRAYREF $fields, ARRAYREF $tables,
# 				STR $condition, ARRAYREF $options ) ;
# Parametres	: $type designe le type de la requete i.e. select, insert, update,
# 		  delete, create ou drop
#		: $fields contient la liste des champs ou des valeurs a inserer ou
#		  modifier ou retourner pour une requete
#		: $tables contient la liste des tables impactees par la requete (une seule
#		  sauf pour select)
#		: $condition est la condition d'application pour la requete dans le cas
#		  de requete conditionnelle (select, update, delete)
#		: $options contient les differentes options applicables pour une requete
#		  de type select
sub __build_sql_query($$$$$) {
	my ( $type, $fields, $tables, $conditions, $options) = @_ ;

	my $table;
	if ( $type ne 'select' ) {
		if ( $#$tables > 0 ) {
			error 'Two many tables for query type '.$type.'. '.
			      'Using table '.$tables->[0].'.';
		}
		$table = $tables->[0];
	}

	my $query;
	if ( $type eq 'insert' ) {
		my $values = __parse_insert($fields,$table);
		$query = 'INSERT INTO '.$table.$values;
	} elsif ( $type eq 'update' ) {
		my $values = join ',', @$fields;
		my $condition = __parse_conditions($conditions,undef);
		$query	= 'UPDATE '.$table.' SET '.$values.' WHERE '.$condition ;
	} elsif ( $type eq 'delete' ) {
		my $condition = __parse_conditions($conditions,undef);
		$query	= 'DELETE FROM '.$table.' WHERE '.$condition ;
	} elsif ( $type eq 'select' ) {
		my $cond  = __parse_conditions($conditions,$tables);
		my $table = join ',',@$tables;
		my $what  = join ',',@$fields;
		$query = __parse_options($options,$what,$table,$cond);
#	} elsif ( $type eq 'add_seq' ) {
#		$values = "'".join ( '\',\'', @{$ref_fields} )."'" ;
#		$query	= 'INSERT INTO SEQ VALUES ( '.$values.' )' ;
#	} elsif ( $type eq 'del_seq' ) {
#		$query	= 'DELETE FROM SEQ WHERE '.$condition ;
#	} elsif ( $type eq 'create' ) {
#		my ( $def_field ) ;
#		foreach $field ( @{$ref_fields} ) {
#			if ( $field->{type} =~ /primary/i ) {
#				$def_field = "$field->{name} $field->{type}"
#			}
#		}
	} elsif ( $type eq 'drop' ) {
		$query	= 'DROP TABLE '.$table ;
	}
	return $query ;
}

# La meme, mais renvoie deux infos
# - en return, la requete prevue pour etre reutilisee (avec des ?)
# - dans la table $values les valeurs a mettre a la place des ?
sub __build_sql_reuse_query($$$$$$) {
	my ( $type, $fields, $tables, $conditions, $options, $values) = @_ ;

	if ( $type ne 'select' ) {
		error 'We only reuse select';
		return undef;
	}

	my $query;
	my $cond  = __parse_conditions($conditions,$tables,$values);
	my $table = join ',',@$tables;
	my $what  = join ',',@$fields;
	$query = __parse_options($options,$what,$table,$cond);
	return $query ;
}

# Desc	: Fonction construisant la liste des tables et des champs de la base a
#	  partir de la liste des elements recherches.
# Proto	: VOID __parse_data ( ARRAYREF $elems, ARRAYREF $fields, ARRAYREF $tables ) ;
# Parametres	: $elems est la liste des elements a parser
#		: $fields est la liste resultat contenant la liste des champs trouves
#		: $tables est la liste resultat contenant la liste des tables trouvees
sub __parse_data {
	my ( $elems, $fields, $tables ) = @_ ;
	# On construit la liste a partir de la liste des objets demandes
	for my $elem ( @$elems ) {
		$elem =~ /^(([^\._\s]+)\.)?(([^\._\s]+)(_(\S+))?)$/;
		my ( $table, $tbfield, $field ) = ( $2, $4, $3 );
		if ( $tbfield eq $field ) {
			# recherche toute la table
			$table = $tbfield;
			$field = '*';
		} elsif ( !defined($table) ) {
			# nom de table implicite
			$table = $tbfield;
		} else {
			# nom de table explicite
		}
		my $qfield; # $field with the * protected
		if ( $field eq '*' ) {
			$qfield = '\*';
		} else {
			$qfield = $field;
		}
		if ( ! grep /^$table\.($qfield|\*)$/ , @$fields ) {
			push @$fields, $table.".".$field;
		}
		if ( ! grep /^(\s*\S+\s+(as|AS)\s+)?$table\s*$/, @$tables ) {
			push @$tables, $table;
		}
	}
}

sub __parse_update {
	my ( $modifs, $fields, $tables ) = @_ ;

	for my $modif ( @$modifs ) {
	 	$modif =~ /^([^=]+)\s*=\s*.*$/m ;
		my $table = $1 ;
		if ( $table =~ /^[^\.]+\.[^\_]+_.*$/ ) {
			$table  =~ s/\..*$// ;
		} elsif ( $table =~ /^[^\_]+\_.*$/ ) {
			$table      =~ s/\_.*$// ;
		}
		if ( ! grep ( /$table/, @$tables ) && $#$tables == 0 ) {
			return FALSE;
		} elsif ( $#$tables < 0 ) {
			push @$tables, $table;
		}
		push @$fields, $modif;
	}
	return TRUE;
}

# Desc    	: Se connecte a la base de donnees et renvoie une hash avec
#		  les elements necessaire a la connexion ( handle, param ... )
# Proto   	: HASHREF connect_db ( STR $type, STR $host, STR $name, STR $user,
# 		                       STR $pass, HASHREF $opts, BOOL $force ) ;
# Parametres	: type est le type du moteur de base e.g. mysql ou pg
#		: $host est le nom de host pour contacter le serveur de base de donnees
#		: $name est le nom de la base de donnees
#		: $user est l'utilisateur de la base de donnees
#		: $pass est le mot de passe pour l'utilisateur $db_user
#		: $opts sont les options de la connexion DBI
#		: $force en mod_perl, force la reconnexion au serveur
sub connect_db($$$$$;$$$) {
	my ( $type, $host, $name, $user, $pass, $opts, $force, $prev_dbh ) = @_ ;
	$force //= FALSE;

	my ( $source, $dbi, $utf, $utf_sql ) ;
	my $dbh = $prev_dbh // {};
	
	my $keep_utf;
	if ( $type eq 'mysql' ) {
		$source	= "dbi\:$type\:database=$name;host=$host" ;
		if ( defined $opts && defined $opts->{'utf8'} && $opts->{'utf8'} ) {
			if ( $opts->{'utf8'} =~ /perl/ ) {
				$utf = 1;
			}
			if ( $opts->{'utf8'} =~ /sql/ )  {
				$utf_sql = 1;
			}
			$keep_utf = delete $opts->{'utf8'};
		}
	} elsif ( $type eq 'pg' ) {
		$source	= "dbi\:$type\:dbname=$name;host=$host" ;
	} else {
		error "Unknown databse type $type.";
		return undef;
	}
	if ( ! defined $opts ) {
		$opts	= {
			RaiseError => 0,
			PrintError => 1
		};
	}

	if ( $type eq 'mysql' && $utf_sql ) {
		$source .= ";mysql_enable_utf8=1";
	}
        if (defined $ENV{'MOD_PERL'}) {
		my $sql_timeout = 180;
		Apache::DBI->connect_on_init( $source, $user, $pass, $opts ) ;
		if ( $force ) {
			__debug "on force la reconnexion";
			Apache::DBI->setPingTimeOut($source, 0);
			$dbi = DBI->connect ( $source, $user, $pass, $opts ) ;
			Apache::DBI->setPingTimeOut($source, $sql_timeout);
		} else {
			Apache::DBI->setPingTimeOut($source, $sql_timeout);
			$dbi = DBI->connect ( $source, $user, $pass, $opts ) ;
                }
        } else {
		$dbi = DBI->connect ( $source, $user, $pass, $opts ) ;
	}
	
	if (not defined $dbi) {
		error 'Unable to connect to database '.$source.' with user '.$user;
		error 'Return code '.$DBI::err.' with message '.$DBI::errstr;
		$dbh->{'err_db'}	= 1 ;
		$dbh->{'err_rc'}	= $DBI::err ;
		$dbh->{'err_mess'}	= $DBI::errstr ;
		$dbh->{'last_try'}	= time();
	} else {
		$dbh->{'Handle'}	= $dbi ;
	}

	# Tous ces champs sont remplis mme si le connect()  chou : ils permettront
	# de rejouer plus tard.
	$dbh->{'Engine'}= $type ;
	$dbh->{'Name'}	= $name ;
	$dbh->{'Utf8'}  = $utf;
	$dbh->{'Utf8-sql'} = $utf_sql;
	$dbh->{'Locks'}	= {};
	$dbh->{'Slock'} = '';
	$dbh->{'host'}	= $host;
	$dbh->{'user'}	= $user;
	$dbh->{'pass'}	= $pass;
	if ( $keep_utf ) {
		$opts->{'utf8'} = $keep_utf;
	}
	$dbh->{'opts'}	= $opts;
	$dbh->{'pid'}   = $$;
	$dbh->{'SQL_REUSE'} = {};
	push @OPEN_DBH, $dbh;
	return $dbh ;
}

# Desc	: Gere la deconnexion a une base de donnees en prevoyant un commit
#		  des transactions avant. Renvoie TRUE si tout est ok et FALSE sinon
# Proto	: BOOL Disconnect_db ( HASHREF $ref_handle ) ;
# Parametres	: $ref_handle contient toutes les infos de la connexion a la base
sub disconnect_db($) {
	my ($dbh) = @_ ;
	my ($rc) ;
	
	if ( ! $dbh->{'Handle'}->{'AutoCommit'} ) {
		$rc	= $dbh->{'Handle'}->commit ;
		if ( ! $rc ) {
			error 'Unable to commit transactions before disconnecting';
			error 'Return code '.$DBI::err.' with message '.$DBI::errstr;
			return FALSE;
		}
	}

	# Il y a eu un fork entre temps...
	if ( $dbh->{'pid'} != $$ ) {
		$dbh->{'Handle'}->{'InactiveDestroy'} = 1;
		# InactiveDestroy, donc pas de disconnect.
		# $dbh->{'Handle'}->disconnect;
		return TRUE;
	}

	# on va fermer tous les handle de requetes stockes en memoire.
	my $count = 0;
	foreach my $sql (keys %{$dbh->{'SQL_REUSE'}}) {
		__debug "Fermeture du handle de requete stocke pour $sql";
		$dbh->{'SQL_REUSE'}->{$sql}->finish();
		$count++;
	}
	delete $dbh->{'SQL_REUSE'};
	__debug "$count descripteurs d'instructions SQL actifs liberes.";
	
	$rc = $dbh->{'Handle'}->disconnect ;
	if ( ! $rc ) {
		error 'Unable to disconnect database';
		error 'Return code '.$DBI::err.' with message '.$DBI::errstr;
		return FALSE;
	}
	return TRUE;
}

# nom           : reconnect_db 
# description   : permet de se reconnecter au serveur MySQL
# parametres    : dbh : handle vers une base 
#		  force : force la reconnexion (en mod_perl)
# retours       : le handle mis a jour
my $in_loop = 0;
sub reconnect_db ($$) {
	my ( $dbh, $force ) = @_;

	unless (defined $dbh) {
		alert "utiliser d'abord connect_db avant reconnect_db";
		return undef;
	}

	if ( defined $dbh->{'Handle'} && $dbh->{'pid'} != $$ ) {
		__debug "This DBH was opened for ".$dbh->{'pid'}." now in $$, need to swap.";
		my $old = $dbh->{'Handle'};
		$dbh->{'Handle'} = $old->clone();
		$dbh->{'pid'} = $$;
		$dbh->{'SQL_REUSE'} = {};
		$old->{InactiveDestroy} = 1;
		# InactiveDestroy, donc pas de disconnect.
		# $old->disconnect;
		return $dbh;
	}

       	my $new_dbh = connect_db( $dbh->{'Engine'}, $dbh->{'host'}, $dbh->{'Name'}, $dbh->{'user'}, $dbh->{'pass'}, $dbh->{'opts'}, $force, $dbh );

	if ( $dbh->{'dbname'} ) { # En cas d'utilisation de base multiple.
		if ( $new_dbh->{'Handle'} ) {
			$in_loop++;
			if ( $in_loop > 10 ) {
				use Carp;
				Carp::confess "Failure";
			}
			# Si la connexion a la base principale est reussie, on switch de base.
			# Sinon, le sql_transaction va appeler reconnect_db et ca va mal finir.
			my $res = sql_transaction($new_dbh, 'use', 'use '.$dbh->{'dbname'});
			if ( !$res->{'ok'} ) {
				error "Failed to switch to database '$dbh->{'dbname'}' : ".$res->{'err_mess'};
				return undef;
			}
			$in_loop--;
		}
		$new_dbh->{'dbname'} = $dbh->{'dbname'};
		$new_dbh->{'suffix'} = $dbh->{'suffix'};
	}

	# Si on est en train de faire un "lock table ...", la hash 'Locks' contient le resultat
	# attendu. Il ne faut pas la perdre dans la reconnexion.
	$new_dbh->{'Locks'} = delete $dbh->{'Locks'};

	return $new_dbh;
}


sub clone_dbh($) { # Test
	my ( $dbh ) = @_;

	if ( ! defined $dbh ) {
		alert "Utiliser d'abord connect_db avant clone_dbh";
		return undef;
	}

	my $new = {};
	$new->{'Engine'} = $dbh->{'Engine'};
	$new->{'host'}   = $dbh->{'host'};
	$new->{'Name'}   = $dbh->{'Name'};
	$new->{'user'}   = $dbh->{'user'};
	$new->{'pass'}   = $dbh->{'pass'};
	$new->{'opts'}   = $dbh->{'opts'};
	$new->{'Utf8'}   = $dbh->{'Utf8'};
	$new->{'Utf8-sql'}   = $dbh->{'Utf8-sql'};
	$new->{'Handle'} = $dbh->{'Handle'}->clone();
	$new->{'pid'}    = $dbh->{'pid'};
	# For multi-instance databases
	$new->{'suffix'} = $dbh->{'suffix'};
	$new->{'dbname'} = $dbh->{'dbname'};
	# Should I purge SQL_REUSE ? Lets suppose I should.
	$new->{'SQL_REUSE'} = {};
	if ( $new->{'dbname'} ) {
		sql_transaction($new, 'use', 'use '.$new->{'dbname'});
	}
	return $new;
}


sub __is_utf8($) {
	my $string = shift;
	return 1 if $string =~ /^[\000-\177]*$/; # ASCII
	Encode::_utf8_on($string);
	return 1 if utf8::valid($string);
	return 0;
}

# Desc  : Make sure everything from DB is utf8-clean.
sub hashref_rewrite($) {
	my ( $hash ) = @_;
	return undef if ! defined $hash;
	return $hash if ref($hash) ne 'HASH';
	for my $k ( keys %$hash ) {
		next if ! defined $hash->{$k};
		next if $hash->{$k} =~ /^[\000-\177]*$/;
		my $v = $hash->{$k};
		Encode::_utf8_on($hash->{$k});
		if ( ! utf8::valid($hash->{$k}) ) {
			error "Found information from MySql, $v, in hash-key $k, which is not utf8";
			$hash->{$k} = Encode::decode("iso-8859-15", $v);
		}
	}
	return $hash;
}

sub __sql_transaction_nohandle {
	my ( $dbh, $result ) = @_;
	if ( $dbh->{'last_try'} < time() ) {
		# Ca fait a peu pres une seconde qu'on a echoue, il faut re-essayer, le
		# serveur est peut-etre revenu
		$result->{'reconnect'} = 1;
	} else {
		# Aucune chance de se reconnecter, par la peine d'essayer
		$result->{'reconnect'} = 0;
	}
	$result->{'err_db'} = 1;
	$result->{'err_rc'} = 666; # or whatever ...
	$result->{'err_mess'} = '__sql_transaction called with undefined DBI object; '.
				'this probably means the SQL server is on steroids!';
	alert $result->{'err_mess'};
	return $result;
}

# Desc	: Gestion de la transaction SQL aupres de la base de donnees, renvoie une
# 	  hash contenant les resultat de la transaction
# Proto	: HASHREF __sql_transaction ( HDL $dbh, STR $type, STR $query );
# Parametres	: $dbi est un handle de connexion a une base de donnees
#		: $type est le type de requete envoyee i.e. select ...
#		: $query est la requete SQL proprement dite
sub __sql_transaction {
	my ( $dbh, $type, $query ) = @_ ;

	my $result = {
	    'reconnect'  => 0,
	    'ok'	 => 0,
	    'count_rows' => 0,
	    #'err_db'	 => 0,
	    #'err_rc'	 => 0,
	    'query'	 => $query,
	};

	# Si on est appel ici avec un $dbh->{'Handle'} pas dfini, l'explication la plus
	# probale (en dehors d'un dveloppeur moisi) est qu'on a tent un
	# reconnect_db() qui a chou ... en gnral, cela signifie que le serveur
	# SQL ne tourne carrment pas.
	# Ce qui n'est pas une raison pour gnrer plein de warnings (jamais) et
	# faire n'importe quoi (quoique...).
	if ( ! defined $dbh->{'Handle'} ) {
		return __sql_transaction_nohandle($dbh,$result);
	}
	if ( $dbh->{'Utf8'} && $query =~ /[\x{10000}-\x{10ffff}]/ ) {
		error "Found ultra-wide char in query: $query";
		$query =~ s/[\x{10000}-\x{10ffff}]//g;
	}
	if ( $dbh->{'Utf8'} && !__is_utf8($query) ) {
		error "Found not utf8 char in query: $query";
	}
	if ( ! $dbh->{'Utf8'} && Encode::is_utf8($query) ) {
		error "Query is in utf8 but db is not. Fixing...";
		$query = Encode::encode("iso-8859-1", $query);
	}

	my $prepare = $dbh->{'Handle'}->prepare ( $query ) ;
	if ( ! defined ( $prepare ) ) {
		error 'Unable to prepare query '.$query;
		error 'With return code '.$DBI::err.' and message '.$DBI::errstr;
		$result->{'err_db'}	= 1 ;
		$result->{'err_rc'}	= $DBI::err ;
		$result->{'err_mess'}	= $DBI::errstr ;
		return $result;
	}
	
	# sous mod_perl, si on tente une requete sql entre la fin du timeout du serveur MySQL
	# et la prochaine tentative de reconnexion de Apache::DBI, DBD genere un warning sur
	# 'execute'. On catch le message de warning pour indiquer qu'il faut se reconnecter
	# au serveur
	my $is_warn = 0;
	local $SIG{__WARN__} = sub { 
		error $_[0]; 
		$is_warn = 1;
	};
	my $nb_rows = $prepare->execute ;
	if ( $is_warn ) {
		warning 'There was a warning during query. Might be disconnected.';
		if ( $DBI::err eq '2006' ) {
			$result->{'reconnect'} = 1;
			return $result;
		} else {
			error "Erreur SQL inattendu: $DBI::err -  $DBI::errstr";
		}
	}
	if ( !defined($nb_rows) ) {
		error 'Unable to execute query '.$query;
		error 'With return code '.$DBI::err.' and message '.$DBI::errstr;
		$result->{'err_db'}	= 1 ;
		$result->{'err_rc'}	= $DBI::err ;
		$result->{'err_mess'}	= $DBI::errstr ;
		if ( $DBI::err eq '2006' || $DBI::err eq '2013') {
			error 'Connexion lost, will try to reconnect.';
			$result->{'reconnect'} = 1;
		} else {
			error "Erreur SQL inattendu: $DBI::err -  $DBI::errstr";
		}
		return $result;
	}
	if ( $nb_rows eq '0E0' || $nb_rows == 0 ) {
		if ( $type =~ /^(select|show|desc)$/ ) {
			$result->{'count_rows'} = 0 ;
		}
		$result->{'ok'} = 1;
		return $result;
	}
	$result->{'count_rows'} = $nb_rows ;
	if ( $type =~ /^(select|show|desc)$/ ) {
		my $rows = [];
		while ( $nb_rows > 0 ) {
			my $row = $prepare->fetchrow_hashref ;
			if ( $dbh->{'Utf8'} ) {
				$row = hashref_rewrite($row);
			}
			push @$rows, $row;
			$nb_rows-- ;
		}
		$result->{'rows'} = $rows;
	}
	$result->{'ok'} = 1;
	return $result;
}

# Desc	: en mod_perl, reconnect a la base si la connexion a ete perdue avant
#	  d'appeler __sql_transaction 
# Proto	: HASHREF sql_transaction ( HDL $dbh, STR $type, STR $query );
# Parametres	: $dbh contient toutes les infos de la connexion a la base
#		: $type est le type de requete envoyee i.e. select ...
#		: $query est la requete SQL proprement dite
sub sql_transaction {
	my ( $dbh, $type, $query ) = @_ ;
	
	# ATTENTION :
	#
	# * En contexte mod_perl, on appelle tout le temps Reconnect_db() (qui
	#   potentiellement ne fera pas une vraie reconnexion, grce 
	#   Apache::DBI).
	#
	# * En contexte non-mod_perl, si le Handle n'est pas dfini, on appelle
	#   Reconnect_db() (qui fera une vraie reconnexion). Mais si le Handle est
	#   dfini, on n'a rien  faire pour l'instant.
	#
	if (defined $ENV{'MOD_PERL'} or not defined $dbh->{'Handle'} or $dbh->{'pid'} != $$) {
	    $dbh = reconnect_db($dbh, 0);
	}

	my $result = __sql_transaction( $dbh, $type, $query );

	# Si la transaction a choue et qu'on veut tenter une reconnexion :
	if ($result->{'reconnect'}) {
		$dbh = reconnect_db($dbh, 1);
		$result = __sql_transaction( $dbh, $type, $query );
	}
	return $result;
}

sub dochar($) {
	my ( $txt ) = @_;
	if ( defined $txt ) {
		$txt =~ s/\\/\\\\/g;
		$txt =~ s/"/\\\"/g;
	} else {
		$txt = '';
	}
	return '"'.$txt.'"';
}

# When an option is specified that can influence the list of requested
# elements, that list is vastly modified:
# - if the option is count(FIELD) and FIELD used to be in the requested,
#   fields, it is replaced by count(FIELD)
# - if the field is not in the list, the list is discarded and replaced
#   by count(FIELD).
# Thus, options like min, max, sum, count, are considered more important
# than requested fields.
sub get_data($$$;$$) {
	my ( $dbh, $elems, $condition, $options, $tables ) = @_ ;
	my $fields = [] ;
	if ( ref($tables) ne 'ARRAY' ) {
		$tables = [];
	}
	my $query;
	__parse_data ( $elems, $fields, $tables );
	$query = __build_sql_query ( 'select', $fields, $tables, $condition, $options );
	my $result;
	$result = sql_transaction ( $dbh, 'select', $query ) ;
	return $result ;
}

sub new_data($$$) {
	my ( $dbh, $tables, $data ) = @_ ;
	my $query;
	my $result;
	
	$query	= __build_sql_query ('insert', $data, $tables, 'none', 'none' ) ;
	$result	= sql_transaction   ( $dbh, 'insert', $query ) ;
	return $result ;
}

sub set_data($$$) {
	my ( $dbh, $modifs, $conds ) = @_ ;
	my $query;
	my $result;
	my $fields = [] ;
	my $tables = [] ;
	
	if ( ! __parse_update ( $modifs, $fields, $tables ) ) {
		error 'Unable to build query : bad argument for update';
		$result = {
			'err_db' => 1,
			'err_mess' => 'Bad argument for set_data'
		};
		return $result;
	}
	$query = __build_sql_query ( 'update', $fields, $tables, $conds, 'none' ) ;
	$result = sql_transaction ( $dbh, 'update', $query ) ;
	return $result ;
}

sub del_data($$$) {
	my ( $dbh, $table, $conds ) = @_ ;
	my $query;
	my $result;
	
	$query = __build_sql_query ( 'delete', 'none', $table, $conds, 'none' );
	$result	= sql_transaction ( $dbh, 'delete', $query );
	return $result ;
}

sub __incr_sequence($$$$$) {
	my ( $val, $inc, $min, $max, $cycle ) = @_;
	$val += $inc;
	if ( $inc < 0 && $val < $min ) {
		if ( $cycle ) {
			$val = $max
		} else {
			error "Sequence out of range (lower).";
			return undef;
		}
	} elsif ( $inc > 0 && $val > $max ) {
		if ( $cycle ) {
			$val = $min;
		} else {
			error "Sequence out of range (upper).";
			return undef;
		}
	}
	return $val;
}

sub sequence_data($$;$$) {
	my ( $dbh, $name, $type, $force_val ) = @_;
	$type = 'next' if ( ! defined $type );
	my $res = undef;

	# On verouille la table
	my $tmp = lock_tables($dbh, ['SEQ WRITE']);
	if ( ! defined $tmp || ! $tmp->{'ok'} ) {
		error 'Failed to lock SEQ table';
		return undef;
	}
	
	# On recupere la valeur actuelle
	my $elems = [ 'SEQ' ];
	my $conds = [ "SEQ_NAME='$name'" ];
	my $result = get_data_reuse($dbh, $elems, $conds, '', '' );
	if ( ! defined $result ||
	     ! $result->{'ok'} ||
	     ! $result->{'count_rows'} == 1 ) {
		error 'Failed to get current value for SEQ '.$name;
		goto failure;
	}
	my $val   = $result->{'rows'}->[0]->{'SEQ_VALUE'};
	my $incr  = $result->{'rows'}->[0]->{'SEQ_INC'};
	my $cycle = $result->{'rows'}->[0]->{'SEQ_CYCLE'};
	my $min   = $result->{'rows'}->[0]->{'SEQ_MINVALUE'};
	my $max   = $result->{'rows'}->[0]->{'SEQ_MAXVALUE'};
	__debug "Sequence info: val=$val, inc=$incr, min=$min, max=$max, cycle=$cycle";

	if ( $type eq 'next' ) {
		$val = __incr_sequence($val,$incr,$min,$max,$cycle);
	} elsif ( $type eq 'force' ) {
		$val = $force_val;
	} elsif ( $type eq 'current' ) {
		$res = $val;
		__debug "Current value of $name sequence is $res.";
		goto failure;
	} elsif ( $type eq 'stats' ) {
		$res = {
			'min' => $min,
			'max' => $max,
			'current' => $val,
			'size' => $max-$min,
			'free' => $max-$val,
			'busy' => $val-$min,
			'ratio' => ($val-$min)/($max-$min)
		};
		goto failure;
	} else {
		error "Operation $type not supported on sequences";
		goto failure;
	}
	if ( ! defined $val ) {
		error "Failed to increment sequence $name.";
		goto failure;
	}

	__debug "Condition is ".$conds->[0];
	$result = set_data($dbh,["SEQ_VALUE=$val"],$conds);
	if ( ! $result->{'ok'} ) {
		error 'Failed to update the sequence table.';
		goto failure;
	}
	$res = $val;
	__debug "Next value of $name sequence is $res.";
	__debug "Update with query ".$result->{'query'};
failure:
	unlock_tables( $dbh, ['SEQ WRITE'] );
	return $res;
}

# Pose des lock sur des tables
sub __change_locks($$$) {
	my ( $dbh, $tables, $delta ) = @_;
	my $res = [];
	for my $t ( @$tables ) {
		if ( ! defined $dbh->{'Locks'}->{$t} ) {
			$dbh->{'Locks'}->{$t} = 0;
		}
		$dbh->{'Locks'}->{$t} += $delta;
		if ( $dbh->{'Locks'}->{$t} < 0 ) {
			error "You try to obtain negative lock on $t.";
			$dbh->{'Locks'}->{$t} = 0;
		}
	}
	for my $t ( keys %{$dbh->{'Locks'}} ) {
		if ( $dbh->{'Locks'}->{$t} > 0 ) {
			push @$res, $t;
		}
	}
	return $res;
}

sub lock_tables($$) {
	my ( $dbh, $tables ) = @_ ;
	my ( $query, $result );
	my $tbls;

	$tbls = __change_locks($dbh,$tables,1);
	$query = 'LOCK TABLES ' . join( ',' , @$tbls ) ;
	__debug "Locking with query ".$query;
	
	$result = sql_transaction ( $dbh, 'lock', $query );
	if ( !defined ( $result ) or exists( $result->{'err_db'} ) ) {
		error "impossible de locker : @$tables";
		$result->{'err_mess'}     .= " - DB error : unable to lock tables ";
	} else {
		__debug "Successful lock: $query";
	}
	return $result;
}

sub unlock_tables($;$) {
	my ( $dbh, $tables ) = @_ ;
	my ( $query, $result );
	my $locks = [];
	if ( defined $tables ) {
		$locks = __change_locks($dbh,$tables,-1);
	} else {
		$dbh->{'Locks'} = {};
	}
	if ( $#$locks >= 0 ) {
		$query = 'LOCK TABLES '.join(',',@$locks);
	} else {
		$query = 'UNLOCK TABLES';
	}
	__debug "Unlocking with query $query";
	$result = sql_transaction ( $dbh, 'lock', $query );
	if ( !defined ( $result ) or exists( $result->{'err_db'} ) ) {
		error "Failed unlock: $query";
		$result->{'err_mess'}     .= " - DB error : unable to unlock tables " ;
	} else {
		__debug "Successfull unlock: $query";
	}
	return $result;
}

sub lock_soft($$;$) {
	my ( $dbh, $name, $timeout ) = @_;
	$timeout //= 10;
	$timeout += 0;
	if ( $dbh->{'Slock'} && $dbh->{'Slock'} ne $name ) {
		error "Already having another lock";
		return FALSE;
	}
	my $query = 'select GET_LOCK('.dochar($name).','.$timeout.') as lck';
	my $tmp = sql_transaction ( $dbh, 'select', $query );
	if ( ! $tmp->{'ok'} ) {
		return FALSE;
	}
	if ( @{$tmp->{'rows'}} != 1 ) {
		error "The select on GET_LOCK is supposed to return exactly one row, with the result.";
		return FALSE;
	}
	if ( ! $tmp->{'rows'}->[0]->{'lck'} ) {
		return FALSE;
	}
	$dbh->{'Slock'} = $name;
	return TRUE;
}

sub unlock_soft($$) {
	my ( $dbh, $name ) = @_;
	if ( ! $dbh->{'Slock'} || $dbh->{'Slock'} ne $name ) {
		error "You do not have this lock";
		return FALSE;
	}
	my $query = 'select RELEASE_LOCK('.dochar($name).')';
	my $tmp = sql_transaction ( $dbh, 'select', $query );
	if ( ! $tmp->{'ok'} ) {
		return FALSE;
	}
	$dbh->{'Slock'} = '';
	return TRUE;
}

sub is_locked_soft($;$) {
	my ( $dbh, $name ) = @_;
	if ( ! defined $name ) {
		return $dbh->{'Slock'};
	}
	if ( $dbh->{'Slock'} eq $name && $name ) {
		return TRUE;
	}
	return FALSE;
}

# Prend une requete optimise pour des sequence DBI (avec des ?)
# et la prepare dans DBI.
# Retourne 1 si la preparation reussi, 0 sinon.
sub sql_transaction_reuse_prepare($$) {
	my ($dbh, $query) = @_;

	unless (defined $dbh and defined $query) {
		error "Mauvais passage d'arguments";
		return undef;
	}
	my $rth = $dbh->prepare($query);
	
	unless (defined $rth) {
		error "Probleme lors de la preparation de la requete : $query";
		return undef;
	}

	__debug "requete $query preparee avec succes.";
	return $rth;
}

# Prend un $rth (handler vers 
# la requete preparee) et une liste d'arguments devant correspondre 
# aux arguments attendus par la requete prepare pour $rth.
sub sql_transaction_reuse_execute($$$) {
	my ($dbh, $rth, $values) = @_;
	unless (defined $rth) {
		error "Mauvais passage d'arguments";
		return undef;
	}
	$values = [] unless defined $values;

	# controle de la connexion pour mod_perl
	local $SIG{__WARN__} = sub { 
		error $_[0]; 
	};

	# execution de la requete
	my $rows = $rth->execute(@$values);
	my $result = {};
	
	# si une erreur est survenue lors de l'execution
	if (not defined $rows) {
		error "Probleme lors de l'execution de la requete";
		if ( $DBI::err eq '2006' ) {
			error 'La connexion avec le serveur a ete perdue';
			$result->{'reconnect'} = 1;
		} else {
			error "Erreur SQL inattendu: $DBI::err -  $DBI::errstr";
		}
		return $result;
	} 
	
	# pas d'erreur mais aucune ligne retournee
	if ($rows eq '0E0' || $rows == 0) {
		$result->{'ok'} = 1;
		return $result;
	} 

	# requete reussie et des lignes sont remontees.
	$result->{'count_rows'} = $rows ;
	while ( $rows > 0 ) {
		my $row = $rth->fetchrow_hashref;
		if ( $dbh->{'Utf8'} ) {
			$row = hashref_rewrite($row);
		}
		push ( @{$result->{'rows'}}, $row);
		$rows-- ;
	}
	$result->{'ok'} = 1 ;
	return $result;
}

# prend un handle de dbh et un rth et va executer la requete
# preparee en s'assurant que la connexion est active.
sub sql_transaction_reuse($$$) {
	my ($dbh, $rth, $values ) = @_ ;
	unless (defined $dbh and defined $rth) {
		error "Mauvais passage d'arguments";
		return undef;
	}
	$values = [] unless defined $values;
	
	if (defined($ENV{'MOD_PERL'}) or $dbh->{'pid'} != $$) {
		# FIXME Il est trop tard, le $rth est cram.
		$dbh = reconnect_db($dbh, 0);
	}

	my $result = sql_transaction_reuse_execute($dbh, $rth, $values);

	if ( defined($result->{'reconnect'}) && $result->{'reconnect'} ) {
		# FIXME Il est trop tard, le $rth est cram.
		$dbh = reconnect_db($dbh, 1);
		$result = sql_transaction_reuse_execute($dbh, $rth, $values);
	} 
	return $result;
}

# Un get_data un peu plus intellignet, et ouais, plus optimise ;)
sub get_data_reuse($$$;$$) {
	my ( $dbh, $elems, $conds, $options, $tables ) = @_ ;
	my $fields = [] ;
	if ( ref($tables) ne 'ARRAY' ) {
		$tables = [];
	}

	__parse_data ( $elems, $fields, $tables ) ;
	my $values = [];
	my $query = __build_sql_reuse_query ('select', $fields, $tables, $conds, $options, $values);
	my $nbtry = 1;

REUSE:
	# patch pour permettre l'optimisation "getadata_sequence"
	my $rth = undef;
	if ( $dbh->{'pid'} != $$ ) {
		$dbh = reconnect_db($dbh,0);
	}
	if ( $dbh->{'Utf8'} && $query =~ /[\x{10000}-\x{10ffff}]/ ) {
		error "Found ultra-wide char in query: $query (really weird, you did try to use that in a table name???)";
		$query =~ s/[\x{10000}-\x{10ffff}]//g;
	}
	if ( $dbh->{'Utf8'} && !__is_utf8($query) ) {
		error "Found not utf8 char in query: $query (really weird, you did try to use that in a table name???)";
	}
	for my $v ( @$values ) {
		next if !$dbh->{'Utf8'};
		if ( $v =~ /[\x{10000}-\x{10ffff}]/ ) {
			error "Found ultra-wide char in value: $v";
			$v =~ s/[\x{10000}-\x{10ffff}]//g;
		}
		if ( !__is_utf8($v) ) {
			error "Found not utf8 char in value: $v";
		}
	}
	if (not defined $dbh->{'SQL_REUSE'}->{$query}) {
		$rth = sql_transaction_reuse_prepare($dbh->{'Handle'}, $query);
		if ( ! defined $rth) {
			error "Failed to prepare $query";
			return { 'reconnect' => 0,
				'ok'=> 0,
				'count_rows' => 0,
	 			'err_db'=> 1,
			};
		}
		__debug "[SQL prepared] $query ($rth)";
		$dbh->{'SQL_REUSE'}->{$query} = $rth;
	} else {
		$rth = $dbh->{'SQL_REUSE'}->{$query};
		__debug "[SQL cached] $query (".$rth.")"
	}
	__debug "[SQL values] ".join(',',@$values);

	#__debug "On va executer la requete preparee en $sth, avec @args";
	my $res = sql_transaction_reuse($dbh,$rth,$values);
	if ( $nbtry == 1 && defined($res->{'reconnect'}) && $res->{'reconnect'} ) {
		$nbtry++;
		$dbh = reconnect_db($dbh,1);
		goto REUSE;
	}
	$res->{'query'} = $query;
	return $res;
}



# En entree:
# - une connexion a une base de donnees
# - un nom de table
# En sortie:
# - undef si quelque chose a echoue
# - la description en SQL de la table dans une chaine de caractere, sans la partie 'CREATE TABLE XX'
sub get_table_create {
	my ( $dbh, $table ) = @_;
	my $result;
	my $create_row;
	my $create;
	$result = sql_transaction ( $dbh, 'show', 'show create table '.$table);
	if ( $result->{'count_rows'} != 1 ) {
		error 'Failed to get '.$table.' create sentence';
		return undef;
	}
	$create_row = $result->{'rows'}[0];
	if ( !defined($create_row) || !defined($create_row->{'Create Table'}) ) {
		error 'Failed to get '.$table.' create sentence in fetched row';
		return undef;
	}
	$create = $create_row->{'Create Table'};
	if ( $create =~ s/^CREATE TABLE `$table`// ) {
		__debug 'Table '.$table.' created with '.$create;
	} else {
		error 'Failed to analyze '.$table.' create sentence '.$create.' in fetched row';
		return undef;
	}
	return $create;
}

# En entree:
# - une connexion a une base de donnees
# - un nom de table A
# - un nom de table B
# En sortie:
# - FALSE en cas d'echec
# - TRUE en cas de succes
# Fonction: Cree la table B ayant la meme structure que la table A (meme champs, indexes de la meme
# facon, etc).
sub clone_table_to($$$) {
	my ( $dbh, $table, $clone ) = @_;
	my $create = get_table_create ( $dbh, $table );
	my $result;
	if ( ! defined($create) ) {
		error 'Failed to get table definition for '.$table;
		return FALSE;
	}
	$result = sql_transaction ( $dbh, 'create', 'create table '.$clone.$create);
	if ( defined $result->{'err_db'} ) {
		error 'Failed to create '.$clone.' with definition '.$create;
		return FALSE;
	} else {
		info 'Cloned '.$table.' to '.$clone;
	}
	return TRUE;
}

sub rename_tables($$) {
	my ( $dbh, $actions ) = @_;
	my $query;
	my $result;
	if ( ref($actions) ne 'ARRAY' ) {
		error 'Param 2 (action) have to be an array-ref';
		return FALSE;
	}
	for my $action ( @$actions ) {
		if ( ref($action) ne 'HASH' ) {
			 error 'Param 2 (action) have to be an array-ref of hash-ref';
			 return FALSE;
		}
		if ( !defined($action->{'old'}) || !defined($action->{'new'}) ) {
			 error 'Param 2 (action) have to be an array-ref with keys new and old';
			 return FALSE;
		}
		if (defined($query)) {
			$query .= ' , ';
		}
		$query .= $action->{'old'}.' TO '.$action->{'new'};
	}
	$result = sql_transaction ( $dbh, 'rename', 'rename table '.$query );
	if ( defined $result->{'err_db'} ) {
		error 'Rename failed using '.$query;
		return FALSE;
	} else {
		info 'Renamed tables using '.$query;
	}
	return TRUE;
}

sub drop_table($$) {
	my ( $dbh, $table ) = @_;
	my $query = __build_sql_query('drop',undef,[$table],undef,undef);
	return sql_transaction($dbh,'drop',$query);
}

###############
#  Iterateur  #
###############

# Gestion des erreurs
sub log_erreur ($$$) {
    my ($session, $operation,  $query) = @_;
    error "Unable to $operation query $query :With return code $DBI::err and message $DBI::errstr";
}

################################################################################
# Fonctions open read et close
#
# Ces 3 fonctions utilisent une structure appele session (parfois qualifie de
# "magic"). Cette structure est renvoye par open et utilise par read et close.
#
# Elle contient :
#    - le rsultat du prepare() de DBI
#    - le dbh lui-mme auquel un champ Verrou est ajout
################################################################################

# Prepare_SQL_query
#
# Equivalent de __sql_transaction utilis par open_SQL.
#
# Argument :
#     $dbh : le manipulateur de la connexion
#     $query_type = type de requte
sub __sql_transaction_open ($$$) {
	my ( $dbh, $query_type, $query ) = @_ ;
	my $session = { query => $query } ; # resultat  renvoyer

        # L'option mysql_use_result n'est pertinente que pour MySQL
	if ($dbh->{Engine} ne 'mysql') {
		warning("$dbh->{Engine} pas encore implment");
		return undef;
	}
	if ( ! defined $dbh->{'Handle'} ) {
		return undef;
	}
	if ( $dbh->{'Utf8'} && $query =~ /[\x{10000}-\x{10ffff}]/ ) {
		error "Found ultra-wide char in query: $query";
		$query =~ s/[\x{10000}-\x{10ffff}]//g;
	}
	if ( $dbh->{'Utf8'} && !__is_utf8($query) ) {
		error "Found not utf8 char in query: $query";
	}
	# mysql_use_result permet de ramener les row une par une
	my $prepare = $dbh->{Handle}->prepare ( $query, { "mysql_use_result" => 1} ) ;
	if ( ! defined( $prepare ) ) {
		log_erreur $session, 'prepare',  $query;
		return undef ;
	}
	$session->{'prepare'} = $prepare;

	my $nb_rows = $prepare->execute ;
	use Data::Dumper;
	__debug "Returned ".Dumper($nb_rows);
	if ( ! defined( $nb_rows ) ) {
		log_erreur $session, 'execute',  $query;
		return undef ;
	}

	$session->{'ok'} = 1;
	$dbh->{Verrou} = 1;
	$session->{dbh} = $dbh;

	$session->{'count_rows'} = $nb_rows ;
	$session->{'count_rows'} = 0 if $nb_rows eq '0E0';
	if ( $query_type eq 'select' ) {
		$session->{'count_rows'} = 0;
	}
	
	return $session ;
}

# retourne une ligne de la requete (undef s'il n'en reste plus)
# Attention, pour mod-perl, l'objet ne sera pas de type 'DBI::st',
# mais plutot 'Apache::DBI::st', donc faire !~ /DBI::st$/
sub read_sql ($) {
	my ( $session ) = @_;
	if( ref($session->{'prepare'}) !~ /DBI::st$/ ) {
		error "mauvais passage de parametre";
		return undef; 
	}
	local $SIG{__WARN__} = sub { 
		error $_[0]; 
	};
	my $row = $session->{'prepare'}->fetchrow_hashref;
	if ( $session->{'dbh'}->{'Utf8'} ) {
		$row = hashref_rewrite($row);
	}
	return $row;
}

# relache le verrou
sub close_sql ($) {
	my ( $session ) = @_;
    
	unless (defined($session->{dbh})) {
		warning "impossible de relacher le verrou d'une session sans dbh";
		return;
	}
	#---------------------------------------------------------------------------
	#   * Fermeture du curser *
	#
	#   Il faut rcuprer les lignes pas encore lues pour que mysql ferme son
	# curseur.
	#
	#   * Commentaires importants *
	# Il serait prfrable de pouvoir fermer le curseur directement, sans avoir
	#  rcuprer les lignes restantes.
	# !!!! Attention !!!!
	# Ne pas utiliser un fetchall_arrayref qui rcuprerait toutes les lignes
	# restantes et occuperait donc beaucoup de mmoire ce qui est contraire au
	# but des fonctions open, read et close.
	#--------------------------------------------------------------------------- 
	# Eviter les problemes...
	local $SIG{__WARN__} = sub { 
		error $_[0]; 
	};
	while($session->{'prepare'}->fetchrow_arrayref) {}
	$session->{dbh}{Verrou} = 0;
	delete $session->{dbh};
}

# parametres : identiques a get_data
# retourne la session (le magic) [$session dans __prepare_sql_query]
#
# Pour le moment, une seule connexion est possible  la fois. C'est le sens
# du verrou pos par open_sql et relach par close_sql.
sub open_sql ($$$$;$) {
	my ( $dbh, $elems, $conds, $options, $tables ) = @_ ;
	my $fields	= [] ;

	if ( exists $dbh->{Verrou} and $dbh->{Verrou} ) {
		error "Another request is already opened";
		return undef;
	}
	if ( $dbh->{'pid'} != $$ ) {
		__debug "I am opening an SQL request on a DBH for pid ".$dbh->{'pid'}." while I am $$";
		$dbh = reconnect_db($dbh,0);
	}
	if ( ! defined $dbh->{'Handle'} && $dbh->{'last_try'} < time() ) {
		__debug "I have a broken DBH, but did not try connecting again recently, will try now";
		$dbh = reconnect_db($dbh,0);
	}
	if ( ! defined $dbh->{'Handle'} ) {
		__debug "My DBH is broken, and cannot be fixed now (sql server does not answer). Will fail.";
		error "No sql server available";
		return undef;
	}
	if ( ref($tables) ne 'ARRAY' ) {
		$tables = [];
	}
	__parse_data ( $elems, $fields, $tables ) ;
	my $query = __build_sql_query ( 'select', $fields, $tables, $conds, $options ) ;
	__debug "Opening SQL request for query $query";
	my $res = __sql_transaction_open ( $dbh, 'select', $query ) ;
	if ( ! defined($res) ) {
		if ( $DBI::err eq '2006' || $DBI::err eq '2013' ) {
			reconnect_db($dbh,1);
			$res = __sql_transaction_open ( $dbh, 'select', $query ) ;
		} elsif ( ! defined $dbh->{'Handle'} && $dbh->{'last_try'} < time() ) {
			# That should never happen. Either I found it before atempting the first
			# SQL transaction, and it have been fixed, either I have no chance to fix
			# it now...
			reconnect_db($dbh,1);
			return undef if ! defined $dbh->{'Handle'};
			$res = __sql_transaction_open ( $dbh, 'select', $query ) ;
		} else {
			error "Erreur SQL inattendu: $DBI::err -  $DBI::errstr";
		}
	}
	return $res;
}

1;
