package FDN::Sql::MultiDBC;

use strict;
use warnings;
use base 'Exporter';

use FDN::Common::Logger;
use FDN::Sql::Basic;
use FDN::Sql::DB_Access;
use FDN::Sql::DBC qw(do_disconnect);
use FDN::Sql::Structure qw(read_config read_files structure create switch_option set_quiet set_output purge);


our @EXPORT = qw(
	do_connect
	do_disconnect
	reset_multi_dbh
	switch_multi_dbh
	purge_multi_dbh

	multi_db_has_module
	multi_db_add_module
	multi_db_del_module
);

my $dbh;
my $DBS;
my $sub_dbh;

sub __tous_db() {
	return __tous($dbh, 'DB');
}

sub __db_par_name($) {
	my ( $name ) = @_;
	return __par_clef($dbh, 'DB', $name, 'DB_NAME');
}

sub __incr_db_cur($) {
	my ( $dbid ) = @_;
	return __set_tbl_val($dbh, 'DB', 'DB_CUR', dochar($dbid), 'DB_CUR+1');
}

sub __decr_db_cur($) {
	my ( $dbid ) = @_;
	return __set_tbl_val($dbh, 'DB', 'DB_CUR', dochar($dbid), 'DB_CUR-1');
}

sub __new_base($$) {
	my ( $suffix, $dbname ) = @_;
	return __new($dbh, 'BASE',
		'BASE_SUFFIX='.dochar($suffix),
		'BASE_DBNAME='.dochar($dbname)
	);
}

sub __del_base($) {
	my ( $id ) = @_;
	return __del($dbh, 'BASE', $id);
}

sub __base_par_suffix($) {
	my ( $suffix ) = @_;
	my $base = __par_clef($dbh, 'BASE', $suffix, 'BASE_SUFFIX');
	if ( ! defined $base ) {
		return undef;
	}
	if ( ! exists $base->{'BASE_MODULES'} ) {
		die "The databse is not up to date. Please check. There should be a BASE_MODULES column in table BASE.";
	}
	if ( ! defined $base->{'BASE_MODULES'} || $base->{'BASE_MODULES'} eq '' ) {
		# Si jamais la liste des modules est vide, on positionne a 'common'
		# qui est la bonne valeur.
		__set_base_modules($base->{'BASE_ID'},{});
		$base->{'BASE_MODULES'} = 'common';
	}
	if ( $base->{'BASE_MODULES'} eq 'common' ) {
		$base->{'modules'} = {};
		return $base;
	}

	my $list = [ split /,/, $base->{'BASE_MODULES'} ];
	my $modules = {};
	for my $name ( @$list ) {
		$modules->{$name} = 1;
	}
	$base->{'modules'} = $modules;
	return $base;
}

sub __set_base_version($$) {
	my ( $bid, $version ) = @_;
	return __set_tbl_val($dbh, 'BASE', 'BASE_VERSION', dochar($bid), dochar($version));
}

sub __set_base_modules($$) {
	my ( $bid, $modules ) = @_;

	my $mods_name = '';
	for my $name ( @{$DBS->{'modules_list'}} ) {
		if ( exists $modules->{$name} && $modules->{$name} ) {
			$mods_name .= ','.$name;
		}
	}
	$mods_name =~ s/^,//;
	if ( ! $mods_name ) {
		$mods_name = 'common';
	}
	return __set_tbl_val($dbh, 'BASE', 'BASE_MODULES', dochar($bid), dochar($mods_name));
}

sub __dispatch($) {
	my ( $suffix ) = @_;
	# Pas besoin de verrouiller pendant le dispatch:
	# - le suffix est deja attribue
	# - l'unicite de l'index sur BASE_SUFFIX garanti que l'ecriture ne peut pas etre en
	#   double dans la table
	my $dbs = [ __tous_db() ];
	if ( ! @$dbs ) {
		die "La table DB de la base principale est vide, je ne peux pas dispatcher un nouveau client. Echec.";
	}
	my $limits = [];
	my $total = 0;
	for my $db ( @$dbs ) {
		$total += $db->{'DB_MAX'};
		$total -= $db->{'DB_CUR'};
		push @$limits, $total;
	}

	my $choice = rand($total);
	while ( $limits->[0] < $choice && $limits->[1] ) {
		shift @$limits;
		shift @$dbs;
	}
	my $dbname = $dbs->[0]->{'DB_NAME'};
	my $bid = __new_base($suffix, $dbname);
	if ( ! $bid ) {
		# Le seul cas d'echec (hors MySql aux fraises) c'est que la ligne a ete
		# ajoutee par un voisin.
		error "Failed to create a line in BASE";
		return __base_par_suffix($suffix);
	}
	__incr_db_cur($dbs->[0]->{'DB_ID'});
	return __base_par_suffix($suffix);
}

sub __switch_db($$) {
	my ( $dbh, $dbname ) = @_;
	$DBS->{'name'} = $dbname;
	create($DBS);
	my $res = sql_transaction($dbh, 'use', 'use '.$dbname);
	if ( !$res->{'ok'} ) {
		error "Failed to switch to database '$dbname' : ".$res->{'err_mess'};
		return undef;
	}
	$dbh->{'dbname'} = $dbname;
}

BEGIN {
	switch_option('dry',0,'all');
	set_quiet('quiet',0);
	set_output('syslog');
}

sub do_connect($) {
	my ( $name ) = @_;
	$dbh = FDN::Sql::DBC::do_connect($name);
	$DBS = read_config("check/$name.cfg","multi");
	read_files($DBS);
	return $dbh;
}

sub reset_multi_dbh {
	return switch_multi_dbh('NONE');
}


sub __valid_suffix($) {
	my ( $suffix ) = @_;
	if ( ! defined $suffix ) {
		return 0;
	}
	if ( $suffix eq 'NONE' ) {
		return 1;
	}
	if ( $suffix eq '' || $suffix =~ /^[a-z]/i ) {
		return 0;
	}
	return 1;
}

sub __switch_multi_dbh($) {
	my ( $suffix ) = @_;
	if ( ! defined $dbh ) {
		die "Il faut se connecter, quand meme, un peu.";
	}
	if ( ! __valid_suffix($suffix) ) {
		$suffix //= '<undef>';
		die "Switcher sur le suffix '$suffix' en MultiDB est une mauvaise idee (lettres, trop court, vide, etc).";
	}
	if ( ! defined $sub_dbh ) {
		$sub_dbh = clone_dbh($dbh);
		$sub_dbh->{'suffix'} //= 'NONE';
		$sub_dbh->{'dbname'} //= 'NONE';
	}
}

sub multi_db_has_module($$) {
	my ( $suffix, $module ) = @_;
	my $base = __base_par_suffix($suffix);
	if ( ! $base ) {
		# Cette base n'existe pas.
		return 0;
	}
	if ( exists $base->{'modules'} && $module eq 'common' ) {
		return 1;
	}
	if ( exists $base->{'modules'}->{$module} && $base->{'modules'}->{$module} ) {
		return 1;
	}
	return 0;
}

sub multi_db_add_module($$) {
	my ( $suffix, $module ) = @_;
	my $base = __base_par_suffix($suffix);
	if ( ! $base ) {
		die "Multi-db $suffix not found";
	}
	if ( $base->{'modules'}->{$module} ) {
		# Deja active
		return;
	}

	$base->{'modules'}->{$module} = 1;
	__set_base_modules($base->{'BASE_ID'}, $base->{'modules'});
}

sub multi_db_del_module($$) {
	my ( $suffix, $module ) = @_;
	my $base = __base_par_suffix($suffix);
	if ( ! $base ) {
		die "Multi-db $suffix not found";
	}
	if ( ! $base->{'modules'}->{$module} ) {
		# Le module n'est pas active, rien a retirer.
		return;
	}
	delete $base->{'modules'}->{$module};
	__set_base_modules($base->{'BASE_ID'}, $base->{'modules'});
}

sub switch_multi_dbh($) {
	my ( $suffix ) = @_;
	__switch_multi_dbh($suffix);
	if ( $sub_dbh->{'suffix'} eq $suffix ) {
		return $sub_dbh;
	}
	if ( $suffix eq 'NONE' ) {
		$sub_dbh->{'suffix'} = 'NONE';
		return undef;
	}
	if ( $suffix =~ /^[A-Z]/ ) {
		die "Le suffix '$suffix' commence par une lettre capitale, il va se cogner dans le nom de la table. On arrete ici.";
	}
	my $base = __base_par_suffix($suffix);
	if ( ! defined $base ) {
		$base = __dispatch($suffix) or return undef;
	}
	if ( $base->{'BASE_DBNAME'} ne $sub_dbh->{'dbname'} ) {
		__switch_db($sub_dbh, $base->{'BASE_DBNAME'}) or return undef;
	}
	$sub_dbh->{'suffix'} = $suffix;
	my $structure;
	if ( $base->{'BASE_MODULES'} eq 'common' ) {
		$structure = $DBS;
	} else {
		# Si la clef n'existe pas, on a un probleme...
		my $mods_name = $base->{'BASE_MODULES'};
		if ( ! exists $DBS->{'modules'}->{$mods_name} ) {
			Carp::confess "La base '$suffix' utilise les modules '$mods_name', qui ne sont PLUS dans le fichier check.cfg associe.";
		}
		$structure = $DBS->{'modules'}->{$mods_name};
		$structure->{'name'} = $DBS->{'name'};
	}
	if ( $base->{'BASE_VERSION'} ne $structure->{'signature'} ) {
		if ( is_locked_soft($dbh) ) {
			my $lock_name = is_locked_soft($dbh);
			die "I need to change the structure for base '$suffix'. But the dbh is already soft-locked on '$lock_name'. Cannot get an aditionnal lock.";
		}
		if ( ! lock_soft($dbh,'structure-'.$suffix) ) {
			# Echec a obtenir le verrou.
			die "Failed to lock for structure upgrade on base '$suffix'. Probably timeout.";
		}
		$base = __base_par_suffix($suffix);
		if ( $base->{'BASE_VERSION'} ne $structure->{'signature'} ) {
			# En cas d'echec, structure va lever une exception (die).
			structure($structure, clone_dbh($sub_dbh));
			__set_base_version($base->{'BASE_ID'}, $structure->{'signature'});
		}
		unlock_soft($dbh,'structure-'.$suffix);
	}
	return $sub_dbh;
}

sub purge_multi_dbh($) {
	my ( $suffix ) = @_;
	__switch_multi_dbh($suffix);
	my $base = __base_par_suffix($suffix);
	if ( ! defined $base ) {
		# La base '$suffix' n'existe pas, rien a purger.
		debug "La base '$suffix' n'existe pas.";
		return 1;
	}
	my $db_name = $base->{'BASE_DBNAME'};
	if ( $db_name ne $sub_dbh->{'dbname'} ) {
		__switch_db($sub_dbh, $db_name) or return 0;
	}
	my $db = __db_par_name($db_name);
	if ( ! defined $db ) {
		error "Alors... L'instance '$suffix' en multi-db est rangee dans la base '$db_name' que je ne trouve pas dans la table DB... WTF?";
		return 0;
	}
	$sub_dbh->{'suffix'} = $suffix;
	purge($DBS, clone_dbh($sub_dbh));
	__del_base($base->{'BASE_ID'});


	__decr_db_cur($db->{'DB_ID'});
	return 1;
}


1;
