# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Alexis Sukrieh
package FDN::Sql::Plutar::API;
##############################################################
# Cette API est destine aux logiciels ayant besoin d'impacter 
# les donnes de Plutar.
# Par exemple une console de lecture des tches en cours, de 
# modification de tche ou encore d'ajout de nouvelle tche 
# pour une base donne.
#
# De mme une interface Web permettant de consulter les tches
# en cours de traitement par Plutar devrait passer par cette API 
# pour lire les tches de Plutar.
#
# On garanti ainsi une intgrit des donnes optimale en 
# canalisant les demandes de lecture par cete API.
#
# En gros personne ne doit selecter ou updater en direct dans la
# table Plutar, tout le monde passe par cette API.
#
# les fonctions de lecture seront prfixes de 'Get'
# les fonctions d'criture seront prfixes de 'Set'
# 
# -> FDN::Plutar::API::GetTaskList();
# 
# Les fonctions ne sont pas exportes volontairment pour obliger
# les applications utilisant l'API  les appeler avec leur chemin
# absolu. (pas de doute pour la provenance des methodes ainsi).
# 
# Chaque fonction prendra comme parametre unique un HREF contenant 
# les variables attendues. Ce hahs contiendra au minimum une entre 
# de la forme suivante :
# 	'dbh' => $dbh ou $dbh est un handler vers la base de donnes 
#
# les variables de retours de chaque fonctions seront sous forme 
# de HREF lequel contiendra au minimum une entre 
# 	'status' => TRUE|FALSE le flag indiquant le succs de l'opration
#
##############################################################
use strict;
use Exporter ;
our @ISA = ( 'Exporter' ) ;
our @EXPORT = qw (
);

use POSIX qw(strftime);
use FDN::Common::Logger;
use FDN::Common::Constantes;
use FDN::Sql::DB_Access;
use FDN::Sql::Basic;
use FDN::Sql::Plutar::Basic;

sub GetTaskList ($);				# renvoie la liste des tches listes dans PLUTAR
sub SetNewTask ($);				# ajoute une nouvelle tche
sub updateArgsTask ($);


# ----------------------------------------------------------------------
# nom 		: GetTaskList
# description 	: permet de lire le contenu de la table PLUTAR du handler
# 		  fourni et ainis de consulter l'intgrali des tches
# parametres 	: HREF contenant les entres suivantes :
# 			'dbh' => le handler vers la base de donnes
# 			'etat'=> l'tat attendu des tche (par dfaut *)
# 			'dtdo'=> la date  laquelle la tache doit etre execute (dfaut void)
# retours 	: HREF contenant les entres suivantes :
# 			'status' => TRUE ou FALSE selon le succs
# 			'erreur' => si erreur rencontre, le message
# 			'ra_task'=> un AREF contenant toutes les tches demande
# ----------------------------------------------------------------------
sub getTaskList ($) {
	my ($args, @conds);
	
	# vrification des arguments
	$args = shift;
	unless (ref ($args) eq 'HASH'  and
		defined ($args->{dbh}) ) {
		error "paramtre dbh attendu dans les arguments";
		return {
			status => FALSE,
			erreur => get_last_error_msg()
		};
	}

	# analyse de ce qu'on me demande  
	if (	exists $args->{etat} and 
		$args->{etat} ne '*'	){
		push @conds, [$args->{etat}, 'PLUTAR_ETAT', '='];
	}

	if (exists $args->{dtdo}) {
		push @conds, [$args->{dtdo}, 'PLUTAR_DTDO', '='];
	}
	
	# excution de la requte
	my @tasks = __par_champ_ext (
				$args->{dbh}, 
				'PLUTAR',
				\@conds,
			        { 'order' => 'up:PLUTAR_DTDO' }
			);

	return {
		status 	=> TRUE,
		ra_task => \@tasks
	};
}


# ----------------------------------------------------------------------
# nom 		: SetNewTask
# description 	: permet d'ajouter une nouvelle tche
# parametres 	: HREF contenant les entres suivantes :
# 			'dbh' => le handler vers la base de donnes
#			'action' => l'action a faire selon le fichier de conf
# 			'dtdo' => la date  laquelle la tache doit etre execute (dfaut now())
#			'args' => un HREF contenant ls arguments ncessaire  la fonction, 
#				  obligatoire donc si pas d'arguments donner {}
# 			
# retours 	: HREF contenant les entres suivantes :
# 			'status' => TRUE ou FALSE selon le succs
# 			'erreur' => si erreur rencontre, le message
# ----------------------------------------------------------------------
sub setNewTask ($)
{
	my ($rh_args, $ra_conds);
	
	# vrification des arguments
	$rh_args = shift;
	unless (ref ($rh_args) eq 'HASH'  and
		defined $rh_args->{dbh} and
		defined $rh_args->{action} and
		(defined $rh_args->{dtdo} or defined $rh_args->{nbjours}) and
		defined $rh_args->{args} and
		ref $rh_args->{args} eq 'HASH')  {
		error "paramtres obligatoires dbh, action et dtdo attendus dans les arguments";
		return {
			status => FALSE,
			erreur => get_last_error_msg()
		};
	}

	# si on nous donne nbjours, on crase la date dtdo avec la date dans nbjours
	if (defined $rh_args->{nbjours}) {
		my $next_time = time() + ($rh_args->{nbjours} * 24 * 3600);
		$rh_args->{dtdo} = strftime ('%Y%m%d%H%M%S', localtime($next_time));
		info "date dtdo gnr : " . $rh_args->{dtdo};
	}
	
	# insertion de la tche
	my $task_id = __new_plutar_task (
		$rh_args->{dbh},
		$rh_args->{action},
		$rh_args->{dtdo}
	);

	# controle des retours
	if (defined $task_id) {
		if ($task_id) {
			#ajout des arguments 
			__set_properties (
				$rh_args->{dbh},
				'PLUTAR',
				$task_id,
				$rh_args->{args}
			);			
			#malheuresuement, aucun moyen de savoir si __set_properties s'est bien pass	
			info "tache correctement insre (id = $task_id)";
			return {status => TRUE};
		}
		else {
			error "mauvais appel  la fonction __new_plutar ($task_id)";
			return {
				status => FALSE,
				erreur => get_last_error_msg()
			};
		}
	}
	else {
		error "impossible d'insrer la tche dans la base";
		return {
			status => FALSE,
			erreur => get_last_error_msg()
		};
	}
}


# ----------------------------------------------------------------------
# nom 		: updateArgsTask
# description 	: mes a jour les arguments contenus dans la table 
#                 PROPPLUTAR
# parametres 	: HREF contenant les entres suivantes :
# 			'dbh' => le handler vers la base de donnes
#			'plutar_id' => l'action a faire selon le fichier de conf
#			'args' => un HREF contenant ls arguments ncessaire  la fonction, 
#				  obligatoire donc si pas d'arguments donner {}
# 			
# retours 	: HREF contenant les entres suivantes :
# 			'status' => TRUE ou FALSE selon le succs
# 			'erreur' => si erreur rencontre, le message
# ----------------------------------------------------------------------
sub updateArgsTask ($)
{
	my ($rh_args);
	
	# vrification des arguments
	$rh_args = shift;
	unless (ref ($rh_args) eq 'HASH'  and
		defined $rh_args->{dbh} and
		defined $rh_args->{plutar_id} and
		defined $rh_args->{args} and
		ref $rh_args->{args} eq 'HASH')  {
		error "paramtres obligatoires dbh, action et dtdo attendus dans les arguments";
		return {
			status => FALSE,
			erreur => get_last_error_msg()
		};
	}

	# verification sur la coherence de la tache
	my $rh_plutar = __plutar_par_id($rh_args->{dbh} ,$rh_args->{plutar_id});
	
	# controle des retours
	if (defined $rh_plutar) {
			#mise a jour des arguments 
			__update_properties (
				$rh_args->{dbh},
				'PLUTAR',
				$rh_args->{plutar_id},
				$rh_args->{args}
			);			
			#malheuresuement, aucun moyen de savoir si __set_properties s'est bien pass	
			info "tache correctement mise  jour (id = ".$rh_args->{plutar_id}.")";
			return {status => TRUE};
	} else {
		error "impossible de retrouver la tache : ".$rh_args->{plutar_id};
		return {
			status => FALSE,
			erreur => get_last_error_msg()
		};
	}
}

TRUE;
