package FDN::Sql::Structure;

use strict;
use warnings;

use Encode;
use Data::Dumper;
use FDN::Sql::DB_Access qw/!drop_table/;
use FDN::Sql::Basic;
use FDN::Sql::DBC;
use FDN::Common::Config;
use FDN::Common::Modules;
use FDN::Common::Logger;
use FDN::Common::Utils;
use FDN::Common::Sign;

use base 'Exporter';

our @EXPORT = qw (
	read_config
	read_files

	switch_option
	set_quiet
	set_output
	check_db
	$subst
);

our @EXPORT_OK = qw (
	create
	structure
	purge
);

my $dbh;

my $quiet = 0;
my $output = 'stdout';
my $dry_alter = 1;
my $dry_db    = 1;
my $dry_table = 1;
my $dry_data  = 1;
my $dry_cfg   = 1;


# The new definition (target), will not change during process.





my $appli_id_par_name;
my $text_par_xid_appli_id;
my $create_text;
my $set_text_simple;
my $set_text_val;
my $new_appli;


my $dir; # Should become deprecated
my $source_cr;
my $source_db;
my $loc_cr;
my $loc_db;
my $applis;
my $connect;
my $texts;



sub __debug($) {
	my ( $msg ) = @_;
	return if $quiet;
	if ( $output eq 'syslog' ) {
		debug $msg;
	} else {
		print $msg."\n";
	}
}

sub __info($) {
	my ( $msg ) = @_;
	if ( $output eq 'syslog' ) {
		info $msg;
	} else {
		print $msg."\n";
	}
}

sub __error($) {
	my ( $msg ) = @_;
	if ( $output eq 'syslog' ) {
		error $msg;
	} else {
		print STDERR $msg."\n";
	}
}

#############################################################################################
# To build the sql scripts
#############################################################################################


sub read_files($);
sub read_files($) {
	my ( $DBS ) = @_;

	# Deja lu la derniere fois.
	return if $DBS->{'create'};

	my $create = '';
	for my $src ( @{$DBS->{'files'}} ) {
		__debug "Reading from file $src";
		$create .= slurp_file($src,@{$DBS->{'filters'}});
	}
	if ( ref($DBS->{'subst'}) eq 'HASH' ) {
		my ( $k, $v );
		while ( ($k,$v) = each %{$DBS->{'subst'}} ) {
			$create =~ s/$k/$v/sg;
		}
	}
	$DBS->{'create'} = $create;
	$DBS->{'signature'} = sign_b64($create,'');
	$DBS->{'queries'} = [ split /;\s*$/m, $create ];
	if ( ! $DBS->{'force'} ) {
		parse_db_structure($DBS);
	}
	if ( $DBS->{'modules'} ) {
		for my $mod ( keys %{$DBS->{'modules'}} ) {
			read_files($DBS->{'modules'}->{$mod});
		}
	}
}

#############################################################################################
# Some utilities...
#############################################################################################

sub run_or_die($$$) {
	my ( $dbh, $type, $req ) = @_;
	my $tmp = sql_transaction($dbh, $type, $req);
	if ( ! $tmp->{'ok'} ) {
		die "Failed with query $req\n".Dumper($tmp);
	}
	return $tmp;
}

sub dry_or_run($$$$$) {
	my ( $dry, $text, $dbh, $type, $req ) = @_;
	if ( $dry ) {
		__info "-- ".$text if defined $text;
		__info $req.";\n";
		return;
	}
	__info $text if defined $text;
	return run_or_die($dbh, $type, $req);
}

sub database_connexion($) {
	my ( $name ) = @_;

	$dbh = undef;
	my $err = [];
	my $cfg = Config_Section('dbc.cfg',$name);
	push @$err, "No configuration for database $name on that host (/etc/fdn/dbc.cfg)" if ! defined $cfg;
	eval { $dbh = do_connect($name) };
	push @$err, "Failed to connect to database $name on that host (/etc/fdn/dbc.cfg)" if ! defined $dbh;
	$dbh->{'suffix'} //= '';
	return __print_errors($err);
}


sub __print_errors($) {
	my ( $err ) = @_;

	return 1 if ! @$err;

	for my $e ( @$err ) {
		__error $e;
	}
	return 0;
}


##############################################################################################
# On the 'config' action
##############################################################################################

sub check_config_options($) {
	my ( $DBS ) = @_;
	my $err = [];
	push @$err, "No databse selected" if ! defined $DBS->{'name'};
	push @$err, "No login provided" if ! defined $DBS->{'login'};
	return __print_errors($err);
}

sub config($) {
	my ( $DBS ) = @_;

	if ( ! check_config_options($DBS) ) {
		die "Cannot perform config action";
	}

	my $cfg = {};
	my $add = "";
	if ( ! -f '/etc/fdn/dbc.cfg' ) {
		if ( ! -w '/etc/fdn' ) {
			die "Je n'ai pas le droit d'ecrire de config, je ne peux donc pas creer le fichier /etc/fdn/dbc.cfg";
		}
	} else {
		$cfg = Load_Config('dbc.cfg');
		if ( ! $cfg ) {
			die "Le fichier /etc/fdn/dbc.cfg est suffisament en vrac pour que je n'y touche pas.";
		}
	}

	my $name = $DBS->{'name'};
	my $host = $DBS->{'host'};
	$host //= 'localhost';
	if ( exists $cfg->{$name} ) {
		my $err;
		$cfg = $cfg->{$name};
		__debug "Configuration for $name already exists.";
		$err .= "Wrong type for $name, should be 'mysql'\n" if ( $cfg->{'type'} ne 'mysql' );
		$err .= "Wrong host for $name, is '$cfg->{'host'}' should be '$host'\n" if ( $cfg->{'host'} ne $host );
		$err .= "Wrong name for $name\n" if ( $cfg->{'name'} ne $name );
		if ( $err ) {
			die "Database $name is wrongly configured in /etc/fdn/dbc.cfg config file:\n$err\n";
		}
		return;
	}

	my $login  = $DBS->{'login'};
	my $utf8   = $DBS->{'utf8'};
	my $passwd = $DBS->{'passwd'};
	$passwd //= 'toto';

	__info "Need to add $name in config file.";
	$add .= "[$name]\n".
		"\ttype\t= mysql\n".
		"\thost\t= $host\n".
		"\tname\t= $name\n".
		"\tuser\t= ".$login."\n".
		"\tpasswd\t= ".$passwd."\n";
	$add .= "\tutf8\t= ".$utf8."\n" if defined $utf8;
	$add .= "\n";

	if ( $dry_cfg && ! $DBS->{'force'} ) {
		__info "Need to update config file /etc/fdn/dbc.cfg by adding:\n".$add;
		return;
	}

	__info "Updating config file";
	Forget_Config('dbc.cfg');
	open OUT, ">>/etc/fdn/dbc.cfg" or die "Cannot update /etc/fdn/dbc.cfg: $!\n";
	print OUT $add;
	close OUT;
}

##############################################################################################
# On the 'create' action
##############################################################################################

my $connected_to_global = 0;
END {
	if ( $connected_to_global ) {
		do_disconnect('GLOBAL');
	}
}
sub check_create_options($) {
	my ( $DBS ) = @_;
	my $err = [];

	push @$err, "No database selected" if ! defined $DBS->{'name'} || ! $DBS->{'name'};
	push @$err, "No login provided" if ! defined $DBS->{'login'} || ! $DBS->{'login'};
	return __print_errors($err) if @$err;

	$connected_to_global = 1;
	return database_connexion('GLOBAL');
}

sub create($) {
	my ( $DBS ) = @_;

	if ( ! check_create_options($DBS) ) {
		return;
	}
	my $name   = $DBS->{'name'};
	my $login  = $DBS->{'login'};
	my $source = $DBS->{'source'};
	my $passwd = $DBS->{'passwd'};
	$source //= 'localhost';
	$passwd //= 'toto';

	my $tmp = run_or_die($dbh,'show','show databases');
	for my $r ( @{$tmp->{'rows'}} ) {
		my $db = $r->{'Database'};
		if ( $db eq $name ) {
			__debug "Database already exists";
			return;
		}
	}

	__info "-- Need to create database $name";
	my ( $rq1, $rq2, $rq3, $rq4 );
	$rq1 = "create database $name";
	#$rq2 = 'grant all privileges on * to \''.$login.'\'@\''.$source.'\' identified by \''.$passwd.'\'';
	$rq3 = 'grant all privileges on '.$name.'.* to \''.$login.'\'@\''.$source.'\' identified by \''.$passwd.'\'';
	$rq4 = 'flush privileges';
	if ( $dry_db && ! $DBS->{'force'} ) {
		__info $rq1.";";
		#print $rq2.";\n";
		__info $rq3.";";
		__info $rq4.";";
	} else {
		run_or_die($dbh,'create',$rq1);
		#run_or_die($dbh,'grant',$rq2);
		run_or_die($dbh,'grant',$rq3);
		run_or_die($dbh,'flush',$rq4);
	}
	disconnect_db($dbh);
	$dbh = undef;
}

##############################################################################################
# Reading the structure of the database
##############################################################################################

sub read_table_structure($$) {
	my ( $table, $sql ) = @_;
	my $res = {};
	$sql =~ s/\s*--.*$//gm;
	if ( $sql !~ /\s*\((.*)\)\s*(\S.*\S)?\s*$/is ) {
		die "Cannot understand table create for $table in <$sql>";
	}
	my ( $columns, $options )  = ( $1, $2 );
	$res->{'sql'} = $sql;
	$res->{'cols'} = { list => []};
	my $err = 0;
	while ( $columns =~ s/^\s*
				`?(\w+)`?
			       \s+
				(  (?: bigint | int | integer | tinyint )(?:\s*\(\d+\))?  (?:\s+unsigned)?
				 | (?: char | varchar	      )(?:\s*\(\d+\))	(?:\s+character\s+set\s+\S+)? (?:\s+collate\s+\S+\s+)?
				 | text						(?:\s+character\s+set\s+\S+)? (?:\s+collate\s+\S+\s+)?
				 | enum\s*\([^)]+\)				(?:\s+character\s+set\s+\S+)? (?:\s+collate\s+\S+\s+)?
				 | (?: binary | varbinary     )(?:\s*\(\d+\))
				 | boolean | bool
				 | datetime | date 
				 | timestamp(?:\s*\(\d+\))?
				)
			       \s*
				(  (?:unique\s+)?
				   (?: null | (?:not\s+null)?  (?:\s*default\s+(?:\S+|'[^']*')(?:\s+on\s+update\s+\S+)?)? | default\s+\S+\s+not\s+null )
				)
			       \s*
				(,|$)//isx ) {
		my ($col, $type, $dflt) = ($1, $2, $3);
		push @{$res->{'cols'}->{'list'}}, $col;
		$res->{'cols'}->{$col} = { type => $type, dflt => $dflt };
		if ( $dflt =~ /default\s+\S+\s+not\s+null/i ) {
			$err = 1;
			__error "ERR not null and default in wrong order, in $table.$col";
		}
	}
	$res->{'idxs'} = {};
	while ( $columns =~ s/^\s*
				( primary\s+key | index | unique(?:\s+index)? | key | unique\skey | fulltext(?:\s+(?:index|key))? )
			       \s*
				(`\w+`)?
			       \s*
				(\( \s*`?\w+`?\s*(?:\(\s*\d+\s*\)\s*)? (?:, \s*`?\w+`?(?:\(\s*\d+\s*\)\s*)?)* \))
			       \s*
				(,|$)//isx ) {
		my ( $type, $name, $idx ) = ( $1, $2, $3 );
		$idx =~ s/`//g;
		$idx =~ s/\s//g;
		if ( defined $res->{'idxs'}->{$idx} ) {
			$err = 1;
			__error "ERR: There are several indexes about $idx in table $table";
		} else {
			$res->{'idxs'}->{$idx} = []
		}
		push @{$res->{'idxs'}->{$idx}}, { type => $type, name => $name };
	}
	if ( $columns !~ /^\s*$/is) {
		die "J'ai essay d'analyser a:\n$sql\n\nIl me reste a que je ne sais pas lire:\n<<<$columns>>>\n\n";
	}

	$res->{'opts'} = {};
	$options //= '';
	while ( $options =~ s/\s*(\S[^=]*\S)\s*=\s*(\S*)\s*// ) {
		my ( $opt, $val ) = ( $1, $2 );
		$opt = 'type' if ( uc($opt) eq 'ENGINE' );
		# bizarrement la doc de MySQL semble indiqu que la syntaxe
		# pour le charset c'est : [DEFAULT] CHARACTER SET
		# pourtant show create table renvois DEFAULT CHARSET...
		$opt = 'default charset' if ( uc($opt) eq 'DEFAULT CHARSET' );
		$res->{'opts'}->{$opt} = $val;
	}
	if ( $err ) {
		__error "ERR J'ai vu des problmes:\n".$sql."\n".Dumper($res);
	}
	return $res;
}

sub get_old_tables_description($$) {
	my ( $dbh, $db ) = @_;
	my $tmp = run_or_die($dbh,'show','show tables');
	my $tables = {};
	my $suffix = $dbh->{'suffix'};
	for my $r ( @{$tmp->{'rows'}} ) {
		if ( ! defined $db ) {
			Carp::confess "Je devrais avoir le nom de la base...";
		}
		my $tbl = $r->{'Tables_in_'.$db};
		next if $suffix && $tbl ne 'SEQ' && $tbl !~ s/^([A-Z]+)$suffix$/$1/;
		$tables->{$tbl} = 1;
	}
	for my $t ( keys %$tables ) {
		my $tname = $t;
		$tname .= $suffix if $suffix && $t ne 'SEQ';
		$tmp = get_table_create($dbh,$tname);
		$tables->{$t} = read_table_structure($tname, $tmp);
	}
	return $tables;
}

sub read_limit_info($) {
	my ( $req ) = @_;
	my ( $type, $limit, $value, $script, $vid );
	my $res = {};
	if ( $req =~ s/-- limit (\S+) script (\S+) (\S+) (\S+)\s*$//m ) {
		( $type, $limit, $value, $script ) = ( $1, $2, $3, $4 );
	} else {
		die "There is a limit on the create sequence I cannot handle";
	}
	
	if ( $limit !~ /^(VERSION_NAME(\((\d+)\))?)$/ ) {
		die "There is a limit, but not about VERSION_NAME, I cannot handle";
	} else {
		$vid = $3;
	}
	if ( $value !~ /^\d+$/ ) {
		die "There is a limit on VERSION_NAME with non numerical version number";
	}
	if ( $script !~ /^\// ) {
		$script = '/usr/bin/'.$script;
	}
	if ( ! -x $script ) {
		die "There is a script I am supposed to run, $script, but I cannot find it";
	}
	$res->{'type'} = $type;
	$res->{'version'} = $value;
	$res->{'script'} = $script;
	$res->{'vid'} = $vid;
	return $res;
}

sub parse_db_structure($) {
	my ( $DBS ) = @_;
	$DBS->{'limits'} = {
		'pre' => [],
		'pre-del' => [],
		'pre-create' => [],
		'post-del' => [],
		'pre-content' => [],
		'post-content' => [],
		'post' => []
	};
	$DBS->{'tables'} = {};

	my $limits = join '|', keys %{$DBS->{'limits'}};

	my $qs = $DBS->{'queries'};
	for my $q ( @$qs ) {
		while ( $q =~ /-- limit ($limits) / ) {
			push @{$DBS->{'limits'}->{$1}}, read_limit_info($q);
			if ( $q !~ s/(-- limit .*)$//m ) {
				die "Pas pu retirer la limite dans $q\n";
			}
		}
		$q =~ s/^-- .*$//m;
		next if $q =~ /^\s*$/s;
		if ( $q =~ s/create\s+table\s+(\w+)\b//is ) {
			my $tname = $1;
			if ( $tname !~ /^[A-Z]+$/ ) {
				die "La structure de la base demande un table '$tname' dont le nom ne contient pas que des lettres capitales. Ca risque fort de ne pas marcher. Donc je m'arrete.";
			}
			$DBS->{'tables'}->{$tname} = read_table_structure($tname,$q);
		}
	}
}

sub get_current_versions($) {
	my ( $dbh ) = @_;
	my $tmp;
	$tmp = sql_transaction($dbh,'select','select VERSION_NAME, VERSION_ID from VERSION'.$dbh->{'suffix'});
	if ( ! $tmp->{'ok'} ) {
		return -1;
	}
	my $res = {};
	for my $l ( @{$tmp->{'rows'}} ) {
		$res->{$l->{'VERSION_ID'}} = $l->{'VERSION_NAME'};
	}
	return $res;
}



#####################################################
# Modifications on the databse structure
#####################################################


sub __suffix_table($$) {
	my ( $table, $suffix ) = @_;
	return 'SEQ' if $table eq 'SEQ';
	return $table.$suffix->{'suffix'} if ref($suffix);
	return $table.$suffix;
}


sub drop_column($$$) {
	my ( $dbh, $table, $col ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "alter table $tname drop column $col";
	dry_or_run($dry_alter, "Dropping column $col from table $tname", $dbh, 'alter', $req);
	return 1;
}

sub add_column($$$$$$) {
	my ( $dbh, $table, $col, $type, $default, $position ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "alter table $tname add column $col $type $default $position";
	dry_or_run($dry_alter, "Adding column $col to table $tname", $dbh, 'alter', $req);
	return 1;
}

sub modify_column($$$$$) {
	my ( $dbh, $table, $col, $type, $default ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "alter table $tname modify $col $type $default";
	dry_or_run($dry_alter, "Changing type of column $tname.$col", $dbh, 'alter', $req);
	return 1;
}

sub set_table_option($$$$) {
	my ( $dbh, $table, $name, $value ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "alter table $tname ";
	$req .= 'ENGINE' if ( $name eq 'type' );
	$req .= 'default charset' if ( $name eq 'default charset' );
	$req .= " = ".$value;
	dry_or_run($dry_alter, "Changing option $name on table $tname", $dbh, 'alter', $req);
	return 1;
}

sub drop_index($$$$) {
	my ( $dbh, $table, $type, $name ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req;
	if ( $type =~ /^(unique |fulltext )?(index|key)$/i ) {
		$req = "alter table $tname drop index $name";
	} elsif ( $type =~ /^primary key$/i ) {
		$req = "alter table $tname drop primary key";
	} else {
		die "Unknown column type $type";
	}
	dry_or_run($dry_alter, "Dropping index $type/$name on table $tname", $dbh, 'alter', $req);
}

sub create_index($$$$) {
	my ( $dbh, $table, $type, $keys ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req;
	$req = "alter table $tname add $type$keys";
	dry_or_run($dry_alter, "Adding index $type$keys on table $tname", $dbh, 'alter', $req);
}

sub create_table($$$) {
	my ( $dbh, $table, $sql ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "create table $tname $sql";
	dry_or_run($dry_table, "Creating new table $tname", $dbh, 'create', $req);
}

sub drop_table($$) {
	my ( $dbh, $table ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "drop table $tname\n";
	dry_or_run($dry_table, "Dropping deprecated table $tname", $dbh, 'drop', $req);
}



#############################################################
# Modifications on the databse content
#############################################################

sub __seq_name($$) {
	my ( $dbh, $name ) = @_;
	if ( $dbh->{'suffix'} ) {
		my $suffix = $dbh->{'suffix'};
		if ( $name =~ /_[A-Z]+$/ ) {
			$name =~ s/(_[A-Z]+)$/$suffix$1/;
		} else {
			$name = $name.$suffix;
		}
	}
	return $name;
}


sub create_sequence($$$$) {
	my ( $dbh, $db, $name, $seq_info ) = @_;
	$name = __seq_name($dbh,$name);
	my $req = "replace into SEQ values ('$name', $seq_info)";
	dry_or_run($dry_data, "Creating SEQ $name", $dbh, 'replace', $req);
}

sub update_seq($$$$$$) {
	my ( $dbh, $db, $name, $min, $cur, $max ) = @_;
	$name = __seq_name($dbh,$name);
	my $req = "update SEQ set ";
	my @elems;
	push @elems, "SEQ_MINVALUE = '$min'" if defined $min;
	push @elems, "SEQ_MAXVALUE = '$max'" if defined $max;
	push @elems, "SEQ_VALUE = '$cur'" if defined $cur;
	$req .= join ", ", @elems;
	$req .= " where SEQ_NAME = '$name'";
	dry_or_run($dry_data, "Updating SEQ $name to [$min,$min,$max]", $dbh, 'update', $req);
}

sub update_line($$$$) {
	my ( $dbh, $db, $table, $values ) = @_;
	my $tname = __suffix_table($table,$dbh);
	my $req = "replace into $tname $values";
	dry_or_run($dry_data, "Missing/Deprecated line in table $db.$tname", $dbh, 'replace', $req);
}


sub control_line_id($$$$) {
	my ( $dbh, $db, $table, $id ) = @_;
	my $suffix = $dbh->{'suffix'};
	my $sname = $table.'_ID';
	my $seq = get_current($dbh,$sname);
	my $nseq;
	while ( $seq < $id ) {
		if ( $dry_alter ) {
			__info "-- Need to update SEQ table, no way to do that with a single request ( $seq < $id )\n";
			return;
		}
		__info "Updating SEQ on $table$suffix: $seq < $id\n";
		$nseq = get_next($dbh,$sname);
		if ( ! $nseq ) {
			die "Cannot update SEQ table for $sname\n";
		}
		if ( $nseq <= $seq ) {
			die "SEQ did not increase during update for $sname\n";
		}
		$seq = $nseq;
	}
}

sub purge_db($$) {
	my ( $dbh, $db_name ) = @_;
	my ( $tmp, @tables );
	__info "Purge content of $db_name";
	$tmp = run_or_die($dbh,'show','show tables');
	return if ! defined $tmp->{'rows'};
	@tables = @{$tmp->{'rows'}};
	for my $t ( @tables ) {
		my $name = $t->{'Tables_in_'.$db_name};
		dry_or_run($dry_table, undef, $dbh, 'drop', "drop table $name");
	}
}

sub execute_queries($$) {
	my ( $dbh, $sql ) = @_;
	my @queries;
	if ( ref($sql) eq 'ARRAY' ) {
		@queries = @$sql;
	} else {
		@queries = split /;\s*$/m, $sql;
	}
	my $tmp;
	for my $q ( @queries ) {
		next if $q =~ /^-- /;
		next if $q =~ /^\s*$/;
		my $text;
		if ( $q =~ /create\s+table\s+([a-z0-9_]+)\W/is ) {
			$text = " $1,";
		}
		dry_or_run($dry_table, $text, $dbh, 'create', $q);
	}
}


#############################################################
# Comparaison de deux structures de base
#############################################################


sub get_base_type($) {
	my ( $type ) = @_;
	if ( $type =~ /tinyint|int|integer|bigint|bool|^enum/i ) {
		return 'int';
	} elsif ( $type =~ /timestamp/i ) {
		return 'ts';
	} elsif ( $type =~ /date/i ) {
		return 'date';
	} else {
		return 'text';
	}
}




sub compare_columns($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $redo = 0;
	for my $c ( @{$old->{'cols'}->{'list'}} ) {
		if ( ! exists $new->{'cols'}->{$c} ) {
			$redo += drop_column($dbh, $table, $c);
			delete $old->{'cols'}->{$c};
		}
	}
	my $prev;
	for my $c ( @{$new->{'cols'}->{'list'}} ) {
		if ( ! exists $old->{'cols'}->{$c} ) {
			my $type = $new->{'cols'}->{$c}->{'type'};
			my $dflt = $new->{'cols'}->{$c}->{'dflt'};
			my $pos;
			$pos = 'first' if ( ! $prev );
			$pos = 'after '.$prev if ( $prev );
			$redo += add_column($dbh, $table, $c, $type, $dflt, $pos);
			$old->{'cols'}->{$c} = $new->{'cols'}->{$c};
		}
		$prev = $c;
	}
	$old->{'cols'}->{'list'} = [ @{$new->{'cols'}->{'list'}} ];
	return $redo;
}

sub compare_column_types($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $old_charset = 'INVALID';
	if ( defined $old->{'opts'}->{'default charset'} ) {
		$old_charset = $old->{'opts'}->{'default charset'};
	}
	# Maintenant, on a les mme colonnes, vrifier les types.
	my $first_date = 1;
	my $redo = 0;
	for my $c ( @{$old->{'cols'}->{'list'}} ) {
		my $new_type = $new->{'cols'}->{$c}->{'type'};
		my $new_dflt = $new->{'cols'}->{$c}->{'dflt'};
		my $old_type = $old->{'cols'}->{$c}->{'type'};
		my $old_dflt = $old->{'cols'}->{$c}->{'dflt'};

		my $old_base = get_base_type($old_type);
		my $new_base = get_base_type($new_type);

		if ( $old_base ne $new_base ) {
			__info "-- Need to change column base_type from '$old_base' to '$new_base' (strict compare)";
			$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt);
			next;
		}

		my ( $re_old, $re_new ) = ( $old_type, $new_type );
		for my $re ( $re_old, $re_new ) {
			$re =~ s/\s*character\s+set\s+$old_charset\s*//;
			$re =~ s/\(/\\\(/g;
			$re =~ s/\)/\\\)/g;
			$re =~ s/'\s*,\s*'/'\\s*,\\s*'/g if ( $re =~ /enum/i );
			$re =~ s/\s*\\\(/\\s*\\\(/g;
			$re =~ s/\s+/\\s+/g;
			$re =~ s/(int\\s\*\\\(10\\\)\\s+unsigned|\bint(eger)?\\s+unsigned)/(int\\s\*\\\(10\\\)\\s+unsigned|\\bint(eger)?\\s+unsigned)/;
			$re =~ s/(int\\s\*\\\(11\\\)|\bint(eger)?)/(int\\s\*\\\(11\\\)|\\bint(eger)?)/;
			$re =~ s/(timestamp(\\s\*\\\(14\\\))?)/(timestamp(\\s\*\\\(14\\\))?)/;
			$re = '(tinyint\(1\)|bool|boolean)' if ( $re =~ /^\s*(tinyint\\s\*\\\(1\\\)|bool|boolean)\s*$/i );
		}
		if ( $old_type !~ /$re_new/i || $new_type !~ /$re_old/i ) {
			my $msg = "-- Need to change column type from '$old_type' to '$new_type' ";
			if ( $old_type !~ /$re_new/i ) {
				$msg .= "(old type does not match $re_new)";
			} else {
				$msg .= "(new type does not match $re_old)";
			}
			__info $msg;
			$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt);
			next;
		}
		if ( $old_base eq $new_base ) {
			my $re = $new_dflt;
			if ( $old_base eq 'ts' ) {
				if ( $first_date ) {
					$first_date = 0;
					if ( $re eq '' ) { $re = 'not\\s+null\\s+default\\s+CURRENT_TIMESTAMP\\s+on\\s+update\\s+CURRENT_TIMESTAMP'; } 
				} else {
					if ( $re eq '' ) { $re = 'not\\s+null\\s+default\\s+\'0000-00-00 00:00:00\''; }
				}
			}

			if ( $re eq '' ) { $re = '((default\s*)?null)?'; }
			if ( $re eq 'null') { $re = '((default\s|)null|)'; }
			if ( $re =~ /default/i ) {
				if ( $re !~ /not null/i ) {
					if ( $re !~ /null/i ) {
						$re = '(not\s+null\s+)?'.$re;
					}
				}
			}
			if ( $re !~ /default (null|CURRENT_TIMESTAMP)/i ) {
				$re =~ s/default ('|)([^']*)\1(\s|$)/default '$2'/;
			}
			$re =~ s/'true'/'1'/;
			$re =~ s/'false'/'0'/;
			if ( $old_dflt !~ /^$re$/i ) {
				__info "-- Need to change column default from '$old_dflt' to '$new_dflt' (old default does not match $re)";
				$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt );
			}
		}
	}
	return $redo;
}

sub compare_table_options($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $redo = 0;
	for my $o ( keys %{$old->{'opts'}} ) {
		if ( exists $new->{'opts'}->{$o} ) {
			$old->{'opts'}->{$o} =~ s/^('|)(.*)\1$/$2/;
			$new->{'opts'}->{$o} =~ s/^('|)(.*)\1$/$2/;
			if ( $old->{'opts'}->{$o} ne $new->{'opts'}->{$o} ) {
				$redo += set_table_option($dbh,$table,$o,$new->{'opts'}->{$o});
			}
		} else {
			next if ( $o eq 'type' && $old->{'opts'}->{$o} =~ /myisam/i );
			next if ( $o eq 'default charset' );
			if ( $dry_alter ) {
				__info "-- Option '$o' deleted from table $table (was ".$old->{'opts'}->{$o}.")";
				__info "-- Cannot remove an option from a table. CHECK.";
			} else {
				die "Cannot remove an option from a table (option $o from table $table)\n";
			}
		}
	}
	for my $o ( keys %{$new->{'opts'}} ) {
		if ( ! exists $old->{'opts'}->{$o} ) {
			$redo += set_table_option($dbh,$table,$o,$new->{'opts'}->{$o});
		}
	}
	return $redo;
}

sub compare_indices($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	for my $i ( keys %{$old->{'idxs'}} ) {
		my $old_type = $old->{'idxs'}->{$i}->[0]->{'type'};
		my $old_name = $old->{'idxs'}->{$i}->[0]->{'name'};
		my $re;
		if ( $old_type eq 'KEY' ) {
			$re = '(?:index|key)';
		} elsif ( $old_type eq 'PRIMARY KEY' ) {
			$re = 'primary key';
		} elsif ( $old_type eq 'UNIQUE KEY' ) {
			$re = 'unique(?:\s+index|\s+key|)';
		} elsif ( $old_type eq 'FULLTEXT KEY' ) {
			$re = 'fulltext(?:\s+index|\s+key|)';
		}
		if ( exists $new->{'idxs'}->{$i} ) {
			my $new_type = $new->{'idxs'}->{$i}->[0]->{'type'};
			if ( $new_type !~ /$re/i ) {
				drop_index($dbh, $table, $old_type, $old_name);
				create_index($dbh, $table, $new_type, $i);
			}
		} else {
			drop_index($dbh, $table, $old_type, $old_name);
		}
	}
	for my $i ( keys %{$new->{'idxs'}} ) {
		my $new_type = $new->{'idxs'}->{$i}->[0]->{'type'};
		if ( ! exists $old->{'idxs'}->{$i} ) {
			create_index($dbh, $table, $new_type, $i);
		}
	}
	return 0;
}

sub compare_tables($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;

	my $redo = 0;
	$redo += compare_columns($dbh, $table, $old, $new);
	$redo += compare_column_types($dbh, $table, $old, $new);
	$redo += compare_table_options($dbh, $table, $old, $new);
	return 1 if ( $redo );
	compare_indices($dbh, $table, $old, $new);
	return 0;
}









sub check_limits($$$) {	
	my ( $dbh, $limits, $curver ) = @_;
	if ( ref $curver ne 'HASH' ) {
		if ( $curver != -1 ) {
			die "Invalid current version for check limit";
		}
		__info "It's new DB, don't check limit";
		return;
	}
	for my $l ( @$limits ) {
		my $current;
		if ( defined $l->{'vid'} ) {
			$current = $curver->{$l->{'vid'}};
			$current //= 0;
		} else {
			my @vals = values %$curver;
			if ( @vals > 1 ) {
				die "I cannot check limits against several versions with so few informations";
			}
			$current = $vals[0];
		}
		if ( $current < $l->{'version'} ) {
			my $script = $l->{'script'};
			if ( $dbh->{'suffix'} ) {
				$script .= ' '.$dbh->{'suffix'};
			}
			if ( $dry_table ) {
				__info "-- I have to run $script";
			} else {
				__info "Running $script";
				system($script) == 0 or die "Failed to run script $script";
			}
		}
	}
}

sub control_sequence($$$$) {
	my ( $dbh, $db, $name, $info ) = @_;

	my $sname = __seq_name($dbh,$name);
	my $tmp = sql_transaction($dbh, 'select', "select * from SEQ where SEQ_NAME = '$sname'");
	if ( ! $tmp->{'ok'} ) {
		die "Cannot read SEQ named $sname, in database $db:".Dumper($tmp);
	}
	my $seq = $tmp->{'rows'}->[0];
	my $omin = $seq->{'SEQ_MINVALUE'};
	my $omax = $seq->{'SEQ_MAXVALUE'};

	if ( $info !~ /^\s*(\d+|0x[0-9a-fA-F]+)\s*,\s*(\d+|0x[0-9a-fA-F]+)\s*,\s*(\d+|0x[0-9a-fA-F]+)\s*,.*$/ ) {
		die "Failed to read SEQ numerical info from <$info> in query about SEQ $name";
	}
	my ( $min, $max ) = ( $2, $3 );
	{
		no warnings 'portable'; # Hexadecimal number > 0xffffffff non-portable
		$min = oct($min) if $min =~ /x/;
		$max = oct($max) if $max =~ /x/;
	}

	return if ( $min == $omin && $max == $omax ); # Range is the same, nothing to be done.

	my $incr = $seq->{'SEQ_INC'};
	my $ocurrent = $seq->{'SEQ_VALUE'};
	# update_seq va ajouter tout seul le suffix s'il y en a besoin
	if ( $incr > 0 ) {
		if ( $ocurrent < $min || $omin > $max ) {
			# There is no overlap between [$min,$max] and [$omin,$ocurrent]
			# -> update the MIN and MAX and VALUE
			return update_seq($dbh,$db,$name,$min,$min,$max);
		}

		if ( $min == $omin ) {
			if ( $max > $omax || $max > $ocurrent ) {
				# Resizing compatible with the current value of the SEQ
				# -> update the MAX
				return update_seq($dbh,$db,$name,undef,undef,$max);
			}
		}
	}

	if ( $incr < 0 ) {
		if ( $max < $ocurrent || $omax < $min ) {
			# There is no overlap between [$min,$max] and [$ocurrent,$omax]
			# -> update the MIN and MAX and VALUE
			return update_seq($dbh,$db,$name,$min,$max,$max);
		}
		if ( $max == $omax ) {
			if ( $min < $omin || $min < $ocurrent ) {
				# Resizing compatible with the current value of SEQ
				# -> update the MIN
				return update_seq($dbh,$db,$name,$min,undef,undef);
			}
		}
	}

	die "You try to change sequence $name from [$omin,$ocurrent,$omax] step $incr to [$min,XXX,$max], which is not a pure resizing, neither a pure move. It would be unsafe to do so.";
}

sub check_sequence($$$) {
	my ( $dbh, $db, $q ) = @_;
	my $suffix = $dbh->{'suffix'};
	if ( $q !~ /^\s*replace into SEQ\s+values\s*\('([a-z0-9_.-]+)'\s*,(.*)\)\s*$/i ) {
		die "Failed to read SEQ info: $q\n";
	}
	my $name = $1;
	my $seq_info = $2;
	my $seq  = get_current($dbh,$name);
	if ( ! defined $seq ) {
		create_sequence($dbh, $db, $name, $seq_info);
	} else {
		# I may need to update the SEQ entry
		control_sequence($dbh, $db, $name, $seq_info);
	}
}

sub is_line_diff($$$) {
	my ( $structure, $new, $old ) = @_;
	for my $k ( keys %$new ) {
		next if ( ! defined $old->{$k} && $new->{$k} =~ /^NULL$/i );
		next if ( $new->{$k} eq 'now()' );
		if ( $structure->{'cols'}->{$k}->{'type'} =~ /^bool(?:ean)?$/i ) {
			next if ( ! $old->{$k} && $new->{$k} =~ /^false$/i );
			next if (   $old->{$k} && $new->{$k} =~ /^true$/i );
			__info "-- Differs on boolean value on $k: is ".$old->{$k}." should be ".$new->{$k};
			return 1;
		}
		if ( ! defined $old->{$k} ) {
			__info "-- Differs at least on $k: is NULL should be ".$new->{$k};
			return 1;
		}
		if ( $old->{$k} ne $new->{$k} ) {
			__info "-- Differs at least on $k: is ".$old->{$k}." should be ".$new->{$k};
			return 1;
		}
	}
	return 0;
}


sub check_insert($$$$) {
	my ( $dbh, $db, $tables, $q ) = @_;

	if ( $q !~ /^\s*insert\s+into\s+(`|)(\w+)\1\s+(values\s*\((.*)\))\s*$/ixs ) {
		die "Cannot check insert: $q\n";
	}
	#my $quote = $1;
	my $table = $2;
	my $values = $3;
	my $tmp = $4.',';
	my $t = {};
	for my $c ( @{$tables->{$table}->{'cols'}->{'list'}} ) {
		if ( $tmp !~ s/^\s*("|'|)(.*?)\1\s*,// ) {
			die "In table $table.\nSo far got ".Dumper($t)."\nCannot read after $tmp";
		}
		my ( $a, $b ) = ( $1, $2 );
		$t->{$c} = $2;
	}
	if ( $tmp ) {
		die "Too many values in $q\nRemains at the end <$tmp>\n";
	}

	if ( ! exists $t->{$table.'_ID'} ) {
		__error "-- Je ne sais pas traiter les donnes des tables de relation (table $table)";
		return;
	}

	my $id = $t->{$table.'_ID'};
	my $row = __par_id($dbh,$table,$id);
	if ( ! defined $row ) {
		update_line($dbh, $db, $table, $values);
		control_line_id($dbh, $db, $table, $id);
	} else {
		my $diff = is_line_diff($tables->{$table},$t,$row);
		return if ! $diff;
		update_line($dbh, $db, $table, $values);
	}
}


sub check_db_structure($$$) {
	my ( $dbh, $DBS, $curver ) = @_;

	my $redo = 1;
	my $nb_run = 0;
	my $tables_old;
	my $tables_new = $DBS->{'tables'};
	check_limits($dbh,$DBS->{'limits'}->{'pre'},$curver);
	while ( $redo && $nb_run < 2 ) {
		$nb_run++;
		$redo = 0;
		$tables_old = get_old_tables_description($dbh, $DBS->{'name'});

		for my $t ( keys %$tables_new ) {
			if ( $tables_old->{$t} ) {
				$redo += compare_tables($dbh, $t, $tables_old->{$t}, $tables_new->{$t});
			}
		}
	}

	check_limits($dbh,$DBS->{'limits'}->{'pre-create'},$curver);
	for my $t ( keys %$tables_new ) {
		if ( ! $tables_old->{$t} ) {
			create_table($dbh, $t, $tables_new->{$t}->{'sql'});
		}
	}

	check_limits($dbh,$DBS->{'limits'}->{'pre-del'},$curver);
	for my $t ( keys %$tables_old ) {
		if ( ! $tables_new->{$t} ) {
			drop_table($dbh, $t);
		}
	}
	check_limits($dbh,$DBS->{'limits'}->{'post-del'},$curver);
}


sub check_db_content($$$) {
	my ( $dbh, $DBS, $curver ) = @_;
	my $dbname = $DBS->{'name'};
	check_limits($dbh,$DBS->{'limits'}->{'pre-content'},$curver);
	for my $q ( @{$DBS->{'queries'}} ) {
		$q =~ s/^-- .*$//mg;
		$q =~ s/\/\*.*\*\///mg;
		$q =~ s/^\s*//s;
		$q =~ s/\s*$//s;
		next if ( $q =~ /^create table/i );
		next if ( $q =~ /^drop table/i );
		if ( $q =~ /^\s*replace into SEQ/i ) {
			check_sequence($dbh, $dbname, $q);
			next;
		}
		if ( $q =~ /^\s*insert\s+into\s+(`|)\w+\1\s+/i ) {
			check_insert($dbh, $dbname, $DBS->{'tables'}, $q);
		}
	}
	check_limits($dbh,$DBS->{'limits'}->{'post-content'},$curver);
}

sub purge_db_structure($$) {
	my ( $dbh, $DBS ) = @_;
	my $tables = get_old_tables_description($dbh, $DBS->{'name'});
	for my $t ( keys %$tables ) {
		next if $t eq 'SEQ';
		drop_table($dbh, $t);
	}
}

###########################################################################
# On the structure action
###########################################################################



sub check_connexion($;$) {
	my ( $DBS, $outer_dbh ) = @_;
	if ( $outer_dbh ) {
		$dbh = $outer_dbh;
		$dbh->{'suffix'} //= '';
		return 1;
	}

	my $err = [];
	my $name = $DBS->{'name'};
	if ( ! -r '/etc/fdn/dbc.cfg' ) {
		push @$err, "No access configured on that host";
	}
	if ( ! defined $name ) {
		push @$err, "No database selected";
	}
	if ( @$err ) {
		return __print_errors($err);
	}

	return database_connexion($name);
}

sub check_structure_options($) {
	my ( $DBS ) = @_;

	my $err = [];
	if ( ref($DBS) ne 'HASH' ) {
		Carp::confess "DBS should be a hash ref. Here is a full stack trace. Please investigate.";
	}
	if ( ! exists $DBS->{'files'} || ref($DBS->{'files'}) ne 'ARRAY' ) {
		Carp::confess "There should be a 'files' key in DBS, which should be an ARRAY-ref. Here is a full stack trace. Please investigage.";
	}
	if ( ! exists $DBS->{'filters'} || ref($DBS->{'filters'}) ne 'ARRAY' ) {
		Carp::confess "There should be a 'filters' key in DBS, which should be an ARRAY-ref. Here is a full stack trace. Please investigage.";
	}
	push @$err, "No sql file selected" if ! @{$DBS->{'files'}};
	for my $f ( @{$DBS->{'files'}} ) {
		if ( ! -e $f ) {
			push @$err, "File not found: $f";
			next;
		}
		if ( ! -r $f ) {
			push @$err, "Permission denied: $f";
		}
	}
	for my $f ( @{$DBS->{'filters'}} ) {
		if ( ! -e $f ) {
			push @$err, "Filter not found: $f";
			next;
		}
		if ( ! -x $f ) {
			push @$err, "Permission denied: $f (cannot execute)";
		}
	}
	return __print_errors($err);
}

sub structure($;$) {
	my ( $DBS, $outer_dbh ) = @_;
	my $name = $DBS->{'name'};

	my $ok = 1;
	if ( ! check_structure_options($DBS) ) {
		__error "Cannot perform structure control";
		$ok = 0;
	}
	if ( ! check_connexion($DBS, $outer_dbh) ) {
		__error "Cannot connect to the databse";
		$ok = 0;
	}
	if ( ! $ok ) {
		die "Cannot perform structure action";
	}

	read_files($DBS);

	if ( $DBS->{'force'} ) {
		if ( $dbh->{'suffix'} ) {
			die "You did ask a FORCE structure check on a multi-instances databse. It would drop everything, and then re-create the multi-instanciated tables, under their short name only, and thus destroy everything and leave only unusable tables. Sounds insane to me. Maybe you do need a medic?";
		}
		__info "FORCE Building the database";
		purge_db($dbh, $name);
		__info "Create $name";
		execute_queries($dbh,$DBS->{'queries'});
		return;
	}

	__debug "-- Checking the database";
	if ( $dry_table || $dry_alter || $dry_data ) {
		__debug "-- Check database $name...";
		__debug "use $name;";
	}
	my $curver = get_current_versions($dbh);
	check_db_structure($dbh, $DBS, $curver);
	check_db_content  ($dbh, $DBS, $curver);
	check_limits($dbh,$DBS->{'limits'}->{'post'},$curver);
}

sub purge($;$) {
	my ( $DBS, $outer_dbh ) = @_;
	my $name = $DBS->{'name'};

	my $ok = 1;
	if ( ! check_structure_options($DBS) ) {
		__error "Cannot perform structure control";
		$ok = 0;
	}
	if ( ! check_connexion($DBS, $outer_dbh) ) {
		__error "Cannot connect to the databse";
		$ok = 0;
	}
	if ( ! $ok ) {
		die "Cannot perform structure action";
	}

	if ( ! $dbh->{'suffix'} ) {
		die "You did ask me to purge a complete database. For that I am useless, use 'drop db' in mysql client. I am supposed to drop client-related instances in a multi-db system, nothing else.";
	}

	read_files($DBS);
	purge_db_structure($dbh, $DBS);
}

##############################################################################################
# About the script itself (options and so on)
##############################################################################################

sub check_db($;$) {
	my ( $DBS, $action ) = @_;
	my @actions;
	if ( defined $action ) {
		@actions = ( $action );
	} else {
		@actions = @{$DBS->{'actions'}};
	}

	for my $act ( @actions ) {
		__debug "Processing action $act";
		if ( $act eq 'structure' ) {
			structure($DBS);
			next;
		}
		if ( $act eq 'create' ) {
			create($DBS);
			next;
		}
		if ( $act eq 'config' ) {
			config($DBS);
			next;
		}
		die "Unknown action <$act> to perform";
	}
}

sub switch_option($$$) {
	my ( $groupe, $on, $select ) = @_;
	my $grps = {
		'dry' => {
			'all' => [ \$dry_db, \$dry_table, \$dry_alter, \$dry_data, \$dry_cfg ],
			'db' => [ \$dry_db ],
			'table' => [ \$dry_table ],
			'alter' => [ \$dry_alter ],
			'data' => [ \$dry_data ]
		}
	};
	die if ! defined $grps->{$groupe};
	die if ! defined $grps->{$groupe}->{$select};
	for my $opt ( @{$grps->{$groupe}->{$select}} ) {
		$$opt = $on;
	}
	return;
}

sub set_quiet($$) {
	my ( $option, $s ) = @_;
	$quiet = $s;
}

sub set_output($) {
	$output = $_[0];
}



sub read_config($$;$) {
	my ( $file, $section, $DBS ) = @_;
	$DBS //= {
		'name' => $section
	};
	$DBS->{'files'}   //= [];
	$DBS->{'filters'} //= [];
	$DBS->{'actions'} //= [];
	$file //= $DBS->{'config_file'};
	$section //= $DBS->{'config'};

	my $cfg = Config_Section($file, $section);
	die "Config not found is $file [$section]" if ! defined $cfg;
	die "Only the 'mysql' type is supported so far" if ! defined $cfg->{'type'} || $cfg->{'type'} ne 'mysql';

	$DBS->{'host'}    = $cfg->{'host'}  if defined $cfg->{'host'};
	$DBS->{'source'}  = $cfg->{'source'} if defined $cfg->{'source'};
	$DBS->{'name'}    = $cfg->{'name'} if defined $cfg->{'name'};
	$DBS->{'login'}   = $cfg->{'login'} if defined $cfg->{'login'};
	$DBS->{'passwd'}  = $cfg->{'passwd'} if defined $cfg->{'passwd'};
	$DBS->{'utf8'}    = $cfg->{'utf8'} if defined $cfg->{'utf8'};
	push @{$DBS->{'files'}},   @{$cfg->{'@files'}}   if defined $cfg->{'@files'};
	push @{$DBS->{'filters'}}, @{$cfg->{'@filters'}} if defined $cfg->{'@filters'};
	push @{$DBS->{'actions'}}, @{$cfg->{'@actions'}} if defined $cfg->{'@actions'};
	if ( $cfg->{'@modules'} ) {
		$DBS->{'modules'} = {};
		for my $module ( @{$cfg->{'@modules'}} ) {
			my $prev = [ keys %{$DBS->{'modules'}} ];
			for my $already ( @$prev ) {
				my $name = "$already,$module";
				my $hash = $DBS->{'modules'}->{$already};
				my $res  = {};

				my $files = [];
				if ( $hash->{'files'} ) {
					push @$files, @{$hash->{'files'}};
				}
				if ( $cfg->{'@files-'.$module} ) {
					push @$files, @{$cfg->{'@files-'.$module}};
				}
				$res->{'files'} = $files if @$files;

				my $filters = [];
				if ( $hash->{'filters'} ) {
					push @$filters, @{$hash->{'filters'}};
				}
				if ( $cfg->{'@filters-'.$module} ) {
					push @$filters, @{$cfg->{'@filters-'.$module}};
				}
				$res->{'filters'} = $filters if @$filters;

				my $actions = [];
				if ( $hash->{'actions'} ) {
					push @$actions, @{$hash->{'actions'}};
				}
				if ( $cfg->{'@actions-'.$module} ) {
					push @$actions, @{$cfg->{'@actions-'.$module}};
				}
				$res->{'actions'} = $actions if @$actions;

				$DBS->{'modules'}->{$name} = $res;
			}
			my $res = {};
			my $files = [ @{$DBS->{'files'}} ];
			if ( $cfg->{'@files-'.$module} ) {
				push @$files, @{$cfg->{'@files-'.$module}};
			}
			$res->{'files'} = $files;

			my $filters = [ @{$DBS->{'filters'}} ];
			if ( $cfg->{'@filters-'.$module} ) {
				push @$filters, @{$cfg->{'@filters-'.$module}};
			}
			$res->{'filters'} = $filters;

			my $actions = [ @{$DBS->{'actions'}} ];
			if ( $cfg->{'@actions-'.$module} ) {
				push @$actions, @{$cfg->{'@actions-'.$module}};
			}
			$res->{'actions'} = $actions;

			$DBS->{'modules'}->{$module} = $res;
		}
		$DBS->{'modules_list'} = [ @{$cfg->{'@modules'}} ];
	}
	return $DBS;
}


### #############################################################################################
### # To check the texts in the localisation database.     XXX DEPRECATED
### #############################################################################################
### 
### sub check_text($$$$) {
### 	my ( $dbh, $appli, $aid, $t ) = @_;
### 	#ID SIMPLE DESC VAL
### 	#print "Je vrifie le texte $appli/$aid/".$t->{'TEXT_ID'}." (".$t->{'TEXT_VAL'}.")\n";
### 	my $text = &$text_par_xid_appli_id($t->{'TEXT_ID'},$aid);
### 	if ( defined $text ) {
### 		if ( $text->{'TEXT_VAL'} ne $t->{'TEXT_VAL'} ) {
### 			&$set_text_val($text->{'TEXT_ID'},$t->{'TEXT_VAL'});
### 		}
### 		if ( $text->{'TEXT_SIMPLE'} ne $t->{'TEXT_SIMPLE'} ) {
### 			&$set_text_simple($text->{'TEXT_ID'},$t->{'TEXT_SIMPLE'});
### 		}
### 	} else {
### 		__info "Il faut crer le texte, application: $appli/$aid, ".$t->{'TEXT_VAL'};
### 		my $tid = &$create_text($aid, $t->{'TEXT_ID'}, $t->{'TEXT_VAL'}, $t->{'TEXT_SIMPLE'}, $t->{'TEXT_DESC'});
### 		if ( ! $tid ) {
### 			die "Failed to create text\n";
### 		}
### 	}
### }
### 
### sub check_texts($$) {
### 	my ( $dbh, $appli ) = @_;
### 	my $sql = $texts->{$appli};
### 	my @queries = split /;\s*$/m, $sql;
### 	my $aid = &$appli_id_par_name($appli);
### 	if ( ! defined $aid ) {
### 		__error "-- Failed to get application $appli, need to insert it";
### 		if ( ! $dry_data ) {
### 			$aid = &$new_appli($appli);
### 		}
### 		if ( ! $aid ) {
### 			# FIXME If in dry run, normal failure (ask for --run)
### 			die "Failed to insert row in table APPLICATION\n";
### 		}
### 	}
### 	__info "-- On application '$appli' / <$aid>";
### 	for my $q ( @queries ) {
### 		next if ( $q !~ /^\s*insert into TEXT/ );
### 		my $t = {};
### 		if ( $q !~ /\s*insert\s+into\s+TEXT\s+
### 				\( \s*(\w+)\s*,   \s*(\w+)\s*,	 \s*(\w+)\s*,	\s*(\w+)\s*  \)
### 				\s* values \s*
### 				\( \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s* \)\s*$/xs ) {
### 			die "Invalid insert in text database for appli $appli: $q\n";
### 		}
### 		$t->{$1} = $5;
### 		$t->{$2} = $6;
### 		$t->{$3} = $7;
### 		$t->{$4} = $8;
### 		$t->{'TEXT_VAL'} =~ s/\\"/"/gs;
### 		check_text( $dbh, $appli, $aid, $t );
### 	}
### }
### 
### sub check_all_the_texts() {
### 	my $login = $connect->{'localisation'}->[0];
### 	my $passwd = $connect->{'localisation'}->[1];
### 	my $opts = { 'PrintError' => 0 };
### 	$opts->{'utf8'} = $connect->{'localisation'}->[2] if defined $connect->{'localisation'}->[2];
### 	my $tmp;
### 	my $dbh = connect_db('mysql','localhost','localisation',$login,$passwd,$opts);
### 	if ( ! $dbh ) {
### 		die "Failed to connect to localisation databse\n";
### 	}
### 	__info "-- Checking the texts\n";
### 	for my $appli ( keys %$texts ) {
### 		print
### 		check_texts($dbh,$appli);
### 	}
### 	disconnect_db($dbh);
### }
### 
### sub utf8_clean($) {
### 	my ( $hash ) = @_;
### 	for my $k ( %$hash ) {
### 		Encode::_utf8_on($hash->{$k});
### 	}
### }

1;
