# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
package FDN::Sql::Workflow::Run;

use strict;
use Exporter ;

our @ISA = ( 'Exporter' ) ;

our @EXPORT = qw (
	launch_task
	send_event
	wait_event
	run_task
	purge_task
	get_task_list
	get_task_infos

	workflow_select_db
	workflow_select_config
);

use FDN::Common::Constantes;
use FDN::Common::Logger;
use FDN::Common::Config;
use FDN::Common::Modules;
use FDN::Sql::Workflow::Basic;
use Data::Dumper;

my $tasks = {};
my $config = 'wf.cfg';
my $modules = {};

sub workflow_select_config($) { # Test
	$config = $_[0];
	my $tmp = Load_Config($config);
	if ( ! defined $tmp ) {
		return FALSE;
	}
	return TRUE;
}

our $debug = 0;
sub __debug($) { # Test
	return if ! $debug;
	debug $_[0];
}

########################################################
## Section 1:
## Le fichier de configuration (lecture et stockage)
########################################################

# Retourne un AREF sur la liste des taches possibles
sub get_task_list() { # Test
	return Config_Sections_List($config);
}

sub get_task_infos($);
sub get_task_infos($) { # Test
	my ( $name ) = @_;
	if ( defined $tasks->{$name} ) {
		__debug "Getting task type '$name' from cache.";
		return $tasks->{$name};
	}
	
	my $cfg = Config_Section($config,$name);
	if ( ! defined ($cfg) ) {
		error "Task type '$name' not defined.";
		return undef;
	}
	__debug "Task type '$name' is defined.";
	
	my $infos = {};
	$infos->{'name'} = $name;
	
	if ( ! defined ($cfg->{'type'}) ) {
		die "La tache $name n'a pas de type";
	}
	$infos->{'type'} = $cfg->{'type'};
	__debug "La tache est de type ".$infos->{'type'}; 
	
	if ( ! defined ($cfg->{'keycalc'}) ) {
		die "La tache $name n'a pas de keycalc";
	}
	my $keycalc = $cfg->{'keycalc'};
	__debug "La methode de calcul de la clef est $keycalc";
	if ( $keycalc ne 'CAT' ) {
		__debug "La clef n'est pas calculee par concatenation";
		if ( $keycalc =~ /^&/ ) {
			$keycalc =~ s/^.//;
			$keycalc = get_foreign_function($keycalc);
			if ( ref($keycalc) ne 'CODE' ) {
				die "Le keycalc de la tache $name n'est pas valide";
			}
			__debug "Valeur de keycalc: $keycalc";
		} else {
			die "Valeur de keycalc incorrecte pour $name";
		}
	}
	$infos->{'keycalc'} = $keycalc;
	
	
	if ( ! defined ($cfg->{'action'}) ) {
		die "La tache $name n'a pas d'action associee";
	}
	my $action = $cfg->{'action'};
	__debug "L'action est ".$action;
	if ( $action =~ /^&/ ) {
		$action =~ s/^.//;
		$action = get_foreign_function($action);
		if ( ref($action) ne 'CODE' ) {
			error "La fonction de $name : $action est indefinie";
			die "";
		}
	} else {
		error "L'action $action est invalide";
		die "";
	}
	$infos->{'action'} = $action;
	
	$infos->{'in'} = [];
	$infos->{'byname'} = {};
	if ( defined ($cfg->{'@in'}) ) {
		for my $k ( @{$cfg->{'@in'}} ) {
			if ( ! defined($cfg->{$k}) ) {
				die "Le parametre d'entree $k de $name n'est pas defini";
			}
			my ( $opt, $syn ) = split /:/,$cfg->{$k};
			my $v = { 'opt' => $opt, 'syn' => $syn, 'name' => $k };
			push @{$infos->{'in'}}, $v;
			$infos->{'byname'}->{$k} = $v;
		}
	} else {
		__debug "Cette tache n'a pas de parametres d'entree";
	}
	
	$infos->{'out'} = [];
	if ( defined ($cfg->{'@out'}) ) {
		for my $k ( @{$cfg->{'@out'}} ) {
			if ( ! defined($cfg->{$k}) ) {
				die "Le parametre de sortie $k de $name n'est pas defini";
			}
			my ( $opt, $syn ) = split /:/,$cfg->{$k};
			my $v = { 'opt' => $opt, 'syn' => $syn, 'name' => $k };
			push @{$infos->{'out'}}, $v;
			$infos->{'byname'}->{$k} = $v;
		}
	} else {
		__debug "Cette tache n'a pas de parametres de sortie";
	}
	
	$infos->{'launch'} = [];
	if ( defined ($cfg->{'@launch'}) ) {
		if ( $infos->{'type'} eq 'atomic' ) {
			die "La tache $name de type atomic a des taches subordonnees.";
		}
		for my $t ( @{$cfg->{'@launch'}} ) {
			my $tmp = get_task_infos($t);
			if ( !defined($tmp) ) {
				error "La tache subordonnee $t de $name n'existe pas";
				return undef;
			}
			
			my $in = {};
			if ( defined ($cfg->{'@'.$t.'-in'}) ) {
				$in = get_subtask_in($cfg->{'@'.$t.'-in'}, $name, $t);
			}
			
			my $out;
			if ( defined ($cfg->{'@'.$t.'-in'}) ) {
				$out = get_subtask_out($cfg->{'@'.$t.'-out'}, $name, $t);
			}
			push @{$infos->{'launch'}}, {
						'name' => $t,
						'in' => $in,
						'out' => $out };
		}
	} else {
		if ( $infos->{'type'} ne 'atomic' ) {
			 error "La tache $name, de type ".$infos->{'type'}." n'a pas de tache subordonnee.";
			 die "";
		}
	}
	$tasks->{$name} = $infos;
	return $infos;
}

sub get_subtask_in($$$) { # Test
	my ( $params, $name, $subname ) = @_;
	my $in = {};
	for my $i ( @$params ) {
		if ( $i =~ /^\s*(\S.*\S)\s*=>\s*(\S.*\S|\S)\s*$/ ) {
			my ( $from, $to ) = ( $1, $2 );
			if ( $from =~ /^&(.*)/ ) {
				$from =~ /^&(.*)$/;
				$from = get_foreign_function($1);
				if ( ref($from) ne 'CODE' ) {
					error "La fonction $from utilisee pour le parametre d'entree $to de la sous-tache $subname de $name n'est pas definie";
					$from = 'INVALID';
				}
			}
			$in->{$to}=$from;
		} else {
			error "Syntaxe invalide: le champs @".$subname."-in ne peux pas contenir ".$i;
		}
	}
	return $in;
}

sub get_subtask_out($$$) { # Test
	my ( $params, $name, $subname ) = @_;
	my $out = {};
	for my $i ( @$params ) {
		if ( $i =~ /^\s*(\S.*\S)\s*=>\s*(\S.*\S|\S)\s*$/ ) {
			my ( $from, $to ) = ( $1, $2 );
			$out->{$from}=$to;
		} else {
			error "Syntaxe invalide: le champs @".$subname."-out ne peux pas contenir ".$i;
		}
	}
	return $out;
}


########################################################
## Section 2:
## Fonctions utilitaires portant sur les taches
########################################################

# Verifie les variables initiales d'une tache d'apres
# le fichier de conf:
# - name est le nom de la tache (la section du fichier de conf)
# - infos est la description formelle de la tache
# - vars est le hash avec les variables de la tache
sub check_start_vars($$$) { # Test
	my ( $name, $infos, $vars ) = @_;
	
	my $key=$name;
	for my $k ( @{$infos->{'in'}} ) {
		my $opt = $k->{'opt'};
		my $syn = $k->{'syn'};
		my $param = $k->{'name'};
		__debug "On verifie le parametre $param";
		
		if ( $opt eq 'mandatory' ) {
			if (!defined $vars->{$param}) {
				error "Pas de parametre $param en entree de $name";
				return FALSE;
			}
			__debug "Le parametre $param obligatoire est present";
			
			if ( $infos->{'keycalc'} eq 'CAT' ) {
				$key.='/'.$vars->{$param};
				__debug "La clef devient $key";
			}
		}
		if ( defined $vars->{$param} ) {
			if ( $vars->{$param} !~ /$syn/ ) {
				error "Valeur invalide pour $param en entree de $name ($syn)";
				return FALSE;
			}
			__debug "Le parametre $param est conforme a $syn";
		}
	}
	
	if ( $infos->{'keycalc'} ne 'CAT' ) {
		my $fct = $infos->{'keycalc'};
		$key = &$fct($name,$vars);
		__debug "Valeur de la clef: $key";
	}
	return $key;
}

# Re-injecte dans les variables du pere les variables de sortie
# du fils qui doivent l'etre. Ca a lieu lors de la fin du fils,
# avant que le pere ne recoive la notification de cette mort.
# - infos: description du pere
# - props: variables du pere
# - props_fils: variables du pere
# - num: numero d'ordre du fils
# Retourne un boolean disant si les proprietes du pere ont ete alterees
# (si oui, il faut updater).
sub traite_sortie_fils($$$$) { # Test
	my ( $infos, $props, $props_fils, $num ) = @_;
	my $out = ${$infos->{'launch'}}[$num]->{'out'};
	my $clean=TRUE;
	for my $o ( keys %$out ) {
		my $dst = $out->{$o};
		$props->{$dst} = $props_fils->{$o};
		my $tmp = '<undef>';
		$tmp = $props_fils->{$o} if defined $props_fils->{$o};
		__debug "Propagation de ".$o."=".$tmp.
			" vers papa sous le nom ".$dst;
		$clean=FALSE;
	}
	return $clean;
}

# Cree les variable du fils d'apres ce que dit la conf (variables
# du pere, fonctions, constantes, etc). Ca a lieu, bien entendu,
# avant le lancement du fils. On ne verifie pas que les conditions
# de lancement sont reunies, on devrait peut-etre.
# - infos: description du pare
# - props: variables du pere
# - num: numero d'ordre du fils
# Retourne les futures proprietes du fils.
sub prepare_entree_fils($$$) { # Test
	my ( $infos, $props, $num ) = @_;
	my $in = ${$infos->{'launch'}}[$num]->{'in'};
	my $props_fils = {};
	__debug "Preparation des entrees du fils numero $num de ".$infos->{'name'};
	__debug "Variables du pere: ".Dumper($props);
	for my $i ( keys %$in ) {
		my $val = $in->{$i};
		if ( ref($val) eq 'CODE' ) {
			__debug "Parametre $i est une fonction";
			$val = &$val($props);
		} elsif ( $val =~ /^\$(.*)$/ ) {
			my $src = $1;
			__debug "Parametre $i est une variable ($val soit $1)";
			if ( defined ($props->{$src}) ) {
				$val = $props->{$1};
			} else {
				error "Cette clef n'est pas presente dans les proprietes";
				$val = 'INVALID';
			}
		} else {
			__debug "Parametre $i est une constante ($val)";
		}
		$props_fils->{$i} = $val;
	}
	return $props_fils;
}

# Positionne une tache a l'etat fail et purge ses executions futures
# et ses attentes d'evenements. On devrait peut-etre verouiller les
# 'event' pour etre sur qu'un RUN n'apparaisse pas pendant qu'on purge...
# Il ne faut pas: il faudrait poser le lock quand la tache decide d'echouer
# et retirer le lock quand son echec est effectif. Soit on pose le lock dans
# la tache elle-meme (trop dangereux!), soit on pose le lock pendant toute son
# execution (trop long!).
# Conclusion: quand une tache echoue, il se peut qu'elle recoive encore des
# evennements (des vrais, pas 'init' ou 'end').
# - tid: la tache a echouer
sub __fail_task($) { # Test
	my ( $tid ) = @_;
	# Echec de la tache
	set_task_state($tid,'failed');
	# Suppression des executions futures
	my @runs = run_par_task($tid);
	for my $r (@runs) {
		next if (!defined($r));
		next if ( $r->{'RUN_STATE'} ne 'free');
		__debug "Suppression du run ".$r->{'RUN_ID'};
		del_run($r->{'RUN_ID'});
	}
	# Suppression des evenements attendus
	my @waits = waiting_par_task($tid);
	for my $w (@waits) {
		next if (!defined($w));
		__debug "Suppression du waiting ".$w->{'WAITING_ID'};
		del_waiting($w->{'WAITING_ID'});
	}
}

# Place une tache en attente d'un evenement. S'il est deja survenu,
# elle est ajoutee dans 'RUN' et son etat passe a 'willrun', sinon
# elle est ajoutee dans 'WAITING' et son etat passe a 'wait-event'.
# - tid: la tache
# - event: l'evenement
# Retour:
# - Vrai si la tache est effectivement mise en attente (creation d'un
#   waiting).
# - Faux sinon: l'evennement est deja survenu, et on cree un nouveau
#   run).
sub wait_event($$) { # Test
	my ( $tid, $event ) = @_;
	my $lck = lock_event();
	__debug "wait_event got lock for $event";
	if ( !$lck ) {
		#FIXME Que faire en pareil cas?
		error "Pas obtenu le verrou soft en 10 secondes!";
		return undef;
	}
	my $eid = event_par_event($event);
	my $id;
	if ( defined($eid) ) {
		# Evenement deja survenu
		$id = new_run($tid,$event);
		__debug "Event $event happened, so new run $id. Release lock.";
		unlock_event();
		return FALSE;
	} else {
		# Evenement attendu
		$id = new_waiting($tid,$event);
		__debug "Event $event not yet happened, so new waiting $id. Release lock.";
		unlock_event();
		return TRUE;
	}
}

# Survenue d'un evennement. Si quelqu'un etait en attente de cet
# evennement, il passe de la table WAITING a la table RUN. Son
# etat devrait passer a willrun.
sub send_event($;$) { # Test
	my ( $event, $props ) = @_;
	my $lck = lock_event();
	__debug "send_event got lock for $event";
	if ( !$lck ) {
		#FIXME Que faire en pareil cas?
		error "Pas obtenu le verrou soft en 10 secondes!";
		return FALSE;
	}
	my $eid = new_event($event);
	if ( defined $props && ref($props) eq 'HASH') {
		update_event_properties($eid,$props);
	}
	__debug "Just created event $event. Release lock.";
	unlock_event();
	my @wait = waiting_par_event($event);
	for my $w ( @wait ) {
		next if ( ! defined($w) );
		new_run($w->{'TASK_ID'},$event);
		del_waiting($w->{'WAITING_ID'});
	}
	return TRUE;
}

sub __try_end_task($) { # Test
	my ( $tid ) = @_;
	my $waited = waiting_par_task($tid);
	my $run = run_par_task($tid);
	if ( !defined $waited && !defined $run ) {
		# Creer un run sur evenement 'end'
		__debug "Envoi du pseudo-evennement 'end";
		new_run($tid,'end');
	} else {
		__debug "Encore au moins un evenement en attente";
	}
}

sub launch_task($$);
sub __launch_subtask($$$$) { # Test
	my ($tid, $infos, $props, $i) = @_;
	my $f_vars = prepare_entree_fils($infos,$props,$i);
	my ($fid, $fkey) = launch_task($infos->{launch}[$i]->{'name'}, $f_vars);
	new_taskself($tid,$fid,$i+1);
	wait_event($tid,'EOT-'.$fkey);
	return $fid;
}

########################################################
## Section N:
## Fonctions de haut niveau
########################################################

# Lancement d'une nouvelle tache: la tache est creee, et
# mise dans 'RUN' sur l'evenement 'init' pour sa premiere
# execution.
# - name: nom de la tache (section de la conf)
# - vars: ses variables initiales
# Retour: l'id de la tache ou le couple ( tid, key ).
sub launch_task($$) { # Test
	my ( $name, $vars ) = @_;

	my $infos = get_task_infos($name);
	if ( ! defined $infos ) {
		error "Cannot launch task $name: undefined task";
		return undef;
	}

	my $key = check_start_vars($name,$infos,$vars);
	if ( ! $key ) {
		error "N'a pas pu lancer la tache $name";
		return undef;
	}
	
	my $tid = new_task($infos->{'type'}, $name, $key);
	if ( ! defined($tid) ) {
		my $task = task_par_key($key);
		if ( defined($task) ) {
			$tid = $task->{'TASK_ID'};
			return $tid;
		} else {
			error "Echec incomprehensible";
			return undef;
		}
	}
	update_task_properties($tid,$vars);
	new_run($tid,'init');
	if ( wantarray ) {
		return ( $tid, $key );
	} else {
		return $tid;
	}
}

sub __propage_fils($$$) { # Test
	my ( $tid, $infos, $event ) = @_;
	if ( $infos->{'type'} eq 'atomic' ) {
		# La propagation n'a pas de sens pour les taches atomiques.
		return TRUE;
	}
	# Les events FEOT-.* qui sont des faux ne sont pas pris en compte:
	# la propagation a deja eu lieu.
	__debug "Tache non atomique, fin d'un fils, il faut propager";
	if ( $event !~ /^EOT-/ ) {
		# L'evennement n'est pas la fin d'un fils.
		return TRUE;
	}
	
	# Remonter des infos sur le fils.
	$event =~ /^EOT-(.*)$/;
	my $key = $1;
	my $fils = task_par_key($key);
	if ( !defined $fils ) {
		error "La tache qui vient de se terminer (event ".$event.") est introuvable.";
		return FALSE;
	}
	my $fid = $fils->{'TASK_ID'};
	my $prop_fils = get_task_properties($fid);
	
	# Trouver le numero du fils:
	# Chercher l'entree dans TASKSELF correspondante:
	my $taskself = taskself_par_parent_fils($tid,$fid);
	if ( !defined $taskself ) {
		error "Je ne trouve pas le lien de parente";
		return FALSE;
	}
	# Attention, la base numerote avec un de plus que perl...
	my $num=$taskself->{'TASKSELF_ORDER'} - 1;

	# Prendre les proprietes du pere
	my $props = get_task_properties($tid);
	
	# Propager les infos, et archiver si necessaire.
	__debug "Debut de la propagation";
	my $still_clean =  traite_sortie_fils($infos,$props,$prop_fils,$num);
	__debug "Fin de la propagation";
	if ( !$still_clean ) {
		__debug "Il y a de nouvelles infos a sauver";
		update_task_properties($tid,$props);
	}
	return TRUE;
}

sub __identification_tache($$) { # Test
	my ( $tid, $props ) = @_;

	if ( ! defined $props->{'tid'} ) {
		$props->{'tid'} = $tid;
	}
	if ( ! defined $props->{'ptid'} ) {
		my $ts = taskself_par_fils($tid);
		if ( defined $ts ) {
			$props->{'ptid'} = $ts->{TASK_ID};
		} else {
			$props->{'ptid'} = -1;
		}
	}
}

sub __traite_sortie_tache($$) { # Test
	my ( $tid, $ret ) = @_;
	if ( !defined($ret) ) {
		__debug "Retour undef -> echec";
		__fail_task($tid);
		return FALSE;
	}
	if ( $ret ) {
		__debug "La tache demande l'attente de l'evennement $ret";
		wait_event($tid,$ret);
	}
	return TRUE;
}

sub __find_last_son($$) { # Test
	my ( $tid, $guess ) = @_;
	__debug "Recherche du dernier fils connu de $tid. C'est probablement le fils numero $guess.";
	$guess++;
	my $son;
	my $son_id = -1;
	do {
		$guess--;
		goto END if ( $guess == 0 );
		$son = taskself_par_parent_numero($tid,$guess);
		my $tmp = 'undef';
		$tmp = $son if defined $son;
		__debug "En supposant que le dernier fils soit le numero $guess, je trouve $tmp";
	} while ( ! defined $son );
	END:
	if ( defined $son ) {
		$son_id = $son->{'SELF_ID'};
	}
	return ( $guess, $son_id );
}

sub __post_run_serial($$$$) { # Test
	my ( $task, $infos, $props, $event ) = @_;
	my $tid = $task->{'TASK_ID'};
	if ( $event eq 'init' ) {
		# Si je viens de terminer mon init, je lance mon premier fils.
		__debug "L'init de $tid s'est bien passe. Lancement du premier fils.";
		__launch_subtask($tid, $infos, $props, 0);
		set_task_nbchild($tid, 1);
		return;
	}
	if ( $event !~ /^F?EOT-(.*)$/ ) {
		__debug "Pas la fin d'un fils, donc pas de fils a lancer.";
		#FIXME Faut-il essayer de terminer la tache? Par exemple si on recoit
		# un evennement apres la mort du dernier fils...
		# Plus complique: si on a annule des taches (cas d'un echec) et qu'on
		# a envoye un FEOT-*, on n'est plus en train d'attendre le dernier fils.
		# Si on recoit un evennement quelconque, alors on n'attend rien de particulier,
		# le __try_end_task arriverait *faussement* a la conclusion qu'il n'y a plus
		# rien a faire.
		# Reponse: il reste un RUN sur FEOT-*. Comme quoi try_end_task
		# ne doit jouer que s'il n'y a plus de RUN!
		__try_end_task($tid);
		return;
	}
	# On s'interesse aux vraies fins et aux fausses fins pour reprise
	# sur erreur.
	$event =~ /^F?EOT-(.*)$/;
	my $ekey = $1;
	my $eid  = task_id_par_key($ekey);
	if ( ! defined $eid ) {
		error "J'ai recu la fin d'une tache qui n'existe pas. Merveilleux!";
		$eid = -2;
	}
	my $childmax = $#{$infos->{'launch'}}+1;
	my $child = $task->{'TASK_NBCHILD'};
	my $child_id = -1;
	
	( $child, $child_id) = __find_last_son($tid,$child);
	if ( $eid != $child_id ) {
		my $tmp = '<undef>';
		$tmp = $child_id if defined $child_id;
		__debug "J'ai recu la fin d'une tache qui n'est pas mon dernier fils connu. ".
			"J'attends encore la tache $tmp. Celle qui vient de finir est ".
			"$eid, clef: <$ekey>.";
		return;
	}
	if ( $child == $childmax ) {
		__debug "C'est la fin du dernier de mes fils. Plus d'autre a lancer.";
		__try_end_task($tid);
		return;
	}

	__debug "Il me reste un fils au moins a lancer";
	__launch_subtask($tid, $infos, $props, $child);
	set_task_nbchild($tid,$child+1);
}

sub __post_run_parallel($$$$) { # Test
	# FIXME: Si un des fils disparait a un moment, il n'y a pas de mecanique
	# de reprise sur erreur. Il faudrait peut-etre prevoir que sur un evennement
	# de fin on fasse un check-up des fils: s'il en manque il faut les relancer,
	# s'il y en a des echoues, il faut prevenir, etc.
	my ( $task, $infos, $props, $event ) = @_;
	my $tid = $task->{'TASK_ID'};
	if ( $event eq 'init' ) {
		# Si on sort de 'init', lancer tous les fils en parallele.
		my $i=0;
		for my $l ( @{$infos->{'launch'}} ) {
			__launch_subtask($tid, $infos, $props, $i);
			$i++;
		}
		set_task_nbchild($tid,$i);
		return;
	}
	if ( $event =~ /^EOT-/ ) {
		# Si on sort de la fin d'un des fils, chercher s'il y a d'autres
		# evenements attendus (dont la fin d'un autre fils).
		__try_end_task($tid);
		return;
	}
}

sub __post_run_atomic($$$$) { # Test
	my ( $task, $infos, $props, $event ) = @_;
	__try_end_task($task->{'TASK_ID'});
}

sub __post_run($$$$) { # Test
	my ( $task, $infos, $props, $event ) = @_;
	# Si c'est tout a fait termine.
	if ( $event eq 'end' ) {
		__debug "C'est la derniere execution d'une tache, on envoie EOT";
		send_event('EOT-'.$task->{'TASK_KEY'});
		set_task_state($task->{'TASK_ID'},'done');
		return;
	}
	# Si c'est le tout debut.
	if ( $event eq 'init' ) {
		__debug "C'est la premiere execution d'une tache. Elle passe 'running'.";
		set_task_state($task->{'TASK_ID'},'running');
		# On continue sur le post-traitement specifique.
	}
	if ( $infos->{'type'} eq 'serial' ) {
		__post_run_serial($task, $infos, $props, $event);
	} elsif ( $infos->{'type'} eq 'parallel' ) {
		__post_run_parallel($task, $infos, $props, $event);
	} elsif ( $infos->{'type'} eq 'atomic' ) {
		__post_run_atomic($task, $infos, $props, $event);
	}
}

# Fonction de base du scheduller, prend un RUN et le joue.
# Le RUN correspondant est supprime definitivement avant que
# la tache ne se soit reelement executee, c'est peut-etre un
# probleme...
sub run_task($) { # Test
	my ( $run ) = @_;
	my $tid = $run->{'TASK_ID'};
	my $event = $run->{'RUN_EVENT'};
	my $date = $run->{'RUN_DATE'};
	
	my $task = task_par_id($run->{'TASK_ID'});
	__debug "Execution de la tache ".$tid." en reponse a l'evennement ".$event;
	if ( !defined($task) ) {
		return undef;
	}
	my $infos = get_task_infos($task->{'TASK_NAME'});
	if ( !defined($infos) ) {
		error "Ne peut executer la tache prevue: run ".$run->{'RUN_ID'}." le type de tache ".$task->{'TASK_TYPE'}." n'est pas defini (disparition de la conf?)";
		return undef;
	}
	
	__debug "Tache de type ".$infos->{'type'};
	if ( ! __propage_fils($tid, $infos, $event) ) {
		error "Il y a eu une erreur pendant la prise en compte du fils.";
		return undef;
	}
	my $props = get_task_properties($tid);
	
	# Permettre a la tache de se connaitre et de connaitre la tache avant.
	__identification_tache($tid,$props);

	# Execution pour de vrai.
	__debug "Lancement de la tache";
	my $ret = &{$infos->{'action'}}($props,$event,$date);
	update_task_properties($tid,$props);
	__debug "Fin de la tache";
	
	# Test sur sa valeur de retour
	if ( ! __traite_sortie_tache($tid, $ret) ) {
		# Echec de la tache. Fin de la partie.
		return undef;
	}
	# Il faut avoir supprime le RUN *avant* de faire le __post_run sinon
	# __try_end_task refusera de terminer les taches.
	# FIXME: si le __post_run echoue (cas type: ne peut pas lancer une sous-tache
	# parce que des variables d'entree sont invalides), alors le RUN a disparu, et
	# le post-traitement n'est pas fait. Il faudrait pouvoir re-prendre au __post_run
	# apres reparation, mais c'est pas tres simple a faire...
	del_run($run->{'RUN_ID'});
	__post_run($task, $infos, $props, $event);
	return 0;
	
}

sub purge_task($);
sub purge_task($) { # Test
	my ( $tid ) = @_;
	my $task = task_par_id($tid);
	# Si des fois elle etait en train de tourner, elle echoue
	if ( $task->{'TASK_STATE'} !~ /^(done|failed)$/ ) {
		__fail_task($tid);
	}
	# On supprime ses proprietes
	del_task_properties($tid);
	# On supprime sa fin
	purge_event_par_event('EOT-'.$task->{'TASK_KEY'});
	# On va chercher les fils
	my @fils = taskself_par_parent($tid);
	# On les supprime:
	for my $f ( @fils ) {
		next if ( !defined $f );
		my $fid = $f->{'SELF_ID'};
		del_taskself($tid,$fid);
		purge_task($fid);
	}
	# On va chercher les parents
	my @papa = taskself_par_fils($tid);
	# On s'en detache, mais on ne les purge pas.
	for my $p ( @papa ) {
		next if ( !defined $p );
		my $pid = $p->{'TASK_ID'};
		del_taskself($pid,$tid);
		# Papa a perdu un fils, on va decrementer
		my $papa = task_par_id($pid);
		set_task_nbchild($pid,$papa->{'TASK_NBCHILD'}-1);
	}
	# On va chercher les gens qui nous attendent
	my @waits = waiting_par_event('EOT-'.$task->{'TASK_KEY'});
	for my $w ( @waits ) {
		next if ( ! defined $w );
		del_waiting($w->{'WAITING_ID'});
	}
	# On degage la tache, maintenant
	del_task($tid);
}

TRUE;

